/*
 * HybServ TS Services, Copyright (C) 1998-1999 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 *
 * $Id: hybserv.c,v 3.4.0 1998/07/22 00:34:37 wnder Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <string.h>
#include <time.h>

#include "alloc.h"
#include "config.h"
#include "data.h"
#include "dcc.h"
#include "defs.h"
#include "hash.h"
#include "hybdefs.h"
#include "init.h"
#include "log.h"
#include "operserv.h"
#include "server.h"
#include "settings.h"
#include "show.h"
#include "sock.h"
#include "timer.h"
#include "timestr.h"

#ifdef HAVE_PTHREADS
#include <pthread.h>
#endif

/* unixtime of when services was started */
time_t                       TimeStarted;

/* offset from UTC time */
long                         gmt_offset;

struct NetworkInfo           *Network;

/* structure containing services' info */
struct MyInfo                Me;

/*
 * Set to 1 "ConnectBurst" seconds after a successful
 * hub connection, so we don't send bursts of client/clone
 * connections to +y users etc, during an initial connect.
 */
int                          SafeConnect = 0;

int
main(int argc, char *argv[])

{
#ifndef DEBUGMODE
  int pid; /* pid of this process */
#endif
  FILE *pidfile; /* to write our pid */
  uid_t uid, /* real user id */
        euid; /* effective user id */
#ifdef HAVE_PTHREADS
  pthread_attr_t attr;
  pthread_t selectid;
  pthread_t timerid;
#endif

  TimeStarted = time(NULL);

  fprintf(stderr,
    "\nHybServ TS services version %s by Patrick Alken (wnder@underworld.net)\n",
    hVersion);

  /*
   * Load SETPATH (settings.conf) - this must be done
   * before the config file is loaded, and before any
   * putlog() calls are made, since LogFile is specified
   * in settings.conf
   */
  if (LoadSettings(0) == 0)
  {
    fprintf(stderr, "Fatal errors encountered and logged parsing %s, exiting\n",
      SETPATH);
    return (0);
  }

  /*
   * If they run ./shownicks or ./showchans rather than ./hybserv
   * display nicknames/channels
   */
  if (strstr(argv[0], "shownicks"))
  {
  #ifdef NICKSERVICES
    ShowNicknames(argc, argv);
  #endif /* NICKSERVICES */
    return (0);
  }
  else if (strstr(argv[0], "showchans"))
  {
  #if defined(NICKSERVICES) && defined(CHANNELSERVICES)
    ShowChannels(argc, argv);
  #endif /* defined(NICKSERVICES) && defined(CHANNELSERVICES) */
    return (0);
  }

  putlog(LOG1, "HybServ TS services started");

  /*
   * Get the offset from GMT (London time)
   */
  gmt_offset = GetGMTOffset(TimeStarted);

#if 0
  tm_gmt = gmtime(&TimeStarted);
  gmt_offset = TimeStarted - mktime(tm_gmt);
#endif /* 0 */

  uid = getuid(); /* the user id of the user who ran the process */
  euid = geteuid(); /* the effective id (different if setuid) */

  if (!uid || !euid)
  {
    /*
     * It is setuid root or running as root
     */
    printf("Please do not run HybServ as root\n");
    putlog(LOG1, "Detected running with root uid, exiting");
    exit (0);
  }

  if (chdir(HPath) != 0)
  {
    fprintf(stderr,
      "HPath is an invalid directory, please check %s\n",
      SETPATH);
    putlog(LOG1, "Invalid HPath (%s), shutting down", HPath);
    exit(0);
  }

  /*
   * the Network list must be initialized before the config
   * file is loaded
   */
  InitLists();

  /* load server, jupe, gline, user, admin info */
  LoadConfig(ConfigFile);

  /* load nick/chan/memo/stat databases */
  LoadData();

#ifdef GLOBALSERVICES

  if (LogonNews)
  {
    Network->LogonNewsFile.filename = LogonNews;
    ReadMessageFile(&Network->LogonNewsFile);
  }

#endif /* GLOBALSERVICES */

  if (LocalHostName)
    SetupVirtualHost();

#ifndef DEBUGMODE

  pid = fork();
  if (pid == (-1))
  {
    printf("Unable to fork(), exiting.\n");
    exit(0);
  }
  if (pid != 0)
  {
    printf("Running in background (pid: %d)\n\n", pid);
    exit(0);
  }
  close(0);
  close(1);
  close(2);

#else

  printf("Entering debug mode\n");  

#endif /* !DEBUGMODE */


  /*
   * Signals must be set up after fork(), since the parent
   * exits
   */
  InitSignals();

  srandom(time(NULL));

  /* write our pid to a file */
  if ((pidfile = fopen(PidFile, "w")) == (FILE *) NULL)
    putlog(LOG1, "Unable to open %s", PidFile);
  else
  {
    fprintf(pidfile, "%i\n", (int) getpid());
    fclose(pidfile);
  }

  /* initialize tcm/user listening ports */
  InitListenPorts();

  /* initialize hash tables */
  ClearHashes(1);

#ifdef BLOCK_ALLOCATION
  InitHeaps();
#endif

  HubSock = NOSOCKET;
  CycleServers();

#ifdef HAVE_PTHREADS

  /*
   * This thread will call DoTimer() every second
   */
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
  pthread_create(&timerid, &attr, p_CheckTime, NULL);

#endif

  while (1)
  {
    /* enter loop waiting for server info */

  #ifdef HAVE_PTHREADS

    pthread_create(&selectid, NULL, (void *) &ReadSocketInfo, NULL);
    pthread_join(selectid, NULL);

  #else

    ReadSocketInfo();

  #endif /* HAVE_PTHREADS */

    if (Me.hub)
      SendUmode(OPERUMODE_Y,
        "*** Disconnected from %s",
        Me.hub->name);
    else
      SendUmode(OPERUMODE_Y,
        "*** Disconnected from hub server");

    if (currenthub)
      if (currenthub->realname)
      {
        MyFree(currenthub->realname);
        currenthub->realname = (char *) NULL;
      }

    close(HubSock); /* There was an error */
    HubSock = NOSOCKET;
    currenthub->connect_ts = 0;

    /* 
     * whenever HybServ connects/reconnects to a server, clear
     * users, servers, and chans
     */
    ClearUsers();
    ClearChans();
    ClearServs();
    /*
     * ClearHashes() must be called AFTER ClearUsers(),
     * or StatServ's unique client counts will be off since
     * cloneTable[] would be NULL while it was trying to find
     * clones
     */
    ClearHashes(0);

    PostCleanup();
  } /* while (1) */

  return (0);
} /* main() */
