/*
 * HybServ TS Services, Copyright (C) 1998-1999 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 *
 * $Id: misc.c,v 1.0.0 1999/09/04 20:56:13 wnder Exp $
 */

#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>

#include "alloc.h"
#include "client.h"
#include "conf.h"
#include "config.h"
#include "data.h"
#include "dcc.h"
#include "defs.h"
#include "hash.h"
#include "hybdefs.h"
#include "log.h"
#include "match.h"
#include "misc.h"
#include "operserv.h"
#include "settings.h"
#include "sock.h"
#include "Strn.h"

extern char *crypt();

/*
 * This is an array of all the service bots. It's useful
 * because it keeps code repitition down. For example, when
 * we want to introduce all the service nicks to the network
 * upon connection, we can just loop through this array, rather
 * than writing out all the introduce() calls. It's also useful
 * for the remote trace and whois queries.
 */

struct aService ServiceBots[] = {
  { &n_OperServ, &id_OperServ, &desc_OperServ, &Me.osptr },

#ifdef NICKSERVICES

  { &n_NickServ, &id_NickServ, &desc_NickServ, &Me.nsptr },

#ifdef CHANNELSERVICES

  { &n_ChanServ, &id_ChanServ, &desc_ChanServ, &Me.csptr },

#endif /* CHANNELSERVICES */

#ifdef MEMOSERVICES

  { &n_MemoServ, &id_MemoServ, &desc_MemoServ, &Me.msptr },

#endif /* MEMOSERVICES */

#endif /* NICKSERVICES */

#ifdef STATSERVICES

  { &n_StatServ, &id_StatServ, &desc_StatServ, &Me.ssptr },

#endif /* STATSERVICES */

#ifdef HELPSERVICES

  { &n_HelpServ, &id_HelpServ, &desc_HelpServ, &Me.hsptr },

#endif /* HELPSERVICES */

#ifdef GLOBALSERVICES

  { &n_Global, &id_Global, &desc_Global, &Me.gsptr },

#endif /* GLOBALSERVICES */

#ifdef SEENSERVICES

  { &n_SeenServ, &id_SeenServ, &desc_SeenServ, &Me.esptr },

#endif /* SEENSERVICES */

  { 0, 0, 0, 0 }
};

/*
debug()
  Output a debug message to stderr
*/

void
debug(char *format, ...)

#ifdef DEBUGMODE

{
  va_list args;

  va_start(args, format);

  fprintf(stderr, "DEBUG> ");
  vfprintf(stderr, format, args);

  va_end(args);
} /* debug() */

#else /* !DEBUGMODE */

{
  va_list args;
  char buf[MAXLINE];

  va_start(args, format);

  vSnprintf(buf, sizeof(buf), format, args);

  va_end(args);

  /* send the string to +d users */
  SendUmode(OPERUMODE_D, "DEBUG: %s",
    buf);
} /* debug() */

#endif /* DEBUGMODE */

/*
fatal()
  Fatal error occured - log it and exit
*/

void
fatal(int keepgoing, char *format, ...)

{
  char buf[MAXLINE];
  va_list args;
  int oldlev;

  va_start(args, format);

  vSnprintf(buf, sizeof(buf), format, args);

  va_end(args);

  /*
   * If this is being called from dparse(), it's possible
   * LogLevel hasn't been set yet. So, make sure this
   * error gets logged, whether LogLevel is set or not
   */
  oldlev = LogLevel;
  LogLevel = 1;

  /* log the error */
  if (!keepgoing)
  {
    putlog(LOG1, "FATAL: %s", buf);
    exit(1);
  }
  else
    putlog(LOG1, "Warning: %s", buf);

  LogLevel = oldlev;
} /* fatal() */

/*
notice()
  purpose: send a NOTICE to 'nick' with 'msg'
  return: none
*/

void
notice(char *from, char *nick, char *format, ...)

{
  char finalstr[MAXLINE];
  char who[MAXLINE];
  va_list args;

  va_start(args, format);

  vSnprintf(finalstr, sizeof(finalstr), format, args);

  if (ServerNotices)
    strcpy(who, Me.name);
  else
    strcpy(who, from);

  toserv(":%s NOTICE %s :%s\n", 
    who,
    nick,
    finalstr);

  va_end(args);
} /* notice() */

/*
DoShutdown()
  Shut down services
*/

void
DoShutdown(char *who, char *reason)

{
  char sendstr[MAXLINE];
  struct PortInfo *pptr;

  putlog(LOG1, "Shutting down services");

  if (!WriteDatabases())
    putlog(LOG1, "Database update failed");

#if defined(NICKSERVICES) && defined(CHANNELSERVICES)
  toserv(":%s QUIT :%s\n",
    n_ChanServ,
    "Shutting Down");
#endif

  /* close listening sockets */
  for (pptr = PortList; pptr; pptr = pptr->next)
    if (pptr->socket != NOSOCKET)
      close(pptr->socket);

  if (reason)
  {
    if (who)
      sprintf(sendstr, "%s (authorized by %s)", reason, who);
    else
      strcpy(sendstr, reason);
  }
  else
    if (who)
      sprintf(sendstr, "Authorized by %s", who);
    else
      sendstr[0] = '\0';

  toserv(":%s QUIT :%s\nSQUIT %s :%s\n",
    n_OperServ,
    "Shutting Down",
    currenthub->realname ? currenthub->realname : currenthub->hostname,
    sendstr);

  exit (1);
} /* DoShutdown() */

/* 
HostToMask()
  Convert a userhost like "user@host.name.org" into "*user@*.name.org"
or "user@1.2.3.4" into "*user@1.2.3.*"
*/

char *
HostToMask (char *username, char *hostname)

{
	char un[USERLEN+20];
	char result[UHOSTLEN+2+20];
	char c;
	char *p1, *p2;
	
	if (!username || !hostname)
		return ((char *) NULL);
	if (username[0] == '~')
		sprintf(un, "*%s", &username[1]);
	else
		sprintf(un, "*%s", username);
	p1=strchr(hostname, '.');
	p2=strrchr(hostname, '.');
	if (p1 == p2) { /* aa@b.com */
		sprintf(result, "%s@%s", un, hostname);
		return (MyStrdup(result));
	}	 
	c=hostname[strlen(hostname)-1];
	if ((c >= '0') && (c <= '9')) /* 1.2.3.* */
		sprintf(result, "%s@%.*s.*", un, p2-hostname, hostname);
	else /* *.x.y.z.com */
		sprintf(result, "%s@*%s", un, p1);
	return MyStrdup(result);
}

/*
Substitute()
  args: char *nick, char *str, int sockfd
  purpose: replace any formatting characters in 'str' with the
           corresponding information.

    Formatting characters:
      %O - Nickname of n_OperServ
      %N - Nickname of n_NickServ
      %C - Nickname of n_ChanServ
      %M - Nickname of n_MemoServ
      %T - Nickname of n_StatServ
      %H - Nickname of n_HelpServ
      %G - Nickname of n_Global
      %S - Name of this server
      %A - Administrative info
      %V - current version
      %B - bold character
      %+<flag> - needs <flag> to read the line

  return: NULL if its a blank line, -1 if 'nick' doesn't have privs
          to read the line; otherwise a ptr to substituted string
*/

char *
Substitute(char *nick, char *str, int sockfd)

{
  char tempstr[MAXLINE], key;
  char *finalstr = NULL;
  int tcnt, fcnt;
  struct Luser *lptr;

  lptr = FindClient(nick);

  strcpy(tempstr, str);
  finalstr = (char *) MyMalloc(MAXLINE);
  memset(finalstr, 0, MAXLINE);
  fcnt = 0;
  tcnt = 0;
  while (tcnt < MAXLINE)
  {
    if (IsEOL(tempstr[tcnt]))
      break;

    if (tempstr[tcnt] == '%')
    {
      key = tempstr[++tcnt];
      switch (key)
      {
        case 'o':
        case 'O':
        {
          strcat(finalstr, n_OperServ);
          fcnt += strlen(n_OperServ) - 1;
          break;
        }
        case 'n':
        case 'N':
        {
          strcat(finalstr, n_NickServ);
          fcnt += strlen(n_NickServ) - 1;
          break;
        }
        case 'c':
        case 'C':
        {
          strcat(finalstr, n_ChanServ);
          fcnt += strlen(n_ChanServ) - 1;
          break;
        }
#ifdef SEENSERVICES
        case 'e':
        case 'E':
        {
          strcat(finalstr, n_SeenServ);
          fcnt += strlen(n_SeenServ) - 1;
          break;
        }
#endif
        case 'm':
        case 'M':
        {
          strcat(finalstr, n_MemoServ);
          fcnt += strlen(n_MemoServ) - 1;
          break;
        }
        case 't':
        case 'T':
        {
          strcat(finalstr, n_StatServ);
          fcnt += strlen(n_StatServ) - 1;
          break;
        }
        case 'h':
        case 'H':
        {
          strcat(finalstr, n_HelpServ);
          fcnt += strlen(n_HelpServ) - 1;
          break;
        }
        case 'g':
        case 'G':
        {
          strcat(finalstr, n_Global);
          fcnt += strlen(n_Global) - 1;
          break;
        }
        case 's':
        case 'S':
        {
          strcat(finalstr, Me.name);
          fcnt += strlen(Me.name) - 1;
          break;
        }
        case 'b':
        case 'B':
        {
          strcat(finalstr, "\002");
          break;
        }
        case 'v':
        case 'V':
        {
          strcat(finalstr, hVersion);
          fcnt += strlen(hVersion) - 1;
          break;
        }
        case 'a':
        case 'A':
        {
          strcat(finalstr, Me.admin);
          fcnt += strlen(Me.admin) - 1;
          break;
        }
        case '+':
        {
          char flag;
          char *cptr, *finstr;
          struct Userlist *tempuser = NULL;

          flag = tempstr[tcnt + 1];
          if (!nick)
            tempuser = DccGetUser(IsDccSock(sockfd));
          else if (lptr)
            tempuser = GetUser(1, lptr->nick, lptr->username, lptr->hostname);
          else
            tempuser = GetUser(1, nick, (char *) NULL, (char *) NULL);

          if ((CheckAccess(tempuser, flag)))
          {
            if (!IsRegistered(lptr, sockfd))
              return ((char *) -1);

            tcnt += 2;
            cptr = &tempstr[tcnt];
            finstr = Substitute(nick, cptr, sockfd);
            if (!finstr)
            {
              /* its a line like "%+a" with no text, so send a \r\n */
              /* MyFree(finstr); */
              finstr = MyStrdup("\r\n");
            }
            return (finstr);
          }
          else
            return ((char *) -1); /* user doesn't have privs to read line */
          break;
        }

        default:
        {
          strcat(finalstr, "%");
          finalstr[++fcnt] = tempstr[tcnt];
          break;
        }
      } /* switch (key) */
    }
    else
      finalstr[fcnt] = tempstr[tcnt];

    fcnt++;
    tcnt++;
  }

  if (finalstr[0])
  {
    finalstr[fcnt++] = '\n';
    finalstr[fcnt] = '\0';
    return (finalstr);
  }
  else
  {
    MyFree(finalstr);
    return ((char *) NULL);
  }
} /* Substitute() */

/*
GetService
  Returns a pointer to luser structure containing n_*Nick
if 'name' matches a Service Bot
*/

struct Luser *
GetService(char *name)

{
  struct aService *sptr;

  if (!name)
    return (NULL);

  for (sptr = ServiceBots; sptr->name; ++sptr)
  {
    if (!strcasecmp(name, *(sptr->name)))
      return (*(sptr->lptr));
  }

  return (NULL);
} /* GetService() */

/*
FindService
  Similar to GetService(), except accept a struct Luser * pointer
and attempt to find a service match. Returns a pointer
to appropriate Luser structure if 'lptr' is a service nick
*/

struct Luser *
FindService(struct Luser *lptr)

{
  struct aService *sptr;

  if (!lptr)
    return (NULL);

  for (sptr = ServiceBots; sptr->name; ++sptr)
  {
    if (lptr == *(sptr->lptr))
      return (*(sptr->lptr));
  }

  return (NULL);
} /* FindService() */

/*
pwmatch()
  Check if 'password' matches 'chkpass' - return 1 if so, 0 if not
Assumes you're trying to match passwords from *Serv databases
*/

int
pwmatch(char *password, char *chkpass)

{
#ifdef CRYPT_PASSWORDS
  char *encr;
#endif

  if (!password || !chkpass)
    return 0;

#ifdef CRYPT_PASSWORDS

  encr = (char *) crypt(chkpass, password);
  assert(encr != 0);

  if (!strcmp(encr, password))
    return 1;
  else
    return 0;

#else

  if (!strcmp(password, chkpass))
    return 1;
  else
    return 0;

#endif
} /* pwmatch() */

/*
operpwmatch()
  Check if 'password' matches 'chkpass' - return 1 if so, 0 if not
Assumes you're trying to match passwords from O: lines in hybserv.conf
*/

int
operpwmatch(char *password, char *chkpass)

{
#ifdef CRYPT_OPER_PASSWORDS
  char *encr;
#endif

  if (!password || !chkpass)
    return 0;

#ifdef CRYPT_OPER_PASSWORDS

  encr = (char *) crypt(chkpass, password);
  assert(encr != 0);

  if (!strcmp(encr, password))
    return 1;
  else
    return 0;

#else

  if (!strcmp(password, chkpass))
    return 1;
  else
    return 0;

#endif
} /* operpwmatch() */

/*
IsInNickArray()
 Determine if 'nickname' is in the array 'nicks'
Return 1 if so, 0 if not
*/

int
IsInNickArray(int nickcnt, char **nicks, char *nickname)

{
  int ii;
  char *ntmp;

  for (ii = 0; ii < nickcnt; ii++)
  {
    ntmp = GetNick(nicks[ii]);
    if (!ntmp)
      continue;

    if (!strcasecmp(ntmp, nickname))
      return (1);
  }

  return (0);
} /* IsInNickArray() */

/*
IsNum()
  Determine if 'str' is a number - return 0 if not, the number if it is
*/

int
IsNum(char *str)

{
  int result = 0;
  char tmp[MAXLINE];
  char *tmp2;

  if (!str)
    return 0;

  strcpy(tmp, str);
  tmp2 = tmp;
  while (*tmp2)
  {
    if (IsDigit(*tmp2))
    {
      result *= 10;
      result += ((*tmp2) & 0xF);
      tmp2++;
    }
    else
      return 0;
  }

  /* if 'str' actually contains the number 0, return 1 */
  if (result == 0)
    result = 1;

  return (result);
} /* IsNum() */

/*
GetCommand()
 Attempt to find the command "name" in the list "cmdlist".
Return a pointer to the index containing "name" if found,
otherwise NULL.  If the command is found, but there is
more than 1 match (ambiguous), return (struct Command *) -1.
*/

struct Command *
GetCommand(struct Command *cmdlist, char *name)

{
  struct Command *cmdptr, *tmp;
  int matches, /* number of matches we've had so far */
      clength;

  if (!cmdlist || !name)
    return (NULL);

  tmp = NULL;
  matches = 0;
  clength = strlen(name);
  for (cmdptr = cmdlist; cmdptr->cmd; cmdptr++)
  {
    if (!strncasecmp(name, cmdptr->cmd, clength))
    {
      if (clength == strlen(cmdptr->cmd))
      {
        /*
         * name and cmdptr->cmd are the same length, so it
         * must be an exact match, don't search any further
         */
        matches = 0;
        break;
      }
      tmp = cmdptr;
      matches++;
    }
  }

  /*
   * If matches > 1, name is an ambiguous command, so the
   * user needs to be more specific
   */
  if ((matches == 1) && (tmp))
    cmdptr = tmp;

  if (cmdptr->cmd)
    return (cmdptr);

  if (matches == 0)
    return (NULL); /* no matches found */
  else
    return ((struct Command *) -1); /* multiple matches found */
} /* GetCommand() */

char* stripctrlsymbols( char * source )
{
	char *p = NULL;

        if( source == NULL )
                return NULL;

	for (p = source; *p; p++) {	
		if (*p == 0x04 && (*(p + 1) == 0x08 ))
		{
			*(p+1) = '*';
			source = p+1;
		        continue;	
		}
		if( *p < 0x21 ){
			*(p) = '*';
                        source = p;
                }
	}
	return source;
}

char* stripformatsymbols( char * source )
{
        char *p ;

        if( source == NULL )
                return NULL;

        for (p = source; *p; p++) 
            if (*p == '%' ) 
                  *p = '*';
        return source;
}
