/*
 * HybServ TS Services, Copyright (C) 1998-1999 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 *
 * $Id: chanserv.c,v 1.0.0 1998/09/04 12:12:32 wnder Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "alloc.h"
#include "channel.h"
#include "chanserv.h"
#include "client.h"
#include "conf.h"
#include "config.h"
#include "dcc.h"
#include "err.h"
#include "hash.h"
#include "helpserv.h"
#include "hybdefs.h"
#include "log.h"
#include "match.h"
#include "memoserv.h"
#include "misc.h"
#include "mystring.h"
#include "nickserv.h"
#include "operserv.h"
#include "server.h"
#include "settings.h"
#include "sock.h"
#include "Strn.h"
#include "timestr.h"

#if defined(NICKSERVICES) && defined(CHANNELSERVICES)

extern char *crypt();

/* hash table of registered channels */
struct ChanInfo *chanlist[CHANLIST_MAX];

static void cs_joinchan(struct ChanInfo *);
static void cs_CheckChan(struct ChanInfo *, struct Channel *);
static int ChangeChanPermPass(struct ChanInfo *, char *);
static int SetChanDefAccess(struct ChanInfo *);
static void AddChan(struct ChanInfo *);
static struct ChanInfo *MakeChan();
static void AddFounder(struct Luser *, struct ChanInfo *);
static int AddAccess(struct ChanInfo *, struct Luser *, char *,
                     struct NickInfo *, int);
static int AddAccessDB(struct ChanInfo *, struct NickInfo *, char *,
                     struct NickInfo *, int);
static int DelAccess(struct ChanInfo *, struct Luser *, char *,
                     struct NickInfo *);
static int AddAkick(struct ChanInfo *, struct Luser *, char *, char *);
static int DelAkick(struct ChanInfo *, char *);
static int IsFounder(struct Luser *, struct ChanInfo *);
static int GetAccess(struct ChanInfo *, struct Luser *);
static struct ChanAccess *OnAccessList(struct ChanInfo *, char *,
                                       struct NickInfo *);
static struct AutoKick *OnAkickList(struct ChanInfo *, char *);

/* default access levels for new channels */
int DefaultAccess[] = {
  -1,          /* CA_AUTODEOP */
  5,           /* CA_AUTOVOICE */
  5,           /* CA_CMDVOICE */
  9000,        /* CA_ACCESS */
  5,           /* CA_CMDINVITE */
  10,          /* CA_AUTOOP */
  10,          /* CA_CMDOP */
  10,          /* CA_CMDUNBAN */
  15,          /* CA_AKICK */
  9000,        /* CA_CMDCLEAR */
  10000,       /* CA_SET */
  9500,        /* CA_SUPEROP */
  10000        /* CA_FOUNDER */
};

static void c_help(struct Luser *, struct NickInfo *, int, char **);
static void c_register(struct Luser *, struct NickInfo *, int, char **);
static void c_drop(struct Luser *, struct NickInfo *, int, char **);
static void c_identify(struct Luser *, struct NickInfo *, int, char **);
static void c_newpass(struct Luser *, struct NickInfo *, int, char **);

static void c_access(struct Luser *, struct NickInfo *, int, char **);
static void c_access_add(struct Luser *, struct NickInfo *, int, char **);
static void c_access_del(struct Luser *, struct NickInfo *, int, char **);
static void c_access_list(struct Luser *, struct NickInfo *, int, char **);

static void c_akick(struct Luser *, struct NickInfo *, int, char **);
static void c_akick_add(struct Luser *, struct NickInfo *, int, char **);
static void c_akick_del(struct Luser *, struct NickInfo *, int, char **);
static void c_akick_list(struct Luser *, struct NickInfo *, int, char **);

static void c_list(struct Luser *, struct NickInfo *, int, char **);
static void c_level(struct Luser *, struct NickInfo *, int, char **);

static void c_set(struct Luser *, struct NickInfo *, int, char **);
static void c_set_topiclock(struct Luser *, struct NickInfo *, int, char **);
static void c_set_private(struct Luser *, struct NickInfo *, int, char **);
static void c_set_secure(struct Luser *, struct NickInfo *, int, char **);
static void c_set_secureops(struct Luser *, struct NickInfo *, int, char **);
static void c_set_restricted(struct Luser *, struct NickInfo *, int, char **);
static void c_set_forget(struct Luser *, struct NickInfo *, int, char **);
static void c_set_guard(struct Luser *, struct NickInfo *, int, char **);
static void c_set_seenserv(struct Luser *, struct NickInfo *, int, char **);
static void c_set_password(struct Luser *, struct NickInfo *, int, char **);
static void c_set_founder(struct Luser *, struct NickInfo *, int, char **);
static void c_set_successor(struct Luser *, struct NickInfo *, int, char **);
static void c_set_mlock(struct Luser *, struct NickInfo *, int, char **);
static void c_set_topic(struct Luser *, struct NickInfo *, int, char **);
static void c_set_entrymsg(struct Luser *, struct NickInfo *, int, char **);
static void c_set_email(struct Luser *, struct NickInfo *, int, char **);
static void c_set_url(struct Luser *, struct NickInfo *, int, char **);

static void c_invite(struct Luser *, struct NickInfo *, int, char **);
static void c_op(struct Luser *, struct NickInfo *, int, char **);
static void c_voice(struct Luser *, struct NickInfo *, int, char **);
static void c_unban(struct Luser *, struct NickInfo *, int, char **);
static void c_info(struct Luser *, struct NickInfo *, int, char **);

static void c_clear(struct Luser *, struct NickInfo *, int, char **);
static void c_clear_ops(struct Luser *, struct NickInfo *, int, char **);
static void c_clear_voices(struct Luser *, struct NickInfo *, int, char **);
static void c_clear_modes(struct Luser *, struct NickInfo *, int, char **);
#ifdef UNIBG_SERVICES
static void c_clear_gecos_bans(struct Luser *, struct NickInfo *, int, char **);
#endif
static void c_clear_bans(struct Luser *, struct NickInfo *, int, char **);
static void c_clear_users(struct Luser *, struct NickInfo *, int, char **);
static void c_clear_all(struct Luser *, struct NickInfo *, int, char **);

#ifdef EMPOWERADMINS
static void c_forbid(struct Luser *, struct NickInfo *, int, char **);
static void c_setpass(struct Luser *, struct NickInfo *, int, char **);
static void c_status(struct Luser *, struct NickInfo *, int, char **);
static void c_forget(struct Luser *, struct NickInfo *, int, char **);
static void c_noexpire(struct Luser *, struct NickInfo *, int, char **);
#endif /* EMPOWERADMINS */

/* main ChanServ commands */
static struct Command chancmds[] = {
  { "HELP", c_help, LVL_NONE },
  { "REGISTER", c_register, LVL_IDENT },
  { "DROP", c_drop, LVL_IDENT },
  { "IDENTIFY", c_identify, LVL_IDENT },
  { "NEWPASS", c_newpass, LVL_NONE },
  { "ACCESS", c_access, LVL_IDENT },
  { "AKICK", c_akick, LVL_IDENT },
  { "LIST", c_list, LVL_NONE },
  { "LEVEL", c_level, LVL_IDENT },
  { "SET", c_set, LVL_IDENT },
  { "INVITE", c_invite, LVL_IDENT },
  { "OP", c_op, LVL_IDENT },
  { "VOICE", c_voice, LVL_IDENT },
  { "UNBAN", c_unban, LVL_IDENT },
  { "INFO", c_info, LVL_NONE },
  { "CLEAR", c_clear, LVL_IDENT },

#ifdef EMPOWERADMINS
  { "FORBID", c_forbid, LVL_ADMIN },
  { "SETPASS", c_setpass, LVL_ADMIN },
  { "STATUS", c_status, LVL_ADMIN },
  { "FORGET", c_forget, LVL_ADMIN },
  { "NOEXPIRE", c_noexpire, LVL_ADMIN },
#endif

  { 0, 0, 0 }
};

/* sub-commands for ChanServ ACCESS */
static struct Command accesscmds[] = {
  { "ADD", c_access_add, LVL_NONE },
  { "DEL", c_access_del, LVL_NONE },
  { "LIST", c_access_list, LVL_NONE },

  { 0, 0, 0 }
};

/* sub-commands for ChanServ AKICK */
static struct Command akickcmds[] = {
  { "ADD", c_akick_add, LVL_NONE },
  { "DEL", c_akick_del, LVL_NONE },
  { "LIST", c_akick_list, LVL_NONE },

  { 0, 0, 0 }
};

/* sub-commands for ChanServ SET */
static struct Command setcmds[] = {
  { "TOPICLOCK", c_set_topiclock, LVL_NONE },
  { "TLOCK", c_set_topiclock, LVL_NONE },
  { "PRIVATE", c_set_private, LVL_NONE },
  { "SECURE", c_set_secure, LVL_NONE },
  { "SECUREOPS", c_set_secureops, LVL_NONE },
  { "RESTRICTED", c_set_restricted, LVL_NONE },
  { "FORGET", c_set_forget, LVL_ADMIN },
  { "GUARD", c_set_guard, LVL_NONE },
  { "SEENSERV", c_set_seenserv, LVL_NONE },
  { "PASSWORD", c_set_password, LVL_NONE },
  { "NEWPASS", c_set_password, LVL_NONE },
  { "FOUNDER", c_set_founder, LVL_NONE },
  { "SUCCESSOR", c_set_successor, LVL_NONE },
  { "MLOCK", c_set_mlock, LVL_NONE },
  { "MODELOCK", c_set_mlock, LVL_NONE },
  { "TOPIC", c_set_topic, LVL_NONE },
  { "ENTRYMSG", c_set_entrymsg, LVL_NONE },
  { "ONJOIN", c_set_entrymsg, LVL_NONE },
  { "EMAIL", c_set_email, LVL_NONE },
  { "MAIL", c_set_email, LVL_NONE },
  { "URL", c_set_url, LVL_NONE },
  { "WEBSITE", c_set_url, LVL_NONE },

  { 0, 0, 0 }
};

/* sub-commands for ChanServ CLEAR */
static struct Command clearcmds[] = {
  { "OPS", c_clear_ops, LVL_NONE },
  { "VOICES", c_clear_voices, LVL_NONE },
  { "MODES", c_clear_modes, LVL_NONE },
  { "BANS", c_clear_bans, LVL_NONE },
#ifdef UNIBG_SERVICES
  { "GECOSBANS", c_clear_gecos_bans, LVL_NONE },
#endif
  { "USERS", c_clear_users, LVL_NONE },
  { "ALL", c_clear_all, LVL_NONE },

  { 0, 0, 0 }
};

typedef struct
{
  int level;
  char *cmd;
  char *desc;
} AccessInfo;

static AccessInfo accessinfo[] = {
  { CA_AUTODEOP, "AUTODEOP", "Automatic deop/devoice" },
  { CA_AUTOVOICE, "AUTOVOICE", "Automatic voice" },
  { CA_CMDVOICE, "CMDVOICE", "Use of command VOICE" },
  { CA_ACCESS, "ACCESS", "Allow ACCESS modification" },
  { CA_CMDINVITE, "CMDINVITE", "Use of command INVITE" },
  { CA_AUTOOP, "AUTOOP", "Automatic op" },
  { CA_CMDOP, "CMDOP", "Use of comand OP" },
  { CA_CMDUNBAN, "CMDUNBAN", "Use of command UNBAN" },
  { CA_AKICK, "AUTOKICK", "Allow AKICK modification" },
  { CA_CMDCLEAR, "CMDCLEAR", "Use of command CLEAR" },
  { CA_SET, "SET", "Modify channel SETs" },
  { CA_SUPEROP, "SUPEROP", "All of the above" },
  { CA_FOUNDER, "FOUNDER", "Full access to the channel" },

  { 0, 0, 0 }
};

/*
cs_process()
  Process command coming from 'nick' directed towards n_ChanServ
*/

void
cs_process(char *nick, char *command)

{
  int acnt;
  char **arv;
  struct Command *cptr;
  struct Luser *lptr;
  struct NickInfo *nptr;
  struct ChanInfo *chptr;
  int isadmin;
  int showchaninfo;

  if (!command || !(lptr = FindClient(nick)))
    return;

  if (Network->flags & NET_OFF)
  {
    notice(n_ChanServ, lptr->nick,
      "Services are currently \002disabled\002");
    return;
  }

  acnt = SplitBuf(command, &arv);
  if (acnt == 0)
  {
    MyFree(arv);
    return;
  }

  cptr = GetCommand(chancmds, arv[0]);

  if (!cptr || (cptr == (struct Command *) -1))
  {
    notice(n_ChanServ, lptr->nick,
      "%s command [%s]",
      (cptr == (struct Command *) -1) ? "Ambiguous" : "Unknown",
      arv[0]);
    MyFree(arv);
    return;
  }

  if ((!(nptr = FindNick(lptr->nick))) &&
      (cptr->level != LVL_NONE))
  {
    /* the command requires a registered nickname */

    notice(n_ChanServ, lptr->nick,
      "Your nickname is not registered");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_NickServ,
      "REGISTER");
    MyFree(arv);
    return;
  }

  /*
   * Check if the command is for admins only - if so,
   * check if they match an admin O: line.  If they do,
   * check if they are EITHER oper'd, or registered with OperServ,
   * if either of these is true, allow the command
   */

  isadmin = IsValidAdmin(lptr);

  if ((cptr->level == LVL_ADMIN) && !(IsValidAdmin(lptr)))
  {
    notice(n_ChanServ, lptr->nick, "Unknown command [%s]",
      arv[0]);
    MyFree(arv);
    return;
  }

  if (nptr)
  {
    if (nptr->flags & NS_FORBID)
    {
      notice(n_ChanServ, lptr->nick,
        "Cannot execute commands for forbidden nicknames");
      MyFree(arv);
      return;
    }

    if (cptr->level != LVL_NONE)
    {
      if (!(nptr->flags & NS_IDENTIFIED))
      {
        notice(n_ChanServ, lptr->nick,
          "Password identification is required for [\002%s\002]",
          cptr->cmd);
        notice(n_ChanServ, lptr->nick, 
          "Type \002/msg %s IDENTIFY <password>\002 and retry",
          n_NickServ);
        MyFree(arv);
        return;
      }
    }
  } /* if (nptr) */

  if ((chptr = FindChan(acnt >= 2 ? arv[1] : (char *) NULL)))
  {
    showchaninfo = 1;

    if ((chptr->flags & (CS_FORBID | CS_FORGET))
            && (strncasecmp(arv[0], "DROP", strlen(arv[0])) != 0))
	showchaninfo = 0;

    if ((isadmin || IsOperator(lptr))
	    && (chptr->flags & (CS_FORBID | CS_FORGET))
	    && (strncasecmp(arv[0], "INFO", strlen(arv[0])) == 0))
	showchaninfo = 1;

    if (!(showchaninfo))
    {
        if (chptr->flags & CS_FORBID)
        {
            notice(n_ChanServ, lptr->nick,
              "Cannot execute commands for forbidden channels");
        }
        else
        {
            notice(n_ChanServ, lptr->nick,
              "Cannot execute commands for forgotten channels");
        }

      MyFree(arv);
      return;
    }
  }

  /* call cptr->func to execute command */
  (*cptr->func)(lptr, nptr, acnt, arv);

  MyFree(arv);

  return;
} /* cs_process() */

/*
cs_loaddata()
  Load ChanServ database - return 1 if successful, -1 if not, and -2
if the errors are unrecoverable
*/

int
cs_loaddata()

{
  FILE *fp;
  char line[MAXLINE], **av;
  char *keyword;
  int ac, ret = 1, cnt;
  struct ChanInfo *cptr = NULL;

  if (!(fp = fopen(ChanServDB, "r")))
  {
    /* ChanServ data file doesn't exist */
    return -3;
  }

  cnt = 0;
  /* load data into list */
  while (fgets(line, MAXLINE - 1, fp))
  {
    ++cnt;

    ac = SplitBuf(line, &av);
    if (!ac)
    {
      /* probably a blank line */
      MyFree(av);
      continue;
    }

    if (av[0][0] == ';')
    {
      MyFree(av);
      continue;
    }

    if (!strncasecmp("->", av[0], 2))
    {
      /* 
       * check if there are enough args
       */
      if (ac < 2)
      {
        fatal(1, "%s:%d Invalid database format (FATAL)",
          ChanServDB,
          cnt);
        ret = -2;
        MyFree(av);
        continue;
      }

      /* check if there is no channel associated with data */
      if (!cptr)
      {
        fatal(1, "%s:%d No channel associated with data",
          ChanServDB,
          cnt);
        if (ret > 0)
          ret = -2;
        MyFree(av);
        continue;
      }

      keyword = av[0] + 2;
      if (!strncasecmp("PASS", keyword, 4))
      {
        if (!cptr->password)
          cptr->password = MyStrdup(av[1]);
        else
        {
          fatal(1,
            "%s:%d ChanServ entry for [%s] has multiple PASS lines (using first)",
            ChanServDB,
            cnt,
            cptr->name);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (!strncasecmp("PERMPASS", keyword, 8))
      {
        if (!cptr->permpass)
          cptr->permpass = MyStrdup(av[1]);
        else
        {
          fatal(1,
            "%s:%d ChanServ entry for [%s] has multiple PERMPASS lines (using first)",
            ChanServDB,
            cnt,
            cptr->name);
          if (ret > 0)
            ret = -1;
        }
      } 
      else if (!strncasecmp("FNDR", keyword, 4))
      {
        if (!cptr->founder)
        {
          struct NickInfo *ni = FindNick(av[1]);
          struct NickInfo *master;
          if (!ni)
             ni = FindNick("unibgchans");
// this is still lame! we must be sure that we have
// registered default nickname!
// decho, 14.04.2000
          if (!ni)
          {
            fatal(1, "%s:%d Founder nickname [%s] for channel [%s] is not registered (FATAL)",
              ChanServDB,
              cnt,
              av[1],
              cptr->name);
            ret = -2;
          }

          else if ((master = GetMaster(ni)))
          {
            cptr->founder = MyStrdup(master->nick);
            AddFounderChannelToNick(&master, cptr);
          }

          else
          {
            fatal(1, "%s:%d Unable to find master nickname for founder [%s] on channel [%s] (FATAL)",
              ChanServDB,
              cnt,
              ni->nick,
              cptr->name);
            ret = -2;
          }
        }
        else
        {
          fatal(1, "%s:%d ChanServ entry for [%s] has multiple FNDR lines (using first)",
            ChanServDB,
            cnt,
            cptr->name);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (!strncasecmp(keyword, "SUCCESSOR", 9))
      {
        if (!cptr->successor)
        {
          struct NickInfo  *ni = FindNick(av[1]);

          if (!ni)
          {
            fatal(1, "%s:%d Successor nickname [%s] for channel [%s] is not registered (ignoring)",
              ChanServDB,
              cnt,
              av[1],
              cptr->name);
            if (ret > 0)
              ret = -1;
          }
          else
            cptr->successor = MyStrdup(ni->nick);
        }
        else
        {
          fatal(1, "%s:%d ChanServ entry for [%s] has multiple SUCCESSOR lines (using first)",
            ChanServDB,
            cnt,
            cptr->name);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (!strncasecmp("ACCESS", keyword, 6))
      {
        if (ac < 3)
        {
          fatal(1, "%s:%d Invalid database format (FATAL)",
            ChanServDB,
            cnt);
          ret = -2;
        }
        else
        {
          struct NickInfo *nptr;
	  struct NickInfo *added;
	  
	  if (ac < 4) {
	    added = NULL;
	  } else {
	    added = FindNick(av[3]);
	  }
	  
          if ((nptr = FindNick(av[1])))
            AddAccessDB(cptr,  added ? added : NULL, NULL, nptr, atoi(av[2]));
          else
            AddAccessDB(cptr, added ? added : NULL, av[1], NULL, atoi(av[2]));
        }
      }
      else if (!strncasecmp("AKICK", keyword, 5))
      {
        if (ac < 3)
        {
          fatal(1, "%s:%d Invalid database format (FATAL)",
            ChanServDB,
            cnt);
          ret = -2;
        }
        else
          AddAkick(cptr, (struct Luser *) NULL, av[1], av[2] + 1);
      }
      else if (!strncasecmp("ALVL", keyword, 4))
      {
        /* channel access levels */
        if (ac != (CA_SIZE + 1))
        {
          fatal(1, "%s:%d Invalid database format (FATAL)",
            ChanServDB,
            cnt);
          ret = -2;
        }
        else if (!cptr->access_lvl)
        {
          int ii;

          cptr->access_lvl = (int *) MyMalloc(sizeof(int) * CA_SIZE);
          for (ii = 0; ii < CA_SIZE; ii++)
            cptr->access_lvl[ii] = atoi(av[ii + 1]);
        }
        else
        {
          fatal(1, "%s:%d ChanServ entry for [%s] has multiple ALVL lines (using first)",
            ChanServDB,
            cnt,
            cptr->name);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (!strncasecmp("TOPIC", keyword, 5))
      {
        if (!cptr->topic)
          cptr->topic = MyStrdup(av[1] + 1);
        else
        {
          fatal(1, "%s:%d ChanServ entry for [%s] has multiple TOPIC lines (using first)",
            ChanServDB,
            cnt,
            cptr->name);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (!strncasecmp("LIMIT", keyword, 5))
      {
        if (!cptr->limit)
          cptr->limit = atoi(av[1]);
        else
        {
          fatal(1, "%s:%d ChanServ entry for [%s] has multiple LIMIT lines (using first)",
            ChanServDB,
            cnt,
            cptr->name);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (!strncasecmp("KEY", keyword, 3))
      {
        if (!cptr->key)
          cptr->key = MyStrdup(av[1]);
        else
        {
          fatal(1, "%s:%d ChanServ entry for [%s] has multiple KEY lines (using first)",
            ChanServDB,
            cnt,
            cptr->name);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (!strncasecmp("MON", keyword, 3))
      {
        if (!cptr->modes_on)
          cptr->modes_on = atoi(av[1]);
        else
        {
          fatal(1, "%s:%d ChanServ entry for [%s] has multiple MON lines (using first)",
            ChanServDB,
            cnt,
            cptr->name);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (!strncasecmp("MOFF", keyword, 5))
      {
        if (!cptr->modes_off)
          cptr->modes_off = atoi(av[1]);
        else
        {
          fatal(1, "%s:%d ChanServ entry for [%s] has multiple MOFF lines (using first)",
            ChanServDB,
            cnt,
            cptr->name);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (!strncasecmp("ENTRYMSG", keyword, 8))
      {
        if (!cptr->entrymsg)
          cptr->entrymsg = MyStrdup(av[1] + 1);
        else
        {
          fatal(1, "%s:%d ChanServ entry for [%s] has multiple ENTRYMSG lines (using first)",
            ChanServDB,
            cnt,
            cptr->name);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (!strncasecmp("EMAIL", keyword, 8))
      {
        if (!cptr->email)
          cptr->email = MyStrdup(av[1]);
        else
        {
          fatal(1, "%s:%d ChanServ entry for [%s] has multiple EMAIL lines (using first)",
            ChanServDB,
            cnt,
            cptr->name);
          if (ret > 0)
            ret = -1;
        }
      }
      else if (!strncasecmp("URL", keyword, 8))
      {
        if (!cptr->url)
          cptr->url = MyStrdup(av[1]);
        else
        {
          fatal(1, "%s:%d ChanServ entry for [%s] has multiple URL lines (using first)",
            ChanServDB,
            cnt,
            cptr->name);
          if (ret > 0)
            ret = -1;
        }
      }
    }
    else
    {
      if (cptr)
      {
        if (!cptr->access_lvl && !(cptr->flags & (CS_FORBID | CS_FORGET)))
        {
	  
	  SetChanDefAccess(cptr);

          fatal(1,
            "%s:%d No access level list for registered channel [%s] (using default)",
            ChanServDB,
            cnt,
            cptr->name);
          if (ret > 0)
            ret = -1;
        }

        if (!cptr->password && !(cptr->flags & (CS_FORBID | CS_FORGET)))
        {
          /* the previous channel didn't have a PASS line */
          fatal(1, "%s:%d No founder password entry for registered channel [%s] (FATAL)",
            ChanServDB,
            cnt,
            cptr->name);
          ret = -2;
        }

        if (!cptr->founder && !(cptr->flags & (CS_FORBID | CS_FORGET)))
        {
          /* the previous channel didn't have a FNDR line */
          fatal(1, "%s:%d No founder nickname entry for registered channel [%s] (FATAL)",
            ChanServDB,
            cnt,
            cptr->name);
          ret = -2;
        }

        /*
         * we've come to a new channel entry, so add the last chan
         * to the list before proceeding
         */
        AddChan(cptr);
      } /* if (cptr) */

      /* 
       * make sure there are enough args on the line
       * <channel> <flags> <time created> <last used>
       */
      if (ac < 4)
      {
        fatal(1, "%s:%d Invalid database format (FATAL)",
          ChanServDB,
          cnt);
        ret = -2;
        cptr = NULL;
        MyFree(av);
        continue;
      }

      if (av[0][0] != '#')
      {
        fatal(1, "%s:%d Invalid channel name [%s] (skipping)",
          ChanServDB,
          cnt,
          av[0]);
        if (ret > 0)
          ret = -1;
        cptr = NULL;
        MyFree(av);
        continue;
      }

      cptr = MakeChan();
      cptr->name = MyStrdup(av[0]);
      cptr->flags = atoi(av[1]);
      cptr->created = atoi(av[2]);
      cptr->lastused = atoi(av[3]);
    }

    MyFree(av);
  } /* while */

  if (cptr)
  {
    if (!cptr->access_lvl && !(cptr->flags & (CS_FORBID | CS_FORGET)))
    {

      SetChanDefAccess(cptr);

      fatal(1,
        "%s:%d No access level list for registered channel [%s] (using default)",
        ChanServDB,
        cnt,
        cptr->name);
      if (ret > 0)
        ret = -1;
    }

    if ((cptr->password && cptr->founder) ||
        (cptr->flags & (CS_FORBID | CS_FORGET)))
    {
      /* 
       * cptr has a founder and password, so it can be added to
       * the table
       */
      AddChan(cptr);
    }
    else
    {
      if (!cptr->password && !(cptr->flags & (CS_FORBID | CS_FORGET)))
      {
        fatal(1, "%s:%d No founder password entry for registered channel [%s] (FATAL)",
          ChanServDB,
          cnt,
          cptr->name);
        ret = -2;
      }

      if (!cptr->founder && !(cptr->flags & (CS_FORBID | CS_FORGET)))
      {
        fatal(1, "%s:%d No founder nickname entry for registered channel [%s] (FATAL)",
          ChanServDB,
          cnt,
          cptr->name);
        ret = -2;
      }
    }
  } /* if (cptr) */

  fclose(fp);

  return (ret);
} /* cs_loaddata */

/*
cs_join()
  Have ChanServ join a channel
*/

void
cs_join(struct ChanInfo *chanptr)

{
  int goodjoin;
  char sendstr[MAXLINE], **av;
  struct Channel *chptr;
  time_t chants;

  if (!chanptr)
    return;

  if (cs_ShouldBeOnChan(chanptr))
    goodjoin = 1;
  else
    goodjoin = 0;

  chptr = FindChannel(chanptr->name);

  if (goodjoin)
  {
    /*
     * chptr might be null, so make sure we check before
     * using it
     */
    if (chptr)
      chants = chptr->since;
    else
      chants = time(NULL);

    sprintf(sendstr,
      ":%s SJOIN %ld %s + :@%s\n",
      Me.name,
      (long) chants,
      chanptr->name,
      n_ChanServ);
    toserv(sendstr);

    SplitBuf(sendstr, &av);

    /*
     * The previous call to FindChannel() may have
     * been null, so reassign chptr to the channel
     * structure if the channel was just created.
     * This way, we will always pass a valid pointer
     * to cs_CheckChan()
     */
    chptr = AddChannel(av, 0, (char **) NULL);

    MyFree(av);
  }

  /* Check that the right ops are opped etc */
  cs_CheckChan(chanptr, chptr);
} /* cs_join() */

/*
cs_joinchan()
 Similar to cs_join, except don't call cs_CheckChan() on the
channel.
*/

static void
cs_joinchan(struct ChanInfo *chanptr)

{
  char sendstr[MAXLINE], **av;
  struct Channel *chptr;

  chptr = FindChannel(chanptr->name);

  sprintf(sendstr,
    ":%s SJOIN %ld %s + :@%s\n",
    Me.name,
    chptr ? (long) chptr->since : (long) time(NULL),
    chanptr->name,
    n_ChanServ);
  toserv(sendstr);

  SplitBuf(sendstr, &av);
  AddChannel(av, 0, (char **) NULL);
  MyFree(av);
} /* cs_joinchan() */

/*
cs_join_ts_minus_1()
  Have ChanServ join a channel with TS - 1
*/

void
cs_join_ts_minus_1(struct ChanInfo *chanptr)

{
  char sendstr[MAXLINE], **av;
  struct Channel *cptr = FindChannel(chanptr->name);
  int ac;

  sprintf(sendstr,
    ":%s SJOIN %ld %s + :@%s\n",
    Me.name,
    cptr ? (long) (cptr->since - 1) : (long) time(NULL),
    cptr->name,
    n_ChanServ);
  toserv(sendstr);

  ac = SplitBuf(sendstr, &av);
  s_sjoin(ac, av);

  MyFree(av);

  /* Check that the right ops are opped etc */
  cs_CheckChan(chanptr, cptr);
} /* cs_join_ts_minus_1() */

/*
cs_part()
  Have ChanServ part 'chptr'
*/

void
cs_part(struct Channel *chptr)

{
  if (!chptr)
    return;

  toserv(":%s PART %s\n",
    n_ChanServ,
    chptr->name);
  RemoveFromChannel(chptr, Me.csptr);
} /* cs_part() */

/*
cs_CheckChan()
  Called right after ChanServ joins a channel, to check that the right
modes get set, the right people are opped, etc
*/

static void
cs_CheckChan(struct ChanInfo *cptr, struct Channel *chptr)

{
  struct ChannelUser *tempu;
  char *dopnicks; /* nicks to deop */
  char modes[MAXLINE]; /* mlock modes to set */

  if (!cptr || !chptr)
    return;

  if (cptr->flags & CS_FORBID)
  {
    char  *knicks; /* nicks to kick */

    knicks = (char *)MyMalloc(sizeof(char));
    knicks[0] = '\0';
    for (tempu = chptr->firstuser; tempu; tempu = tempu->next)
    {
      if (FindService(tempu->lptr))
        continue;
      knicks = (char *) MyRealloc(knicks, strlen(knicks) + strlen(tempu->lptr->nick) + (2 * sizeof(char)));
      strcat(knicks, tempu->lptr->nick);
      strcat(knicks, " ");
    }
    SetModes(n_ChanServ, 0, 'o', chptr, knicks);

    /*
     * Have ChanServ join the channel to enforce the +i
     */
    if (!IsChannelMember(chptr, Me.csptr))
      cs_joinchan(cptr);

    toserv(":%s MODE %s +i\n",
      n_ChanServ,
      cptr->name);
    UpdateChanModes(Me.csptr, n_ChanServ, chptr, "+i");
    KickBan(0, n_ChanServ, chptr, knicks, "Forbidden Channel");
    MyFree(knicks);

    return;
  }

  cptr->lastused = time(NULL);

  if ((cptr->flags & CS_SECUREOPS) || (cptr->flags & CS_RESTRICTED))
  {
    /* SECUREOPS is set - deop all non autoops */
    dopnicks = (char *) MyMalloc(sizeof(char));
    dopnicks[0] = '\0';
    for (tempu = chptr->firstuser; tempu; tempu = tempu->next)
    {
      if (FindService(tempu->lptr))
        continue;
      
      if ((tempu->flags & CH_OPPED) && 
          !HasAccess(cptr, tempu->lptr, CA_AUTOOP))
      {
        dopnicks = (char *) MyRealloc(dopnicks, strlen(dopnicks) + strlen(tempu->lptr->nick) + (2 * sizeof(char)));
        strcat(dopnicks, tempu->lptr->nick);
        strcat(dopnicks, " ");
      }
    }
    SetModes(n_ChanServ, 0, 'o', chptr, dopnicks);
    MyFree(dopnicks);
  }

  if ((cptr->flags & CS_RESTRICTED))
  {
    char *kbnicks; /* nicks to kickban */

    /* channel is restricted - kickban all non-autoops */
    kbnicks = (char *) MyMalloc(sizeof(char));
    kbnicks[0] = '\0';
    for (tempu = chptr->firstuser; tempu; tempu = tempu->next)
    {
      if (FindService(tempu->lptr))
        continue;
      if (!HasAccess(cptr, tempu->lptr, CA_AUTOOP))
      {
        kbnicks = (char *) MyRealloc(kbnicks, strlen(kbnicks) + strlen(tempu->lptr->nick) + (2 * sizeof(char)));
        strcat(kbnicks, tempu->lptr->nick);
        strcat(kbnicks, " ");
      }
    }

    /*
     * Have ChanServ join the channel to enforce the bans
     */
    if (!IsChannelMember(chptr, Me.csptr))
      cs_joinchan(cptr);

    KickBan(1, n_ChanServ, chptr, kbnicks, "Restricted Channel");
    MyFree(kbnicks);
  }

  strcpy(modes, "+");
  if (cptr->modes_on || cptr->key || cptr->limit)
  {
    if ((cptr->modes_on & MODE_S) &&
        !(chptr->modes & MODE_S))
      strcat(modes, "s");
    if ((cptr->modes_on & MODE_P) &&
        !(chptr->modes & MODE_P))
      strcat(modes, "p");
    if ((cptr->modes_on & MODE_N) &&
        !(chptr->modes & MODE_N))
      strcat(modes, "n");
    if ((cptr->modes_on & MODE_T) &&
        !(chptr->modes & MODE_T))
      strcat(modes, "t");
    if ((cptr->modes_on & MODE_M) &&
        !(chptr->modes & MODE_M))
      strcat(modes, "m");
    if ((cptr->modes_on & MODE_C) &&
        !(chptr->modes & MODE_C))
      strcat(modes, "c");
    if ((cptr->modes_on & MODE_I) &&
        !(chptr->modes & MODE_I))
      strcat(modes, "i");
    if (cptr->limit)
      strcat(modes, "l");
    if (cptr->key)
      strcat(modes, "k");

    if (cptr->limit)
    {
      char temp[MAXLINE];

      sprintf(temp, "%s %ld", modes, cptr->limit);
      strcpy(modes, temp);
    }
    if (cptr->key)
    {
      char temp[MAXLINE];

      if (chptr->key)
      {
        sprintf(temp, "-k %s", chptr->key);
        toserv(":%s MODE %s %s\n",
          n_ChanServ,
          chptr->name,
          temp);
        UpdateChanModes(Me.csptr, n_ChanServ, chptr, temp);
      }
      sprintf(temp, "%s %s", modes, cptr->key);
      strcpy(modes, temp);
    }
  }
  if (modes[1])
  {
    toserv(":%s MODE %s %s\n",
      n_ChanServ,
      cptr->name,
      modes);
    UpdateChanModes(Me.csptr, n_ChanServ, chptr, modes);
  }

  strcpy(modes, "-");
  if (cptr->modes_off)
  {
    if ((cptr->modes_off & MODE_S) &&
        (chptr->modes & MODE_S))
      strcat(modes, "s");
    if ((cptr->modes_off & MODE_P) &&
        (chptr->modes & MODE_P))
      strcat(modes, "p");
    if ((cptr->modes_off & MODE_N) &&
        (chptr->modes & MODE_N))
      strcat(modes, "n");
    if ((cptr->modes_off & MODE_T) &&
        (chptr->modes & MODE_T))
      strcat(modes, "t");
    if ((cptr->modes_off & MODE_M) &&
        (chptr->modes & MODE_M))
      strcat(modes, "m");
    if ((cptr->modes_off & MODE_C) &&
        (chptr->modes & MODE_C))
      strcat(modes, "c");
    if ((cptr->modes_off & MODE_I) &&
        (chptr->modes & MODE_I))
      strcat(modes, "i");
    if ((cptr->modes_off & MODE_L) &&
        (chptr->limit))
      strcat(modes, "l");
    if ((cptr->modes_off & MODE_K) &&
        (chptr->key))
    {
      strcat(modes, "k ");
      strcat(modes, chptr->key);
    }
  }
  if (modes[1])
  {
    toserv(":%s MODE %s %s\n",
      n_ChanServ,
      cptr->name,
      modes);
    UpdateChanModes(Me.csptr, n_ChanServ, chptr, modes);
  }

  cs_SetTopic(FindChannel(cptr->name), cptr->topic);
} /* cs_CheckChan() */

/*
cs_SetTopic()
 Set the topic on the given channel to 'topic'
*/

void
cs_SetTopic(struct Channel *chanptr, char *topic)

{
  if (!chanptr || !topic)
    return;

  if (cs_ShouldBeOnChan(FindChan(chanptr->name)))
  {
    /*
     * ChanServ should already be on the channel - just set
     * topic normally
     */
    toserv(":%s TOPIC %s :%s\n",
      n_ChanServ,
      chanptr->name,
      topic);
  }
  else
  {
    /*
     * Hybrid won't accept a TOPIC from a user unless they are
     * on the channel - have ChanServ join and leave. Do it all
     * in one line so ChanServ can't get kicked out etc.
     */
    toserv(":%s SJOIN %ld %s + :@%s\n:%s TOPIC %s :%s\n:%s PART %s\n",
      Me.name,
      chanptr->since,
      chanptr->name,
      n_ChanServ,
      n_ChanServ,
      chanptr->name,
      topic,
      n_ChanServ,
      chanptr->name);
  }
} /* cs_SetTopic() */

/*
cs_CheckModes()
  Called after a mode change, to check if the mode is in conflict
with the mlock modes etc
*/

void
cs_CheckModes(struct Luser *source, struct ChanInfo *cptr,
              int isminus, int mode, struct Luser *lptr)

{
  int slev; /* user level for source */
  char modes[MAXLINE];
  struct Channel *chptr;

  if (!cptr || !source)
    return;

  if ((cptr->flags & (CS_FORBID | CS_FORGET)))
    return;

  /*
   * Allow OperServ and ChanServ to set any mode
   */
  if ((source == Me.csptr) || (source == Me.osptr))
    return;

  /* founders can set any mode */
  if (IsFounder(source, cptr))
    return;

  if (!(chptr = FindChannel(cptr->name)))
    return;

  slev = GetAccess(cptr, source);

  if (mode == MODE_O)
  {
    int    ulev; /* user level for (de)opped nick */

    if (!lptr)
      return;

    ulev = GetAccess(cptr, lptr);
    if (isminus)
    {
      /*
       * don't let someone deop a user with a greater user level,
       * but allow a user to deop him/herself
       */
      if ((lptr != source) &&
          HasAccess(cptr, lptr, CA_AUTOOP) &&
          (ulev >= slev))
      {
        if (HasFlag(lptr->nick, NS_NOCHANOPS))
          return;

        sprintf(modes, "+o %s",
          lptr->nick);
        toserv(":%s MODE %s %s\n",
          n_ChanServ,
          cptr->name,
          modes);
        UpdateChanModes(Me.csptr, n_ChanServ, chptr, modes);
      }
    }
    else /* mode +o */
    {
      if (((cptr->flags & CS_SECUREOPS) &&
          !HasAccess(cptr, lptr, CA_AUTOOP)) ||
          (GetAccess(cptr, lptr) == cptr->access_lvl[CA_AUTODEOP]))
      {
        sprintf(modes, "-o %s",
          lptr->nick);
        toserv(":%s MODE %s %s\n",
          n_ChanServ,
          cptr->name,
          modes);
        UpdateChanModes(Me.csptr, n_ChanServ, chptr, modes);
      }
      else if (HasFlag(lptr->nick, NS_NOCHANOPS))
      {
        sprintf(modes, "-o %s",
          lptr->nick);
        toserv(":%s MODE %s %s\n",
          n_ChanServ,
          cptr->name,
          modes);
        UpdateChanModes(Me.csptr, n_ChanServ, chptr, modes);

        notice(n_ChanServ, lptr->nick,
          "You are not permitted to have channel op privileges");
        SendUmode(OPERUMODE_Y,
          "Flagged user %s!%s@%s was opped on channel [%s] by %s",
          lptr->nick,
          lptr->username,
          lptr->hostname,
          cptr->name,
          source->nick);
      }
    }
    return;
  } /* if (mode == MODE_O) */

  if (mode == MODE_V)
  {
    if (!isminus)
    {
      /* autodeop people aren't allowed voice status either */
      if (GetAccess(cptr, lptr) == cptr->access_lvl[CA_AUTODEOP])
      {
        sprintf(modes, "-v %s",
          lptr->nick);
        toserv(":%s MODE %s %s\n",
          n_ChanServ,
          cptr->name,
          modes);
        UpdateChanModes(Me.csptr, n_ChanServ, chptr, modes);
      }
    }

    return;
  } /* if (mode == MODE_V) */

  /*
   * Check if the mode conflicts with any enforced modes for the
   * channel
   */
  if (cptr->modes_on && isminus)
  {
    modes[0] = '\0';
    if ((mode == MODE_S) &&
        (cptr->modes_on & MODE_S))
      sprintf(modes, "+s-p");
    if ((mode == MODE_P) &&
        (cptr->modes_on & MODE_P))
      sprintf(modes, "+p-s");
    if ((mode == MODE_N) &&
        (cptr->modes_on & MODE_N))
      sprintf(modes, "+n");
    if ((mode == MODE_T) &&
        (cptr->modes_on & MODE_T))
      sprintf(modes, "+t");
    if ((mode == MODE_M) &&
        (cptr->modes_on & MODE_M))
      sprintf(modes, "+m");
    if ((mode == MODE_C) &&
        (cptr->modes_on & MODE_C))
      sprintf(modes, "+c");
    if ((mode == MODE_I) &&
        (cptr->modes_on & MODE_I))
      sprintf(modes, "+i");

    if (modes[0])
    {
      toserv(":%s MODE %s %s\n",
        n_ChanServ,
        cptr->name,
        modes);
      UpdateChanModes(Me.csptr, n_ChanServ, chptr, modes);
    }
  }
  if ((mode == MODE_L) &&
      (cptr->limit))
  {
    sprintf(modes, "+l %ld", cptr->limit);
    toserv(":%s MODE %s %s\n",
      n_ChanServ,
      cptr->name,
      modes);
    UpdateChanModes(Me.csptr, n_ChanServ, chptr, modes);
  }
  if ((mode == MODE_K) &&
      (cptr->key))
  {
    if (!isminus)
    {
      sprintf(modes, "-k %s", chptr->key);
      toserv(":%s MODE %s %s\n",
        n_ChanServ,
        cptr->name,
        modes);
      UpdateChanModes(Me.csptr, n_ChanServ, chptr, modes);
    }
    sprintf(modes, "+k %s", cptr->key);
    toserv(":%s MODE %s %s\n",
      n_ChanServ,
      cptr->name,
      modes);
    UpdateChanModes(Me.csptr, n_ChanServ, chptr, modes);
  }

  if (cptr->modes_off && !isminus)
  {
    modes[0] = '\0';
    if ((mode == MODE_S) &&
        (cptr->modes_off & MODE_S))
      sprintf(modes, "-s");
    if ((mode == MODE_P) &&
        (cptr->modes_off & MODE_P))
      sprintf(modes, "-p");
    if ((mode == MODE_N) &&
        (cptr->modes_off & MODE_N))
      sprintf(modes, "-n");
    if ((mode == MODE_T) &&
        (cptr->modes_off & MODE_T))
      sprintf(modes, "-t");
    if ((mode == MODE_M) &&
        (cptr->modes_off & MODE_M))
      sprintf(modes, "-m");
    if ((mode == MODE_C) &&
        (cptr->modes_off & MODE_C))
      sprintf(modes, "-c");
    if ((mode == MODE_I) &&
        (cptr->modes_off & MODE_I))
      sprintf(modes, "-i");
    if ((mode == MODE_L) &&
        (cptr->modes_off & MODE_L))
      sprintf(modes, "-l");
    if ((mode == MODE_K) &&
        (cptr->modes_off & MODE_K))
    {
      if (chptr->key)
        sprintf(modes, "-k %s", chptr->key);
    }

    if (modes[0])
    {
      toserv(":%s MODE %s %s\n",
        n_ChanServ,
        cptr->name,
        modes);
      UpdateChanModes(Me.csptr, n_ChanServ, chptr, modes);
    }
  }
} /* cs_CheckModes() */

/*
cs_CheckTopic()
 Called when a topic is changed - if topiclock is enabled, change
it back
*/

void
cs_CheckTopic(char *who, char *channel)

{
  struct ChanInfo *cptr;

  if (!(cptr = FindChan(channel)))
    return;

  if (cptr->flags & CS_TOPICLOCK)
  {
    if (IsFounder(FindClient(who), cptr))
      return;

    if (cptr->topic)
      cs_SetTopic(FindChannel(cptr->name), cptr->topic);
    else
      cs_SetTopic(FindChannel(cptr->name), "");
  }
} /* cs_CheckTopic() */

/*
cs_CheckOp()
  Check if nick has AutoOp access on 'chanptr', if so
op nick
*/

void
cs_CheckOp(struct Channel *chanptr, struct ChanInfo *cptr, char *nick)

{
  struct ChannelUser *tempuser;
  char modes[MAXLINE];

  if (!chanptr || !cptr || !nick)
    return;

    /*
     * If the user is flagged to be denied channel op
     * privileges, do nothing
     */
  if (HasFlag(nick, NS_NOCHANOPS))
    return;

  /*
   * If ChanServ should be on the channel, but isn't, don't
   * do anything
   */
  if (cs_ShouldBeOnChan(cptr) && !IsChannelOp(chanptr, Me.csptr))
    return;

  if (!(tempuser = FindUserByChannel(chanptr, FindClient(nick))))
    return;

  if (tempuser->flags & CH_OPPED)
  {
    /*
     * They're already opped on the channel - we don't need
     * to do anything
     */
    return;
  }

  if (HasAccess(cptr, tempuser->lptr, CA_AUTOOP))
  {
    sprintf(modes, "+o %s",
      tempuser->lptr->nick);
    toserv(":%s MODE %s %s\n",
      n_ChanServ,
      chanptr->name,
      modes);
    UpdateChanModes(Me.csptr, n_ChanServ, chanptr, modes);
  }
  else if (!(tempuser->flags & CH_VOICED) &&
        HasAccess(cptr, tempuser->lptr, CA_AUTOVOICE))
  {
    sprintf(modes, "+v %s",
      tempuser->lptr->nick);
    toserv(":%s MODE %s %s\n",
      n_ChanServ,
      chanptr->name,
      modes);
    UpdateChanModes(Me.csptr, n_ChanServ, chanptr, modes);
  }
} /* cs_CheckOp() */

/*
cs_CheckJoin()
  Check if 'nick' is on the AutoKick list for chanptr; if not, send
the entry message to 'nickname' (if there is one)
*/

void
cs_CheckJoin(struct Channel *chanptr, struct ChanInfo *cptr, char *nickname)

{
  struct Luser *lptr;
  struct AutoKick *ak;
  char chkstr[MAXLINE];
  char modes[MAXLINE];
  char *nick;

  if (!chanptr || !cptr || !nickname)
    return;

  nick = nickname;
  if (IsNickPrefix(*nick))
  {
    if (IsNickPrefix(*(++nick)))
      ++nick;
  }

  if (!(lptr = FindClient(nick)))
    return;

  if (FindService(lptr))
    return;

  /*
   * If the channel is restricted, and lptr just created it,
   * cs_CheckChan(), called from cs_CheckSjoin() would have
   * already kickbanned lptr, in which case they would no
   * longer be on the channel, but s_sjoin() would not know
   * that - catch it here.
   */
  if (!IsChannelMember(chanptr, lptr))
    return;

#ifndef HYBRID_ONLY
  /*
   * Check if ChanServ has joined the channel yet - this might
   * be the initial SJOIN burst in which case we can't do anything
   */
  if (!IsChannelMember(chanptr, Me.csptr))
    return;
#endif /* !HYBRID_ONLY */

  if (cptr->flags & CS_FORBID)
  {
    /*
     * Normally, we would never get here since ChanServ would have
     * kicked everyone out of the forbidden channel when services
     * was first started.  However, if an administrator uses
     * OMODE -i on the channel, they could conceivably join again,
     * so rekick them out
     */

    if (!IsChannelMember(chanptr, Me.csptr))
      cs_joinchan(cptr);

    toserv(":%s MODE %s +i\n",
      n_ChanServ,
      cptr->name);
    UpdateChanModes(Me.csptr, n_ChanServ, chanptr, "+i");
    KickBan(0, n_ChanServ, chanptr, lptr->nick, "Forbidden Channel");
    return;
  }

  /* someone joined - update lastused time */
  cptr->lastused = time(NULL);

  sprintf(chkstr, "%s!%s@%s",
    lptr->nick,
    lptr->username,
    lptr->hostname);

  if ((ak = OnAkickList(cptr, chkstr)))
  {
    /*
     * Don't kick founders who have identified - even if they match
     * an AKICK entry
     */
    if (IsFounder(lptr, cptr))
      return;

    if (chanptr->numusers == 1)
    {
      /*
       * This user is the only one on the channel -
       * if we kick him out, he could just rejoin
       * since the channel would have been dropped,
       * have ChanServ join the channel for a few
       * minutes
       */
      cs_joinchan(cptr);
    }

    sprintf(modes, "+b %s",
      ak->hostmask);
    toserv(":%s MODE %s %s\n",
      n_ChanServ,
      chanptr->name,
      modes);
    UpdateChanModes(Me.csptr, n_ChanServ, chanptr, modes);

    toserv(":%s KICK %s %s :%s\n",
      n_ChanServ,
      chanptr->name,
      lptr->nick,
      ak->reason ? ak->reason : "");
    RemoveFromChannel(chanptr, lptr);
  }
  else if ((cptr->flags & CS_RESTRICTED) &&
           !HasAccess(cptr, lptr, CA_AUTOOP))
  {
    char  *mask = HostToMask(lptr->username, lptr->hostname);

    if (!IsChannelMember(chanptr, Me.csptr))
      cs_joinchan(cptr);

    sprintf(modes, "+b *!%s",
      mask);

    toserv(":%s MODE %s %s\n",
      n_ChanServ,
      chanptr->name,
      modes);

    UpdateChanModes(Me.csptr, n_ChanServ, chanptr, modes);

    MyFree(mask);

    toserv(":%s KICK %s %s :%s\n",
      n_ChanServ,
      chanptr->name,
      lptr->nick,
      "Restricted Channel");

    RemoveFromChannel(chanptr, lptr);
  }
  else if (cptr->entrymsg)
  {
    /*
     * Send cptr->entrymsg to lptr->nick in the form of a NOTICE
     */
    toserv(":%s NOTICE %s :[%s] %s\n",
      n_ChanServ,
      lptr->nick,
      chanptr->name,
      cptr->entrymsg);
  }
} /* cs_CheckJoin() */

/*
cs_CheckSjoin()
 Called from s_sjoin() to check if the SJOINing nicks are
allowed to have operator status on "chptr->name"

nickcnt : number of SJOINing nicks
nicks   : array of SJOINing nicks
newchan : 1 if a newly created channel
*/

void
cs_CheckSjoin(struct Channel *chptr, struct ChanInfo *cptr,
              int nickcnt, char **nicks, int newchan)

{
  struct Luser *nlptr;
  char *dnicks, /* nicks to deop */
       *currnick;
  int ii; /* looping */

  if (!chptr || !cptr || !nicks)
    return;

  if (cptr->flags & CS_FORGET)
    return;

#ifndef HYBRID_ONLY

  if (IsChannelMember(chptr, Me.csptr))
  {
    /*
     * ChanServ is already on the channel so it is not
     * a brand new channel - we don't have to worry about
     * the new user(s) who joined
     */
    return;
  }

#endif /* !HYBRID_ONLY */

 /*
  * This is where ChanServ will join the channel if its
  * registered.  We just got the TS, so we know how to get
  * ChanServ opped. The thing is, ChanServ won't join any
  * registered channels which are empty on startup, since
  * there won't be an SJOIN associated with it; thus the
  * moment someone joins such a channel, an SJOIN will
  * be sent, causing ChanServ to join (theres little point in
  * ChanServ sitting in an empty channel anyway)
  */

#ifdef HYBRID_ONLY

  /*
   * Only call cs_join() if the channel is brand new,
   * or ChanServ will be changing the topic or setting
   * mlock modes everytime someone joins.
   */
  if (newchan)
    cs_join(cptr);

#else

  cs_join(cptr);

#endif /* HYBRID_ONLY */

  /*
   * First, if AllowAccessIfSOp is enabled, check if there
   * is a SuperOp currently opped on the channel. If there
   * is, return because we don't need to deop anyone in
   * that case. If the channel has RESTRICTED or SECUREOPS
   * enabled, we deop non-autops regardless of having
   * a SuperOp on the channel.
   */
  if (AllowAccessIfSOp &&
      !(cptr->flags & CS_RESTRICTED) &&
      !(cptr->flags & CS_SECUREOPS))
  {
    struct ChannelUser *tempu;

    for (tempu = chptr->firstuser; tempu; tempu = tempu->next)
    {
      if (HasAccess(cptr, tempu->lptr, CA_SUPEROP))
      {
        /*
         * There is at least 1 SuperOp opped on the channel
         * and SECUREOPS/RESTRICTED is off, so allow the 
         * other ops to stay opped, regardless of being
         * an Aop or higher
         */
        return;
      }
    }
  } /* if (AllowAccessIfSOp) */

 /*
   * deop the nick(s) who created the channel if they
   * aren't an autoop or higher
   */
  dnicks = (char *) MyMalloc(sizeof(char));
  dnicks[0] = '\0';

  for (ii = 0; ii < nickcnt; ii++)
  {
    currnick = GetNick(nicks[ii]);
    nlptr = FindClient(currnick);

    if (!nlptr)
      continue;

    if (FindService(nlptr))
      continue;

    if (!IsChannelOp(chptr, nlptr))
      continue;

    if (!HasFlag(nlptr->nick, NS_NOCHANOPS) &&
        HasAccess(cptr, nlptr, CA_AUTOOP))
      continue;

    dnicks = (char *) MyRealloc(dnicks, strlen(dnicks) + strlen(nlptr->nick) + (2 * sizeof(char)));
    strcat(dnicks, nlptr->nick);
    strcat(dnicks, " ");

    /*
     * Send them a notice so they know why they're being
     * deoped
     */
    if (GiveNotice)
    {
      toserv(":%s NOTICE %s :You do not have AutoOp access to [%s]\n",
        n_ChanServ,
        nlptr->nick,
        cptr->name);
    }
  }

  /*
   * Deop the non-autoops
   */
  SetModes(n_ChanServ, 0, 'o', chptr, dnicks);

  MyFree(dnicks);
} /* cs_CheckSjoin() */

/*
cs_ShouldBeOnChan()
 Determine whether ChanServ should be on the channel 'cptr'.
If so, return 1, otherwise 0
*/

int
cs_ShouldBeOnChan(struct ChanInfo *cptr)

{
  if (!cptr)
    return (0);

#ifndef HYBRID_ONLY

  /*
   * Since HYBRID_ONLY is not defined, ChanServ should be on
   * every single monitored channel
   */
  return (1);

#endif /* !HYBRID_ONLY */

  /*
   * Now, HYBRID_ONLY is defined, so generally ChanServ should NOT
   * be on the channel. However, if the channel has SET GUARD
   * enabled, ChanServ should be monitoring the channel.
   */
  if (cptr->flags & CS_GUARD)
    return (1);

  return (0);
} /* cs_ShouldBeOnChan() */

/*
cs_RejoinChannels()
 Have ChanServ rejoin all channel's it should be in
*/

void
cs_RejoinChannels()

{
  int ii;
  struct ChanInfo *cptr;

  for (ii = 0; ii < CHANLIST_MAX; ii++)
  {
    for (cptr = chanlist[ii]; cptr; cptr = cptr->next)
    {
      if (cptr->flags & CS_FORGET)
        continue;
      if (!cs_ShouldBeOnChan(cptr))
        continue;
      if (!FindChannel(cptr->name))
        continue;

      cs_join(cptr);
    }
  }
} /* cs_RejoinChannels() */

/*
PromoteSuccessor()
 Promote the successor on cptr to founder
*/

void
PromoteSuccessor(struct ChanInfo *cptr)

{
  assert(cptr && cptr->successor);

  putlog(LOG2,
    "%s: promoting successor [%s] to founder on [%s]",
    n_ChanServ,
    cptr->successor,
    cptr->name);

  if (cptr->founder)
    MyFree(cptr->founder);

  cptr->founder = cptr->successor;
  cptr->successor = NULL;
} /* PromoteSuccessor() */

/*
MakeChan()
 Create a ChanInfo structure and return it
*/

static struct ChanInfo *
MakeChan()

{
  struct ChanInfo *chanptr;

  chanptr = (struct ChanInfo *) MyMalloc(sizeof(struct ChanInfo));

  memset(chanptr, 0, sizeof(struct ChanInfo));

  return (chanptr);
} /* MakeChan() */

/*
ExpireChannels()
 Delete any channels that have expired.
*/

void
ExpireChannels(time_t unixtime)

{
  int ii;
  struct ChanInfo *cptr, *next;
  struct Channel *chptr;

  if (!ChannelExpire)
    return;

  for (ii = 0; ii < CHANLIST_MAX; ++ii)
  {
    for (cptr = chanlist[ii]; cptr; cptr = next)
    {
      next = cptr->next;

      if ((!(cptr->flags & (CS_FORBID | CS_FORGET | CS_NOEXPIRE))) &&
          ((unixtime - cptr->lastused) >= ChannelExpire))
      {
        putlog(LOG2,
          "%s: Expired channel: [%s]",
          n_ChanServ,
          cptr->name);

        chptr = FindChannel(cptr->name);
        if (IsChannelMember(chptr, Me.csptr))
          cs_part(chptr);

        DeleteChan(cptr);
      }
      if ((cptr->flags & CS_TFORBID) && (unixtime >= cptr->lastused))
      {
        putlog(LOG2,
          "%s: Expired forbidden channel: [%s]",
          n_ChanServ,
          cptr->name);

        chptr = FindChannel(cptr->name);
        if (IsChannelMember(chptr, Me.csptr))
          cs_part(chptr);

        DeleteChan(cptr);
      }
    } /* for (cptr = chanlist[ii]; cptr; cptr = next) */
  }
} /* ExpireChannels() */

/* #ifndef HYBRID_ONLY */

/*
CheckEmptyChans()
 Part any channels in which ChanServ is the only user
*/

void
CheckEmptyChans()

{
  struct UserChannel *tempc;
  struct UserChannel *temp;

  if (Me.csptr)
  {
    for (tempc = Me.csptr->firstchan; tempc; )
    {
      if (tempc->chptr->numusers <= 2)
      {
        /* ChanServ is the only client on the channel, part */
        SendUmode(OPERUMODE_Y,
          "%s: Parting empty channel [%s]",
          n_ChanServ,
          tempc->chptr->name);
	  
	/* take care of seenserv */
	toserv(":%s PRIVMSG SeenServ :!ssctl part %s\n",
	    n_ChanServ,
            tempc->chptr->name);

        putlog(LOG1,
	    "%s: Parting empty channel [%s]",
            n_ChanServ,
            tempc->chptr->name);

        temp = tempc->next;
        cs_part(tempc->chptr);
        tempc = temp;
      }
      else
        tempc = tempc->next;
    }
  }
} /* CheckEmptyChans() */

/* #endif */ /* !HYBRID_ONLY */

/*
FindChan()
  Return a pointer to registered channel 'channel'
*/

struct ChanInfo *
FindChan(char *channel)

{
  struct ChanInfo *cptr;
  unsigned int hashv;

  if (!channel)
    return (NULL);

  hashv = CSHashChan(channel);
  for (cptr = chanlist[hashv]; cptr; cptr = cptr->next)
    if (!strcasecmp(cptr->name, channel))
      return (cptr);

  return (NULL);
} /* FindChan() */

/*
AddFounder()
  Add founder 'lptr' to 'chanptr'
*/

static void
AddFounder(struct Luser *lptr, struct ChanInfo *chanptr)

{
  struct aChannelPtr *fcptr;
  struct f_users *fuptr;

  if (!lptr || !chanptr)
    return;

  fcptr = (struct aChannelPtr *) MyMalloc(sizeof(struct aChannelPtr));
  fcptr->cptr = chanptr;
  fcptr->next = lptr->founder_channels;
  lptr->founder_channels = fcptr;

  fuptr = (struct f_users *) MyMalloc(sizeof(struct f_users));
  fuptr->lptr = lptr;
  fuptr->next = chanptr->founders;
  chanptr->founders = fuptr;
} /* AddFounder() */

/*
RemFounder()
  Remove lptr->nick from the list of channel founders for 'cptr'
*/

void
RemFounder(struct Luser *lptr, struct ChanInfo *cptr)

{
  struct aChannelPtr *fcptr, *cprev;
  struct f_users *fuptr, *uprev;

  if (!cptr || !lptr)
    return;

  cprev = NULL;
  for (fcptr = lptr->founder_channels; fcptr; )
  {
    if (fcptr->cptr == cptr)
    {
      if (cprev)
      {
        cprev->next = fcptr->next;
        MyFree(fcptr);
        fcptr = cprev;
      }
      else
      {
        lptr->founder_channels = fcptr->next;
        MyFree(fcptr);
        fcptr = NULL;
      }

      /*
       * cptr should occur only once in the founder list -
       * break after we find it
       */
      break;
    }

    cprev = fcptr;

    if (fcptr)
      fcptr = fcptr->next;
    else
      fcptr = lptr->founder_channels;
  }

  uprev = NULL;
  for (fuptr = cptr->founders; fuptr; )
  {
    if (fuptr->lptr == lptr)
    {
      if (uprev)
      {
        uprev->next = fuptr->next;
        MyFree(fuptr);
        fuptr = uprev;
      }
      else
      {
        cptr->founders = fuptr->next;
        MyFree(fuptr);
        fuptr = NULL;
      }

      break;
    }

    uprev = fuptr;

    if (fuptr)
      fuptr = fuptr->next;
    else
      fuptr = cptr->founders;
  }
} /* RemFounder() */

static int
SetChanDefAccess(struct ChanInfo *cptr)
{

  cptr->access_lvl = (int *) MyMalloc(sizeof(int) * CA_SIZE);
  
  cptr->access_lvl[CA_AUTODEOP] = -1;
  cptr->access_lvl[CA_AUTOVOICE] = 5;
  cptr->access_lvl[CA_CMDVOICE] = 5;
  cptr->access_lvl[CA_ACCESS] = 9000;
  cptr->access_lvl[CA_CMDINVITE] = 5;
  cptr->access_lvl[CA_AUTOOP] = 10;
  cptr->access_lvl[CA_CMDOP] = 10;
  cptr->access_lvl[CA_CMDUNBAN] = 10;
  cptr->access_lvl[CA_AKICK] = 15;
  cptr->access_lvl[CA_CMDCLEAR] = 9000;
  cptr->access_lvl[CA_SET] = 10000;
  cptr->access_lvl[CA_SUPEROP] = 9500;
  cptr->access_lvl[CA_FOUNDER] = 10000;

  return 0;  
}

/*
ChangeChanPass()
  Change channel founder password for cptr->name to 'newpass'
*/

int
ChangeChanPass(struct ChanInfo *cptr, char *newpass)

{
  #ifdef CRYPT_PASSWORDS
    char  *encr;
  #endif

  if (!cptr || !newpass)
    return 0;

  if (!cptr->password)
  {
    /*
     * The password hasn't been set yet, so we're probably
     * registering the channel right now, thus we need to make
     * our own salt
     */
  #ifdef CRYPT_PASSWORDS

    static char saltChars[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";
    char salt[3];

    salt[0] = saltChars[random() % 64];
    salt[1] = saltChars[random() % 64];
    salt[2] = 0;

    encr = crypt(newpass, salt);
    assert(encr != 0);

    cptr->password = MyStrdup(encr);

  #else

    /* just use plaintext */
    cptr->password = MyStrdup(newpass);

  #endif
  }
  else
  {
    /* the password is being changed */

  #ifdef CRYPT_PASSWORDS

    encr = crypt(newpass, cptr->password);
    assert(encr != 0);

    MyFree(cptr->password);
    cptr->password = MyStrdup(encr);

  #else

    MyFree(cptr->password);
    cptr->password = MyStrdup(newpass);

  #endif

  } /* else */

  return 1;
} /* ChangeChanPass() */

/*
ChangeChanPermPass()
  Change channel perm password for cptr->name to 'newpass'
*/

static int
ChangeChanPermPass(struct ChanInfo *cptr, char *newpass)

{
  #ifdef CRYPT_PASSWORDS
    char  *encr;
  #endif

  if (!cptr || !newpass)
    return 0;

  if (!cptr->permpass)
  {
    /*
     * The perm password hasn't been set yet, so we're probably
     * registering the channel right now, thus we need to make
     * our own salt
     */
  #ifdef CRYPT_PASSWORDS

    static char saltChars[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";
    char salt[3];

    salt[0] = saltChars[random() % 64];
    salt[1] = saltChars[random() % 64];
    salt[2] = 0;

    encr = crypt(newpass, salt);
    assert(encr != 0);

    cptr->permpass = MyStrdup(encr);

  #else

    /* just use plaintext */
    cptr->permpass = MyStrdup(newpass);

  #endif
  }
  else
  {
    /* the password is being changed */

  #ifdef CRYPT_PASSWORDS

    encr = crypt(newpass, cptr->permpass);
    assert(encr != 0);

    MyFree(cptr->permpass);
    cptr->permpass = MyStrdup(encr);

  #else

    MyFree(cptr->permpass);
    cptr->permpass = MyStrdup(newpass);

  #endif

  } /* else */

  return 1;
} /* ChangeChanPermPass() */


/*
AddChan()
  Add 'chanptr' to chanlist table
*/

static void
AddChan(struct ChanInfo *chanptr)

{
  unsigned int hashv;

  /* 
   * all the fields of chanptr were filled in already, so just 
   * insert it in the list
   */
  hashv = CSHashChan(chanptr->name);

  chanptr->prev = NULL;
  chanptr->next = chanlist[hashv];
  if (chanptr->next)
    chanptr->next->prev = chanptr;

  chanlist[hashv] = chanptr;

  ++Network->TotalChans;
  if (chanptr->flags & CS_FORBID)
  {
    if (chanptr->flags & CS_TFORBID)
        ++Network->TForbiddenChans;
    else
	++Network->ForbiddenChans;
  }
  if (chanptr->flags & CS_FORGET)
    ++Network->ForgottenChans;
  if (chanptr->flags & CS_NOEXPIRE)
    ++Network->NoExpireChans;
  if (chanptr->flags & CS_GUARD)
    ++Network->GuardChans;
} /* AddChan() */

/*
DeleteChan()
 Deletes channel 'chanptr'
*/

void
DeleteChan(struct ChanInfo *chanptr)

{
  struct NickInfo *nptr;
  struct AutoKick *ak;
  struct ChanAccess *ca;
  struct f_users *fdrs;
  int hashv;
#ifdef MEMOSERVICES
  struct MemoInfo *mi;
#endif

  if (!chanptr)
    return;

  if (chanptr->flags & CS_FORBID)
  {
    if (chanptr->flags & CS_TFORBID)
        --Network->TForbiddenChans;
    else
	--Network->ForbiddenChans;
  }
  if (chanptr->flags & CS_FORGET)
    --Network->ForgottenChans;
  if (chanptr->flags & CS_NOEXPIRE)
    --Network->NoExpireChans;
  if (chanptr->flags & CS_GUARD)
    --Network->GuardChans;

  hashv = CSHashChan(chanptr->name);

  /* take care about seenserv */
    toserv(":%s PRIVMSG SeenServ :!ssctl leave %s\n",
        n_ChanServ,
        chanptr->name);

#ifdef MEMOSERVICES
  /* check if the chan had any memos - if so delete them */
  if ((mi = FindMemoList(chanptr->name)))
    DeleteMemoList(mi);
#endif

  if (chanptr->password)
    MyFree(chanptr->password);
  if (chanptr->permpass)
    MyFree(chanptr->permpass);
  if (chanptr->topic)
    MyFree(chanptr->topic);
  if (chanptr->key)
    MyFree(chanptr->key);

  while (chanptr->akick)
  {
    ak = chanptr->akick->next;
    MyFree(chanptr->akick->hostmask);
    if (chanptr->akick->reason)
      MyFree(chanptr->akick->reason);
    MyFree(chanptr->akick);
    chanptr->akick = ak;
  }

  while (chanptr->access)
  {
    ca = chanptr->access->next;

    /*
     * If there is an nptr entry in this access node, we must
     * delete it's corresponding AccessChannel entry.
     */
    if (chanptr->access->nptr && chanptr->access->acptr)
      DeleteAccessChannel(chanptr->access->nptr, chanptr->access->acptr);

    if (chanptr->access->hostmask)
      MyFree(chanptr->access->hostmask);
      
    if (chanptr->access->added)
      chanptr->access->added = NULL; 
    
    MyFree(chanptr->access);
    chanptr->access = ca;
  }

  while (chanptr->founders)
  {
    fdrs = chanptr->founders->next;
    RemFounder(chanptr->founders->lptr, chanptr);
    chanptr->founders = fdrs;
  }

  MyFree(chanptr->name);

  if (chanptr->founder)
  {
    if (!(chanptr->flags & (CS_FORBID | CS_FORGET)) &&
        (nptr = GetLink(chanptr->founder)))
      RemoveFounderChannelFromNick(&nptr, chanptr);

    MyFree(chanptr->founder);
  }

  if (chanptr->successor)
    MyFree(chanptr->successor);
  if (chanptr->email)
    MyFree(chanptr->email);
  if (chanptr->url)
    MyFree(chanptr->url);
  if (chanptr->entrymsg)
    MyFree(chanptr->entrymsg);
  if (chanptr->access_lvl)
    MyFree(chanptr->access_lvl);

  if (chanptr->next)
    chanptr->next->prev = chanptr->prev;
  if (chanptr->prev)
    chanptr->prev->next = chanptr->next;
  else
    chanlist[hashv] = chanptr->next;

  MyFree(chanptr);

  --Network->TotalChans;
} /* DeleteChan() */

/*
AddAccess()
  Add 'mask' or 'nptr' to the access list for 'chanptr' with 'level'
Return 1 if successful, 0 if not
*/

static int
AddAccess(struct ChanInfo *chanptr, struct Luser *lptr, char *mask,
          struct NickInfo *nptr, int level)

{
  struct ChanAccess *ptr;
  struct ChanAccess *ca;
  struct NickInfo   *tptr;
  char chkstr[MAXLINE];

  if (mask && strstr(mask, "**"))
    return 0;

  if (!chanptr || (!mask && !nptr))
    return 0;

  if (lptr) {
    sprintf(chkstr, "%s!%s@%s",
	lptr->nick,
        lptr->username,
	lptr->hostname);
    tptr = FindNick(lptr->nick);
    for (ca = chanptr->access; ca; ca = ca->next)
    {
	if (ca->hostmask)
	    if (match(ca->hostmask, chkstr))
		tptr = NULL;
    }
  } else {
    tptr = NULL;
  }

  if ((ptr = OnAccessList(chanptr, mask, nptr)))
  {
    /*
     * 'mask' is already on the access list - just change the
     * level
     */
    if (lptr)
      if (GetAccess(chanptr, lptr) <= ptr->level)
        return 0;

    ptr->level = level;

    if (tptr) {
	ptr->added = FindNick(lptr->nick);
    } else {
	ptr->added = NULL;
    }
    return 1;
  }

  ptr = (struct ChanAccess *) MyMalloc(sizeof(struct ChanAccess));
  memset(ptr, 0, sizeof(struct ChanAccess));

  if (nptr)
  {
    ptr->nptr = nptr;
    ptr->hostmask = NULL;

    /*
     * We also want the NickInfo structure to keep records
     * of what channels it has access on, so if we ever delete
     * the nick, we will know where to remove it's access.
     */
    ptr->acptr = AddAccessChannel(nptr, chanptr, ptr);
  }
  else
    ptr->hostmask = MyStrdup(mask);

  if (tptr) {
      ptr->added = FindNick(lptr->nick);
   } else {
      ptr->added = NULL;
   }

  ptr->level = level;

  ptr->prev = NULL;
  ptr->next = chanptr->access;
  if (ptr->next)
    ptr->next->prev = ptr;

  chanptr->access = ptr;

  return 1;
} /* AddAccess() */

/*
AddAccessDB()
  Add 'mask' or 'nptr' to the access list for 'chanptr' with 'level'
Return 1 if successful, 0 if not
*/

static int
AddAccessDB(struct ChanInfo *chanptr, struct NickInfo *lptr, char *mask,
          struct NickInfo *nptr, int level)

{
  struct ChanAccess *ptr;

  if (mask && strstr (mask, "**"))
      return 0;

  if (!chanptr || (!mask && !nptr))
    return 0;

  if ((ptr = OnAccessList(chanptr, mask, nptr)))
  {
    /*
     * 'mask' is already on the access list - just change the
     * level
     */
    ptr->level = level;
    return 1;
  }

  ptr = (struct ChanAccess *) MyMalloc(sizeof(struct ChanAccess));
  memset(ptr, 0, sizeof(struct ChanAccess));

  if (nptr)
  {
    ptr->nptr = nptr;
    ptr->hostmask = NULL;
    
    if (lptr) {
	ptr->added = lptr;
    } else {
	ptr->added = NULL;
    }
    
    /*
     * We also want the NickInfo structure to keep records
     * of what channels it has access on, so if we ever delete
     * the nick, we will know where to remove it's access.
     */
    ptr->acptr = AddAccessChannel(nptr, chanptr, ptr);
  }
  else
    ptr->hostmask = MyStrdup(mask);

  ptr->level = level;

  ptr->prev = NULL;
  ptr->next = chanptr->access;
  if (ptr->next)
    ptr->next->prev = ptr;

  chanptr->access = ptr;

  return 1;
} /* AddAccessDB() */

/*
DelAccess()
  Delete 'mask' from cptr's access list
Return -1 if 'mask' has a higher user level than lptr, otherwise
the number of matches
*/

static int
DelAccess(struct ChanInfo *cptr, struct Luser *lptr, char *mask,
          struct NickInfo *nptr)

{
  struct ChanAccess *temp, *next;
  int ret, cnt, ulev;
  int found;

  if (!cptr || !lptr || (!mask && !nptr))
    return 0;

  if (!cptr->access)
    return 0;

  ulev = GetAccess(cptr, lptr);
  ret = 0;
  cnt = 0;

  for (temp = cptr->access; temp; temp = next)
  {
    next = temp->next;

    found = 0;

    if (nptr && temp->nptr && (nptr == temp->nptr))
      found = 1;

    if (mask && temp->hostmask)
      if (match(mask, temp->hostmask))
        found = 1;

    if (found)
    {
        /*
        * don't let lptr->nick delete a hostmask that has a
        * >= level than it does
        */
      if (temp->level >= ulev)
      {
        ret = (-1);
        continue;
      }

      ++cnt;

      /*
       * nptr will have an AccessChannel entry which is set to 'temp',
       * so it must be deleted
       */
      if (nptr && temp->acptr)
        DeleteAccessChannel(nptr, temp->acptr);

      DeleteAccess(cptr, temp);
    } /* if (found) */
  }
  
  if (cnt > 0)
    return (cnt);

  return (ret);
} /* DelAccess() */

/*
DeleteAccess()
 Free the ChanAccess structure 'ptr'
*/

void
DeleteAccess(struct ChanInfo *cptr, struct ChanAccess *ptr)

{
  int idx;
  struct ChanAccess *ca;
  
  assert(cptr != 0);
  assert(ptr != 0);

  if (ptr->next)
    ptr->next->prev = ptr->prev;
  if (ptr->prev)
    ptr->prev->next = ptr->next;
  else
    cptr->access = ptr->next;

  if (ptr->hostmask)
    MyFree(ptr->hostmask);
    
  for (ca = cptr->access; ca; ca = ca->next, idx++)
  {
    if ((ca->added) && (ca->added == ptr->nptr))
	ca->added = NULL;
  }
  MyFree(ptr);
} /* DeleteAccess() */

/*
AddAkick()
  Add 'mask' to 'chanptr' 's autokick list
*/

static int
AddAkick(struct ChanInfo *chanptr, struct Luser *lptr, char *mask,
         char *reason)

{
  struct AutoKick *ptr;

  if (mask && strstr (mask, "**"))
      return 0;

  if (!chanptr || !mask)
    return 0;

  if (lptr)
  {
    struct ChanAccess *ca;
    int found = 0;

    /*
     * Check if lptr is trying to add an akick that matches a
     * level higher than their own
     */
    for (ca = chanptr->access; ca; ca = ca->next)
    {
      if (ca->hostmask)
        if (match(mask, ca->hostmask))
          found = 1;

      if (ca->nptr)
      {
        struct NickHost *hptr;
        char chkstr[MAXLINE];

        for (hptr = ca->nptr->hosts; hptr; hptr = hptr->next)
        {
          if (strchr(hptr->hostmask, '!'))
            strcpy(chkstr, hptr->hostmask);
          else
            sprintf(chkstr, "*!%s", hptr->hostmask);

          if (match(mask, chkstr))
          {
            found = 1;
            break;
          }
        }
      }

      if (found)
      {
        if (ca->level >= GetAccess(chanptr, lptr))
          return 0;
        found = 0;
      }
    }
  }

  ptr = (struct AutoKick *) MyMalloc(sizeof(struct AutoKick));
  ptr->hostmask = MyStrdup(mask);
  if (reason)
    ptr->reason = MyStrdup(reason);
  else
    ptr->reason = NULL;
  ptr->next = chanptr->akick;
  chanptr->akick = ptr;
  chanptr->akickcnt++;

  return 1;
} /* AddAkick() */

/*
DelAkick()
  Delete 'mask' from cptr's akick list; return number of matches
*/

static int
DelAkick(struct ChanInfo *cptr, char *mask)

{
  struct AutoKick *temp, *prev, *next;
  int cnt;

  if (!cptr || !mask)
    return 0;

  if (!cptr->akick)
    return 0;

  cnt = 0;

  prev = NULL;
  for (temp = cptr->akick; temp; temp = next)
  {
    next = temp->next;

    if (match(mask, temp->hostmask))
    {
      ++cnt;

      MyFree(temp->hostmask);
      if (temp->reason)
        MyFree(temp->reason);

      if (prev)
      {
        prev->next = temp->next;
        MyFree(temp);
        temp = prev;
      }
      else
      {
        cptr->akick = temp->next;
        MyFree(temp);
        temp = NULL;
      }
    }

    prev = temp;
  }

  cptr->akickcnt -= cnt;

  return (cnt);
} /* DelAkick() */

/*
IsFounder()
  Determine if lptr->nick has identified as a founder for 'cptr'
*/

static int
IsFounder(struct Luser *lptr, struct ChanInfo *cptr)

{
  struct f_users *temp;

#ifdef EMPOWERADMINS
  if (IsValidAdmin(lptr))
    return 1;
#endif

  for (temp = cptr->founders; temp; temp = temp->next)
    if (temp->lptr == lptr)
      return 1;

  return 0;
} /* IsFounder() */

/*
GetAccess()
  Return highest access level 'lptr' has on 'cptr'
*/

static int
GetAccess(struct ChanInfo *cptr, struct Luser *lptr)

{
  struct ChanAccess *ca;
  char chkstr[MAXLINE];
  int level = (-9999);
  struct NickInfo *nptr;
  int found;

  if (!cptr || !lptr)
    return 0;

  if (cptr->flags & (CS_FORBID | CS_FORGET))
    return 0;

  if (IsFounder(lptr, cptr))
    return (cptr->access_lvl[CA_FOUNDER]);

  sprintf(chkstr, "%s!%s@%s",
    lptr->nick,
    lptr->username,
    lptr->hostname);

  if ((nptr = FindNick(lptr->nick)))
  {
    /*
     * If nptr has not yet identified, don't give them access
     */
    if (!(nptr->flags & NS_IDENTIFIED))
    {
      if (cptr->flags & CS_SECURE)
      {
        /*
         * If the channel is secured, they MUST be
         * identified to get access - return 0
         */
        return (0);
      }

      if (!AllowAccessIfSOp)
        nptr = NULL;
      else if (!IsChannelOp(FindChannel(cptr->name), lptr))
        nptr = NULL;
      else
      {
        /*
         * They're opped on the channel, and AllowAccessIfSOp
         * is enabled, so allow them access. This will only
         * work during a netjoin, so ChanServ doesn't wrongfully
         * deop whole channels. If a channel op tries to
         * use this feature to get access without identifying,
         * cs_process() will stop them cold.
         */
      }
    }
  }
  else if (cptr->flags & CS_SECURE)
  {
    /*
     * The channel is secured, and they do not have a registered
     * nickname - return 0
     */
    return (0);
  }

  for (ca = cptr->access; ca; ca = ca->next)
  {
    found = 0;
    if (nptr && ca->nptr)
    {
      if (nptr == ca->nptr)
        found = 1;
      else
      {
      #ifdef LINKED_NICKNAMES
        /*
         * If nptr and ca->nptr are the in the same link,
         * give nptr the exact same access ca->nptr has
         */
        if (IsLinked(nptr, ca->nptr))
          found = 1;
      #endif /* LINKED_NICKNAMES */
      }
    }

    if (ca->hostmask)
      if (match(ca->hostmask, chkstr))
        found = 1;

    if (found)
      if (ca->level >= level)
      {
        /*
         * if ca->level is greater then level, keep going because
         * there may be another hostmask in the access list
         * which matches lptr's userhost which could have an even
         * bigger access level
         */
         level = ca->level;
      }
  }

  if (level == (-9999))
    return 0;
  else
    return (level);
} /* GetAccess() */

/*
HasAccess()
  Determine if lptr->nick has level on 'cptr'
*/

int
HasAccess(struct ChanInfo *cptr, struct Luser *lptr, int level)

{
  if (!cptr || !lptr)
    return 0;

  if (cptr->flags & (CS_FORBID | CS_FORGET))
    return 0;

  if ((cptr->flags & CS_SECURE) || (cptr->flags & CS_SECUREOPS))
  {
    struct NickInfo *nptr;

    if ((nptr = FindNick(lptr->nick)))
    {
      if (!(nptr->flags & NS_IDENTIFIED))
        return 0;
    }
    else
      return 0;
  }

#ifdef EMPOWERADMINS
  /*
   * If AutoOpAdmins is disabled, check if lptr is an admin and if
   * level is AUTOOP (or AUTOVOICE) - if so, check if they are on
   * the channel's access list for the appropriate level; if they
   * are, they are allowed to get opped/voiced (return 1), otherwise
   * return 0. This extra check is needed because GetAccess()
   * will return true if lptr is an admin, regardless of
   * AutoOpAdmins being enabled.
   */
  if (!AutoOpAdmins && IsValidAdmin(lptr) && ((level == CA_AUTOOP) || (level == CA_AUTOVOICE)))
  {
    struct ChanAccess *ca;
    char nmask[MAXLINE];

    sprintf(nmask, "%s!%s@%s",
      lptr->nick,
      lptr->username,
      lptr->hostname);
    if ((ca = OnAccessList(cptr, nmask, FindNick(lptr->nick))))
      if (ca->level >= cptr->access_lvl[level])
        return 1;

    return 0;
  }
#endif /* EMPOWERADMINS */

  if (GetAccess(cptr, lptr) >= cptr->access_lvl[level])
    return 1;

  return 0;
} /* HasAccess() */

/*
OnAccessList()
  Return 1 if 'hostmask' is on cptr's access list
*/

static struct ChanAccess *
OnAccessList(struct ChanInfo *cptr, char *hostmask, struct NickInfo *nptr)

{
  struct ChanAccess *ca;

  for (ca = cptr->access; ca; ca = ca->next)
  {
    if (nptr && ca->nptr)
      if (nptr == ca->nptr)
        return (ca);

    if (hostmask && ca->hostmask)
      if (match(ca->hostmask, hostmask))
        return (ca);
  }

  return (NULL);
} /* OnAccessList() */

/*
OnAkickList()
  Return 1 if 'hostmask' is on cptr's akick list
*/

static struct AutoKick *
OnAkickList(struct ChanInfo *cptr, char *hostmask)

{
  struct AutoKick *ak;

  for (ak = cptr->akick; ak; ak = ak->next)
    if (match(ak->hostmask, hostmask))
      return (ak);

  return (NULL);
} /* OnAkickList() */

/*
c_help()
  Give lptr->nick help on av[1]
*/

static void
c_help(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
/*  return 0;
   Screamer */

  if (ac >= 2)
   if ((av[1] != "") && (strlen(av[1]) > 0))
    if (strchr(av[1], '.') || strchr(av[1], '/')) {
      notice(n_ChanServ, lptr->nick,
	"You can NOT use ""."" or ""/"" character. Please remove it");
      return;
    }
			  
  if (ac >= 2)
  {
    char str[MAXLINE];

    if (ac >= 3)
      sprintf(str, "%s %s", av[1], av[2]);
    else
    {
      if ((!strcasecmp(av[1], "ACCESS")) ||
          (!strcasecmp(av[1], "AKICK")) ||
          (!strcasecmp(av[1], "LEVEL")) ||
          (!strcasecmp(av[1], "SET")))
        sprintf(str, "%s index", av[1]);
      else
      {
        struct Command *cptr;

        for (cptr = chancmds; cptr->cmd; cptr++)
          if (!strcasecmp(av[1], cptr->cmd))
            break;

        if (cptr->cmd)
          if ((cptr->level == LVL_ADMIN) &&
              !(IsValidAdmin(lptr)))
          {
            notice(n_ChanServ, lptr->nick,
              "No help available on \002%s\002",
              av[1]);
            return;
          }

        sprintf(str, "%s", av[1]);
      }
    }

    GiveHelp(n_ChanServ, lptr->nick, str, NODCC);
  }
  else
    GiveHelp(n_ChanServ, lptr->nick, (char *) NULL, NODCC);

  return;
} /* c_help() */

/*
c_register()
  Register channel av[1] with password av[2]
*/

static void
c_register(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  struct Userlist *userptr;
  struct NickInfo *master;
  time_t timetry;
  time_t datetry;
  char   dstr[MAXLINE];

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002REGISTER <#channel> <founder password> <perm password>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "REGISTER");
    return;
  }

  if (HasFlag(lptr->nick, NS_NOREGISTER))
  {
    notice(n_ChanServ, lptr->nick,
      "You are not permitted to register channels");
    putlog(LOG1,
      "Flagged user %s!%s@%s attempted to register channel [%s]",
      lptr->nick,
      lptr->username,
      lptr->hostname,
      av[1]);
    return;
  }

  userptr = GetUser(1, lptr->nick, lptr->username, lptr->hostname);
  if (RestrictRegister)
  {
    if (!(lptr->flags & L_OSREGISTERED) || !IsOper(userptr))
    {
      notice(n_ChanServ, lptr->nick,
        "Use of the [\002REGISTER\002] command is restricted");
      RecordCommand("%s: %s!%s@%s failed REGISTER [%s]",
        n_ChanServ,
        lptr->nick,
        lptr->username,
        lptr->hostname,
        av[1]);
      return;
    }
  }

  master = GetMaster(nptr);
  if (!IsValidAdmin(lptr) && MaxChansPerUser &&
      (master->fccnt >= MaxChansPerUser))
  {
    notice(n_ChanServ, lptr->nick,
      "You are only allowed to register [\002%d\002] channels",
      MaxChansPerUser);
    return;
  }

  if ((master->created + 604800 >= time(NULL))
  #ifdef EMPOWERADMINS
    && !(IsValidAdmin(lptr)))
  #else
    )
  #endif
  {
    timetry = (master->created + 604800) - time(NULL);
    datetry = master->created + 604800;
    strcpy(dstr, ctime(&datetry));
    dstr[strlen(dstr) - 1] = '\0';
    notice(n_ChanServ, lptr->nick,
      "Your nickname is TOO YOUNG.");
    notice(n_ChanServ, lptr->nick,
      "Only nicks older than 1 week can register channels.");
    notice(n_ChanServ, lptr->nick,
      "Try again after %s [%s]", timeago(time(NULL) - timetry, 1),
       dstr);
    return;  
  }

  if ((cptr = FindChan(av[1])))
  {
    if (cptr->flags & (CS_FORBID | CS_FORGET))
      RecordCommand("%s: Attempt to register forbidden channel [%s] by %s!%s@%s",
        n_ChanServ,
        cptr->name,
        lptr->nick,
        lptr->username,
        lptr->hostname);

    notice(n_ChanServ, lptr->nick,
      "The channel [\002%s\002] is already registered",
      av[1]);
    return;
  }

  if (!IsChannelOp(FindChannel(av[1]), lptr) && !(IsValidAdmin(lptr)))
  {
    notice(n_ChanServ, lptr->nick,
      "You are not a channel operator on [\002%s\002]",
      av[1]);
    return;
  }

  cptr = MakeChan();

  if (!ChangeChanPass(cptr, av[2]))
  {
    notice(n_ChanServ, lptr->nick,
      "Register failed");
    RecordCommand("%s: failed to create password for registered channel [%s]",
      n_ChanServ,
      av[1]);
    MyFree(cptr);
    return;
  }

  if (!ChangeChanPermPass(cptr, av[3]))
  {
    notice(n_ChanServ, lptr->nick,
      "Register failed");
    RecordCommand("%s: failed to create perm password for registered channel [%s]",
      n_ChanServ,
      av[1]);
    MyFree(cptr);
    return;
  }

  cptr->name = MyStrdup(av[1]);
  cptr->founder = MyStrdup(nptr->nick);
  cptr->created = cptr->lastused = time(NULL);

  SetChanDefAccess(cptr);
  
  /*
   * If the channel is registered by an operator or higher,
   * make the channel NoExpire
   */
  if ((lptr->flags & L_OSREGISTERED) && IsOper(userptr))
    cptr->flags |= CS_NOEXPIRE;

  AddAccess(cptr, 0, 0, nptr, cptr->access_lvl[CA_FOUNDER]);

  /* add cptr to channel list */
  AddChan(cptr);

  /* add channel to the master's founder channel list */
  AddFounderChannelToNick(&master, cptr);

  /* give lptr founder access */
  AddFounder(lptr, cptr);

  /*
   * Allow this call to cs_join even if HYBRID_ONLY is defined
   * so we can call cs_CheckChan() on it
   */
  cs_join(cptr);

  notice(n_ChanServ, lptr->nick,
    "The channel [\002%s\002] is now registered under your nickname",
    av[1]);
  notice(n_ChanServ, lptr->nick, "You have been added as Founder");
  notice(n_ChanServ, lptr->nick,
    "The founder password for %s is [\002%s\002] - Use it for identify",
    av[1],
    av[2]);

  notice(n_ChanServ, lptr->nick,
    "The permanent password for %s is [\002%s\002] - Remember this when using DROP and NEWPASS commands.",
    av[1], av[3]);

  notice(n_ChanServ, lptr->nick,
    "If someone stole your \002identify password\002 you can change it by typing \002/MSG %s NEWPASS %s <newpass> %s\002",
    n_ChanServ, av[1], av[3]);


  RecordCommand("%s: %s!%s@%s REGISTER [%s]",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    av[1]);
} /* c_register() */

/*
c_drop()
  Drop registered channel av[1]
*/

static void
c_drop(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  struct Channel *chptr;
  struct Userlist *utmp;
  int    isadmin = 0;


  if (lptr->flags & L_OSREGISTERED)
    utmp = GetUser(1, lptr->nick, lptr->username, lptr->hostname);
  else
    utmp = GetUser(0, lptr->nick, lptr->username, lptr->hostname);

  if (IsAdmin(utmp) && IsValidAdmin(lptr))
	    isadmin = 1;

  if ((ac < 3 && !isadmin) || ac < 2)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002DROP <channel> <perm password>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "DROP");
    return;
  }

  if (!(cptr = FindChan(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_CH_NOT_REGGED,
      av[1]);
    return;
  }

  if ((ac >= 3) && !isadmin)
  {
    if (!pwmatch(cptr->permpass, av[2]))
    {
      notice(n_ChanServ, lptr->nick,
        ERR_BAD_PASS);

      RecordCommand("%s: %s!%s@%s failed DROP [%s]",
        n_ChanServ,
        lptr->nick,
        lptr->username,
        lptr->hostname,
        cptr->name);

      return;
    }
  }
  else if (!IsFounder(lptr, cptr) && !isadmin)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002DROP <channel> <perm password>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "DROP");
    return;
  }

  if (isadmin)
    RecordCommand("%s: Administrator %s!%s@%s DROP [%s]",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name);
  else
    RecordCommand("%s: %s!%s@%s DROP [%s]",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name);

#ifdef UNIBG_SERVICES
 if (!(cptr->flags & CS_GUARD))
     cs_part(FindChannel(cptr->name));
#endif

  chptr = FindChannel(cptr->name);
  if (IsChannelMember(chptr, Me.csptr))
    cs_part(chptr);

  /* everything checks out ok...delete the channel */
  DeleteChan(cptr);

  notice(n_ChanServ, lptr->nick,
    "The channel [\002%s\002] has been dropped",
    av[1]);

  if (isadmin)
    toserv(":%s WALLOPS :%s!%s@%s DROP [%s]\n",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      av[1]);
} /* c_drop() */

/*
c_access()
  Modify the channel access list for av[1]
*/

static void
c_access(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  struct Command *cmdptr;

  if (ac < 3)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002ACCESS <channel> {ADD|DEL|LIST} [mask [level]]\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "ACCESS");
    return;
  }

  if (!(cptr = FindChan(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_CH_NOT_REGGED,
      av[1]);
    return;
  }

  if (!HasAccess(cptr, lptr, CA_ACCESS))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_NEED_ACCESS,
      cptr->access_lvl[CA_ACCESS],
      "ACCESS",
      cptr->name);
    RecordCommand("%s: %s!%s@%s failed ACCESS [%s] %s %s %s",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      StrToupper(av[2]),
      (ac >= 4) ? av[3] : "",
      (ac >= 5) ? av[4] : "");
    return;
  }  

  cmdptr = GetCommand(accesscmds, av[2]);

  if (cmdptr && (cmdptr != (struct Command *) -1))
  {
    /* call cmdptr->func to execute command */
    (*cmdptr->func)(lptr, nptr, ac, av);
  }
  else
  {
    /* the option they gave was not valid */
    notice(n_ChanServ, lptr->nick,
      "%s switch [\002%s\002]",
      (cmdptr == (struct Command *) -1) ? "Ambiguous" : "Unknown",
      av[2]);
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "ACCESS");
    return;
  }

  return;
} /* c_access() */

static void
c_access_add(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  char hostmask[MAXLINE];
  struct NickInfo *nickptr;
  int level; /* lptr->nick's access level */
  int newlevel;

  if (!(cptr = FindChan(av[1])))
    return;

  if (ac < 5)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002ACCESS <channel> ADD <hostmask> <level>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "ACCESS ADD");
    return;
  }

#ifdef UNIBG_SERVICES
	if ((atoi(av[4]) < DefaultAccess[CA_FOUNDER] * -1) || (atoi(av[4]) > DefaultAccess[CA_FOUNDER])) {
            notice(n_ChanServ, lptr->nick,
                "Invalid level, can't be greater/smaller than %ld", DefaultAccess[CA_FOUNDER] );
	    return;
	}
#endif

  if (((level = GetAccess(cptr, lptr)) <= atoi(av[4])) ||
      (atoi(av[4]) >= cptr->access_lvl[CA_FOUNDER]))
  {
    notice(n_ChanServ, lptr->nick,
      "You cannot add an access level greater than [\002%d\002]",
      level - 1);
    RecordCommand("%s: %s!%s@%s failed ACCESS [%s] ADD %s %s",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      av[3],
      av[4]);
    return;
  }

  nickptr = NULL;
  hostmask[0] = '\0';
  if (match("*!*@*", av[3]))
    strcpy(hostmask, av[3]);
  else if (match("*!*", av[3]))
  {
    strcpy(hostmask, av[3]);
    strcat(hostmask, "@*");
  }
  else if (match("*@*", av[3]))
  {
    strcpy(hostmask, "*!");
    strcat(hostmask, av[3]);
  }
  else if (match("*.*", av[3]))
  {
    strcpy(hostmask, "*!*@");
    strcat(hostmask, av[3]);
  }
  else
  {
    /* it must be a nickname */
    if (!(nickptr = FindNick(av[3])))
    {
      strcpy(hostmask, av[3]);
      strcat(hostmask, "!*@*");
    }
  }

  newlevel = atoi(av[4]);

  /* add hostmask (or nickptr) to the access list */
  if (AddAccess(cptr, lptr, hostmask, nickptr, newlevel ) )
  {
    notice(n_ChanServ, lptr->nick,
      "[\002%s\002] has been added to the access list for %s with level [\002%d\002]",
      nickptr ? nickptr->nick : hostmask,
      cptr->name,
      newlevel );

    RecordCommand("%s: %s!%s@%s ACCESS [%s] ADD %s %d",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      nickptr ? nickptr->nick : hostmask ? hostmask : "unknown!",
      newlevel );

  }
  else
  {
    notice(n_ChanServ, lptr->nick,
      "You may not modify the access entry [\002%s\002]",
      nickptr ? nickptr->nick : hostmask);

    RecordCommand("%s: %s!%s@%s failed ACCESS [%s] ADD %s %s",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      av[3],
      av[4]);
  }
} /* c_access_add() */

static void
c_access_del(struct Luser *lptr, struct NickInfo *nptr,
             int ac, char **av)

{
  struct ChanInfo *cptr;
  struct ChanAccess *temp;
  char *host;
  int cnt, idx;
  struct NickInfo *nickptr;

  if (!(cptr = FindChan(av[1])))
    return;

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002ACCESS <channel> DEL <hostmask | index>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "ACCESS DEL");
    return;
  }

  host = NULL;
  nickptr = NULL;

  if ((idx = IsNum(av[3])))
  {
    cnt = 0;
    for (temp = cptr->access; temp; temp = temp->next, ++cnt)
    {
      if (idx == (cnt + 1))
      {
        if (temp->hostmask)
          host = MyStrdup(temp->hostmask);
        else
          nickptr = temp->nptr;

        break;
      }
    }

    if (!host && !nickptr)
    {
      notice(n_ChanServ, lptr->nick,
        "[\002%d\002] is not a valid index",
        idx);

      notice(n_ChanServ, lptr->nick,
        ERR_MORE_INFO,
        n_ChanServ,
        "ACCESS LIST");

      return;
    }
  }
  else if (!(nickptr = FindNick(av[3])))
    host = MyStrdup(av[3]);

  if ((cnt = DelAccess(cptr, lptr, host, nickptr)) > 0)
  {
    notice(n_ChanServ, lptr->nick,
      "[\002%s\002] has been removed from the access list for %s",
      nickptr ? nickptr->nick : host,
      cptr->name);

    RecordCommand("%s: %s!%s@%s ACCESS [%s] DEL %s",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      nickptr ? nickptr->nick : host);
  }
  else
  {
    if (cnt == (-1))
    {
      notice(n_ChanServ, lptr->nick,
        "[\002%s\002] matches an access level higher than your own",
        nickptr ? nickptr->nick : host);

      RecordCommand("%s: %s!%s@%s failed ACCESS [%s] DEL %s",
        n_ChanServ,
        lptr->nick,
        lptr->username,
        lptr->hostname,
        cptr->name,
        nickptr ? nickptr->nick : host);
    }
    else
    {
      notice(n_ChanServ, lptr->nick,
        "[\002%s\002] was not found on the access list for %s",
        nickptr ? nickptr->nick : host,
        cptr->name);

      if (host)
        MyFree(host);

      return;
    }
  }

  if (host)
    MyFree(host);
} /* c_access_del() */

static void
c_access_list(struct Luser *lptr, struct NickInfo *nptr,
              int ac, char **av)

{
  struct ChanInfo *cptr;
  struct ChanAccess *ca;
  int idx;
  char *mask = NULL;

  if (!(cptr = FindChan(av[1])))
    return;

  if (ac >= 4)
    mask = av[3];

  RecordCommand("%s: %s!%s@%s ACCESS [%s] LIST %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    mask ? mask : "");

  if (cptr->access)
  {
    notice(n_ChanServ, lptr->nick,
      "-- Access List for [\002%s\002] --",
      cptr->name);
    notice(n_ChanServ, lptr->nick,
      "Num Level Hostmask");
    notice(n_ChanServ, lptr->nick,
      "--- ----- --------");
    idx = 1;
    for (ca = cptr->access; ca; ca = ca->next, idx++)
    {
      char  addnick[100];
	
      if (mask)
        if (match(mask, ca->hostmask ? ca->hostmask : ca->nptr->nick) == 0)
          continue;
      if ((ca->added) && (ca->added->nick)) {
          struct ChanAccess *cadd;
	  
          sprintf(addnick, "- Added by %s", ca->added->nick);
	  for (cadd = cptr->access; cadd; cadd = cadd->next) {
	      if (cadd->nptr)
	          if (match(cadd->nptr->nick, ca->added->nick)) {
		      sprintf(addnick, "%s (%d)", addnick, cadd->level);
		      break;
		  }
	  }
      } else {
	  if (ca->level == (10000)) {
	      strcpy(addnick, "- Added as Founder");
	  } else {
	      strcpy(addnick, "- Added by unknown");
	  }
      }
      notice(n_ChanServ, lptr->nick,
        "%-3d %-5d %-16s %s",
        idx,
        ca->level,
        ca->hostmask ? ca->hostmask : (ca->nptr ? ca->nptr->nick : ""),
	addnick);
    }
    notice(n_ChanServ, lptr->nick,
      "-- End of list --");
  }
  else
  {
    notice(n_ChanServ, lptr->nick,
      "The access list for [\002%s\002] is empty",
      cptr->name);
  }
} /* c_access_list() */

/*
c_akick()
  Modify the AKICK list for channel av[1]
*/

static void
c_akick(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  struct Command *cmdptr;

  if (ac < 3)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002AKICK <channel> {ADD|DEL|LIST} [mask]\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "AKICK");
    return;
  }

  if (!(cptr = FindChan(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_CH_NOT_REGGED,
      av[1]);
    return;
  }

  if (!HasAccess(cptr, lptr, CA_AKICK))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_NEED_ACCESS,
      cptr->access_lvl[CA_AKICK],
      "AKICK",
      cptr->name);
    RecordCommand("%s: %s!%s@%s failed AKICK [%s] %s %s %s",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      StrToupper(av[2]),
      (ac >= 4) ? av[3] : "",
      (ac >= 5) ? av[4] : "");
    return;
  }

  cmdptr = GetCommand(akickcmds, av[2]);

  if (cmdptr && (cmdptr != (struct Command *) -1))
  {
    /* call cmdptr->func to execute command */
    (*cmdptr->func)(lptr, nptr, ac, av);
  }
  else
  {
    /* the option they gave was not valid */
    notice(n_ChanServ, lptr->nick,
      "%s switch [\002%s\002]",
      (cmdptr == (struct Command *) -1) ? "Ambiguous" : "Unknown",
      av[2]);
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "AKICK");
    return;
  }

  return;
} /* c_akick() */

static void
c_akick_add(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  char hostmask[MAXLINE];
  char *reason;
  char freason[MAXLINE];

  if (!(cptr = FindChan(av[1])))
    return;

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002AKICK <channel> ADD <hostmask> [reason]\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "AKICK ADD");
    return;
  }

  if (MaxAkicks && (cptr->akickcnt >= MaxAkicks))
  {
    notice(n_ChanServ, lptr->nick,
      "Maximum autokick count for [\002%s\002] has been reached",
      cptr->name);
    return;
  }

  if (match("*!*@*", av[3]))
    strcpy(hostmask, av[3]);
  else if (match("*!*", av[3]))
  {
    strcpy(hostmask, av[3]);
    strcat(hostmask, "@*");
  }
  else if (match("*@*", av[3]))
  { 
    strcpy(hostmask, "*!");
    strcat(hostmask, av[3]);
  }
  else if (match("*.*", av[3]))
  {
    strcpy(hostmask, "*!*@");
    strcat(hostmask, av[3]);
  }
  else
  {
    struct Luser *ptr;
    char *mask;

    /* it must be a nickname - try to get hostmask */
    if ((ptr = FindClient(av[3])))
    {
      mask = HostToMask(ptr->username, ptr->hostname);
      sprintf(hostmask, "*!%s", mask);
      MyFree(mask);
    }
    else
    {
      strcpy(hostmask, av[3]);
      strcat(hostmask, "!*@*");
    }
  } 

  if (OnAkickList(cptr, hostmask))
  {
    notice(n_ChanServ, lptr->nick,
      "[\002%s\002] matches a hostmask already on the AutoKick list for %s",
        hostmask,
        cptr->name);
    return;
  }

  if (ac < 5)
  {
    sprintf(freason, "(by %s)", lptr->nick);
    reason = MyStrdup(freason);
  } else {
    reason = GetString(ac - 4, av + 4);
    sprintf(freason, "%s (by %s)", reason, lptr->nick);
    reason = MyStrdup(freason);
  }
  /* add hostmask to the akick list */
  if (AddAkick(cptr, lptr, hostmask, reason))
  {
    notice(n_ChanServ, lptr->nick,
      "[\002%s\002] has been added to the autokick list for %s with reason [%s]",
      hostmask,
      cptr->name,
      reason ? reason : "");

    RecordCommand("%s: %s!%s@%s AKICK [%s] ADD %s %s",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      hostmask,
      reason ? reason : "");
  }
  else
  {
    notice(n_ChanServ, lptr->nick,
      "You may not add [\002%s\002] to the autokick list for %s",
      hostmask,
      cptr->name);

    RecordCommand("%s: %s!%s@%s failed AKICK [%s] ADD %s %s",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      av[3],
      reason ? reason : "");
    if (reason)
      MyFree(reason);
    return;
  }

  if (reason)
    MyFree(reason);
} /* c_akick_add() */

static void
c_akick_del(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  struct AutoKick *temp;
  char *host = NULL;
  int cnt, idx;

  if (!(cptr = FindChan(av[1])))
    return;

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002AKICK <channel> DEL <hostmask | index>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "AKICK DEL");
    return;
  }

  if ((idx = IsNum(av[3])))
  {
    cnt = 0;
    for (temp = cptr->akick; temp; temp = temp->next, cnt++)
      if (idx == (cnt + 1))
      {
        host = MyStrdup(temp->hostmask);
        break;
      }

    if (!host)
    {
      notice(n_ChanServ, lptr->nick,
        "[\002%d\002] is not a valid index",
        idx);
      notice(n_ChanServ, lptr->nick,
        ERR_MORE_INFO,
        n_ChanServ,
        "AKICK LIST");
      return;
    }
  }
  else
    host = MyStrdup(av[3]);

  if (!DelAkick(cptr, host))
  {
    notice(n_ChanServ, lptr->nick,
      "[\002%s\002] was not found on the autokick list for %s",
      host,
      cptr->name);
    MyFree(host);
    return;
  }

  notice(n_ChanServ, lptr->nick,
    "[\002%s\002] has been removed from the autokick list for %s",
    host,
    cptr->name);

  RecordCommand("%s: %s!%s@%s AKICK [%s] DEL %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    host);

  MyFree(host);
} /* c_akick_del() */

static void
c_akick_list(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo  *cptr;
  struct AutoKick *ak;
  int idx;
  char *mask = NULL;

  if (!(cptr = FindChan(av[1])))
    return;

  if (ac >= 4)
    mask = av[3];

  RecordCommand("%s: %s!%s@%s AKICK [%s] LIST %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    mask ? mask : "");

  if (cptr->akick)
  {
    notice(n_ChanServ, lptr->nick,
      "-- AutoKick List for [\002%s\002] --",
      cptr->name);
    notice(n_ChanServ, lptr->nick,
      "Num %-35s Reason",
      "Hostmask");
    notice(n_ChanServ, lptr->nick,
      "--- %-35s ------",
      "--------");
    idx = 1;
    for (ak = cptr->akick; ak; ak = ak->next, idx++)
    {
      if (mask)
        if (match(mask, ak->hostmask) == 0)
          continue;
      notice(n_ChanServ, lptr->nick,
        "%-3d %-35s %s",
        idx,
        ak->hostmask,
        ak->reason ? ak->reason : "");
    }
    notice(n_ChanServ, lptr->nick,
      "-- End of list --");
  }
  else
  {
    notice(n_ChanServ, lptr->nick,
      "The autokick list for [\002%s\002] is empty",
      cptr->name);
  }
} /* c_akick_list() */

/*
c_list()
  Display list of channels matching av[1]
*/

static void
c_list(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *temp;
  int IsAnAdmin;
  int ii,
      mcnt, /* total matches found */
      acnt; /* total matches - private nicks */

  if (ac < 2)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002LIST <pattern>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO, 
      n_ChanServ,
      "LIST");
    return;
  }

  RecordCommand("%s: %s!%s@%s LIST %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    av[1]);

  if (IsValidAdmin(lptr))
    IsAnAdmin = 1;
  else
    IsAnAdmin = 0;

#ifdef UNIBG_SERVICES
  if ( !IsAnAdmin ) {
    notice(n_NickServ, lptr->nick,
        "List command is available only for services admins!");
    return;
  }
#endif

  acnt = mcnt = 0;
  notice(n_ChanServ, lptr->nick,
    "-- Listing channels matching [\002%s\002] --",
    av[1]);

  for (ii = 0; ii < CHANLIST_MAX; ++ii)
  {
    for (temp = chanlist[ii]; temp; temp = temp->next)
    {
      if (match(av[1], temp->name))
      {
        mcnt++;
        if ((IsAnAdmin) || !(temp->flags & CS_PRIVATE))
        {
          char  str[20];

          acnt++;
          if (temp->flags & CS_FORBID)
            strcpy(str, "<< FORBIDDEN >>");
          else if (temp->flags & CS_FORGET)
            strcpy(str, "<< FORGOTTEN >>");
          else if (temp->flags & CS_PRIVATE)
            strcpy(str, "<< PRIVATE >>");
          else if (FindChannel(temp->name))
            strcpy(str, "<< ACTIVE >>");
          else
            str[0] = '\0';

          notice(n_ChanServ, lptr->nick, 
            "%-10s %15s created %s ago",
            temp->name,
            str,
            timeago(temp->created, 1));
        }
      }
    }
  }

  notice(n_ChanServ, lptr->nick, 
    "-- End of list (%d/%d matches shown) --",
    acnt,
    mcnt);
} /* c_list() */

/*
c_level()
  Change the level(s) for different privileges on channel av[1]
*/

static void
c_level(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;

  if (ac < 3)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002LEVEL <channel> {SET|RESET|LIST} [index|type [level]]\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "LEVEL");
    return;
  }

  if (!(cptr = FindChan(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_CH_NOT_REGGED,
      av[1]);
    return;
  }

  if (!IsFounder(lptr, cptr))
  {
    notice(n_ChanServ, lptr->nick,
      "Founder access is required for [\002LEVEL\002]");
    RecordCommand("%s: %s!%s@%s failed LEVEL [%s] %s %s %s",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      av[2],
      (ac >= 4) ? av[3] : "",
      (ac >= 5) ? av[4] : "");
    return;
  }  

  if (!strcasecmp(av[2], "SET"))
  {
    int    newlevel, index;

    if (ac < 5)
    {
      notice(n_ChanServ, lptr->nick,
        "Syntax: \002LEVEL <channel> SET <index|type> <level>\002");
      notice(n_ChanServ, lptr->nick,
        ERR_MORE_INFO,
        n_ChanServ,
        "LEVEL SET");
      return;
    }

    index = (-1);
    if (!IsNum(av[3]))
    {
      if (!strcasecmp(av[3], "AUTODEOP"))
        index = CA_AUTODEOP;
      else if (!strcasecmp(av[3], "AUTOVOICE"))
        index = CA_AUTOVOICE;
      else if (!strcasecmp(av[3], "CMDVOICE"))
        index = CA_CMDVOICE;
      else if (!strcasecmp(av[3], "ACCESS"))
        index = CA_ACCESS;
      else if (!strcasecmp(av[3], "CMDINVITE"))
        index = CA_CMDINVITE;
      else if (!strcasecmp(av[3], "AUTOOP"))
        index = CA_AUTOOP;
      else if (!strcasecmp(av[3], "CMDOP"))
        index = CA_CMDOP;
      else if (!strcasecmp(av[3], "CMDUNBAN"))
        index = CA_CMDUNBAN;
      else if (!strcasecmp(av[3], "AUTOKICK"))
        index = CA_AKICK;
      else if (!strcasecmp(av[3], "CMDCLEAR"))
        index = CA_CMDCLEAR;
      else if (!strcasecmp(av[3], "SET"))
        index = CA_SET;
      else if (!strcasecmp(av[3], "SUPEROP"))
        index = CA_SUPEROP;

      if (index == (-1))
      {
        notice(n_ChanServ, lptr->nick,
          "Invalid type specified [\002%s\002]",
          av[3]);
        notice(n_ChanServ, lptr->nick,
          ERR_MORE_INFO,
          n_ChanServ,
          "LEVEL LIST");
        return;
      }
    }
    else
    {
      index = atoi(av[3]) - 1;
      if ((index >= CA_FOUNDER) || (index < CA_AUTODEOP))
      {
        notice(n_ChanServ, lptr->nick,
          "The index [\002%s\002] is not valid",
          av[3]);
        return;
      }
    }

    newlevel = atoi(av[4]);
    if (newlevel > cptr->access_lvl[CA_FOUNDER])
    {
      notice(n_ChanServ, lptr->nick,
        "You cannot create a level greater than [\002%d\002]",
        cptr->access_lvl[CA_FOUNDER]);
      return;
    }
    if (newlevel < -1)
    {
      notice(n_ChanServ, lptr->nick,
          "You cannot create a level less than -1",
          cptr->access_lvl[CA_FOUNDER]);
        return;
    }
    
    cptr->access_lvl[index] = newlevel;
    notice(n_ChanServ, lptr->nick,
      "The level required for [\002%s\002] has been changed to \002%d\002 on %s",
      accessinfo[index].cmd,
      newlevel,
      cptr->name);

    RecordCommand("%s: %s!%s@%s LEVEL [%s] SET %s %s",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      av[3],
      av[4]);

    return;
  }

  if (!strcasecmp(av[2], "RESET"))
  {
    int index;

    if (ac < 4)
    {
      notice(n_ChanServ, lptr->nick,
        "Syntax: \002LEVEL <channel> RESET <index|type|ALL>\002");
      notice(n_ChanServ, lptr->nick,
        ERR_MORE_INFO,
        n_ChanServ,
        "LEVEL RESET");
      return;
    }

    index = (-1);
    if (!IsNum(av[3]))
    {
      if (!strcasecmp(av[3], "AUTODEOP"))
        index = CA_AUTODEOP;
      else if (!strcasecmp(av[3], "AUTOVOICE"))
        index = CA_AUTOVOICE;
      else if (!strcasecmp(av[3], "CMDVOICE"))
        index = CA_CMDVOICE;
      else if (!strcasecmp(av[3], "ACCESS"))
        index = CA_ACCESS;
      else if (!strcasecmp(av[3], "CMDINVITE"))
        index = CA_CMDINVITE;
      else if (!strcasecmp(av[3], "AUTOOP"))
        index = CA_AUTOOP;
      else if (!strcasecmp(av[3], "CMDOP"))
        index = CA_CMDOP;
      else if (!strcasecmp(av[3], "CMDUNBAN"))
        index = CA_CMDUNBAN;
      else if (!strcasecmp(av[3], "AUTOKICK"))
        index = CA_AKICK;
      else if (!strcasecmp(av[3], "CMDCLEAR"))
        index = CA_CMDCLEAR;
      else if (!strcasecmp(av[3], "SET"))
        index = CA_SET;
      else if (!strcasecmp(av[3], "SUPEROP"))
        index = CA_SUPEROP;
      else if (!strcasecmp(av[3], "ALL"))
        index = (-2);

      if (index == (-1))
      {
        notice(n_ChanServ, lptr->nick,
          "Invalid type specified [\002%s\002]",
          av[3]);
        notice(n_ChanServ, lptr->nick,
          ERR_MORE_INFO,
          n_ChanServ,
          "LEVEL LIST");
        return;
      }
    }
    else
    {
      index = atoi(av[3]) - 1;
      if (index >= CA_FOUNDER)
      {
        notice(n_ChanServ, lptr->nick,
          "The index [\002%s\002] is not valid",
          av[3]);
        return;
      }
    }

    if (index == (-2))
    {

      /* reset everything */
      SetChanDefAccess(cptr);
    
      notice(n_ChanServ, lptr->nick,
        "The access level list for %s has been reset",
        cptr->name);
    }
    else
    {
      cptr->access_lvl[index] = DefaultAccess[index];
      notice(n_ChanServ, lptr->nick,
        "The level required for [\002%s\002] has been reset to \002%d\002 on %s",
        accessinfo[index].cmd,
        cptr->access_lvl[index],
        cptr->name);
    }

    RecordCommand("%s: %s!%s@%s LEVEL [%s] RESET %s",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      av[3]);

    return;
  } /* if (!strcasecmp(av[2], "RESET")) */

  if (!strcasecmp(av[2], "LIST"))
  {
    int ii;

    notice(n_ChanServ, lptr->nick,
      "-- Access Levels for [\002%s\002] --",
      cptr->name);
    notice(n_ChanServ, lptr->nick,
      "Index Level %-9s Description",
      "Type");
    notice(n_ChanServ, lptr->nick,
      "----- ----- %-9s -----------",
      "----");
    for (ii = 0; ii < CA_FOUNDER; ++ii)
    {
      notice(n_ChanServ, lptr->nick,
        "%-5d %-5d %-9s %s",
        ii + 1,
        cptr->access_lvl[ii],
        accessinfo[ii].cmd,
        accessinfo[ii].desc);
    }
    notice(n_ChanServ, lptr->nick,
      "-- End of list --");
    return;
  }

  /* the option they gave was not a SET, RESET, or LIST */
  notice(n_ChanServ, lptr->nick,
    "Syntax: \002LEVEL <channel> {SET|RESET|LIST} [index|type [level]]\002");
  notice(n_ChanServ, lptr->nick,
    ERR_MORE_INFO,
    n_ChanServ,
    "LEVEL");

  return;
} /* c_level() */

/*
c_identify()
  Check if lptr->nick supplied the correct founder password for av[1]
*/

static void
c_identify(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  char nmask[MAXLINE];

  if (ac < 3)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002IDENTIFY <channel> <password>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "IDENTIFY");
    return;
  }

  if (!(cptr = FindChan(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_CH_NOT_REGGED,
      av[1]);
    return;
  }

  if (!pwmatch(cptr->password, av[2]))
  {
    notice(n_ChanServ, lptr->nick, ERR_BAD_PASS);

    RecordCommand("%s: %s!%s@%s failed IDENTIFY [%s]",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      av[1]);

    return;
  }

  if (IsFounder(lptr, cptr))
  {
    notice(n_ChanServ, lptr->nick,
      "You already have founder access to [\002%s\002]",
      av[1]);
    return;
  }

  AddFounder(lptr, cptr);

  notice(n_ChanServ, lptr->nick,
    "Password accepted - you are now recognized as a founder for [\002%s\002]",
    av[1]);

  sprintf(nmask, "%s!%s@%s",
    lptr->nick,
    lptr->username,
    lptr->hostname);
  if (!OnAccessList(cptr, nmask, nptr))
  {
    AddAccess(cptr, (struct Luser *) NULL, (char *) NULL, nptr, cptr->access_lvl[CA_SUPEROP]);
    notice(n_ChanServ, lptr->nick,
      "You have been added to %s as a SuperOp",
      cptr->name);
  }

  RecordCommand("%s: %s!%s@%s IDENTIFY [%s]",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name);
} /* c_identify() */

static void
c_newpass(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002NEWPASS <channel> <newpass> <perm password>\002");
        notice(n_ChanServ, lptr->nick,
        ERR_MORE_INFO,
        n_ChanServ,
	"NEWPASS");
    return;
  }

  if (!(cptr = FindChan(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_CH_NOT_REGGED,
        av[1]);
    return;
  }

  if (cptr->flags & CS_FORBID)
  {
    notice(n_ChanServ, lptr->nick,
	"Cannot execute commands for forbidden channels");
    return;
  }

  if (cptr->flags & CS_FORGET)
  {
    notice(n_ChanServ, lptr->nick,
	"Cannot execute commands for forgotten channels");
    return;
  }

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002NEWPASS <channel> <newpass> <perm password>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "SET PASSWORD");
    return;
  }

  if (!pwmatch(cptr->permpass, av[3]))
  {
    notice(n_ChanServ, lptr->nick, ERR_BAD_PASS);
    RecordCommand("%s: %s!%s@%s failed NEWPASS [%s]",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      av[1]);
    return;
  }

  RecordCommand("%s: %s!%s@%s NEWPASS [%s] ...",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name);

  if (!ChangeChanPass(cptr, av[2]))
  {
    notice(n_ChanServ, lptr->nick,
      "Password change failed");
    return;
  }
  else
  {
    notice(n_ChanServ, lptr->nick,
      "The password for %s has been changed to [\002%s\002]",
      cptr->name,
      av[2]);
    return;
  }
} /* c_newpass() */

/*
c_set()
  Set various channel options
*/

static void
c_set(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  struct Command *cmdptr;

  if (ac < 3)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002SET <channel> <option> [parameter]\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "SET");
    return;
  }

  if (!(cptr = FindChan(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_CH_NOT_REGGED,
      av[1]);
    return;
  }

  if (!HasAccess(cptr, lptr, CA_SET))
  {
    char tmp[MAXLINE];

    notice(n_ChanServ, lptr->nick,
      ERR_NEED_ACCESS,
      cptr->access_lvl[CA_SET],
      "SET",
      av[1]);

    /* static buffers .. argh */
    Strncpy(tmp, StrToupper(av[2]), sizeof(tmp) - 1);
    tmp[sizeof(tmp) - 1] = '\0';

    RecordCommand("%s: %s!%s@%s failed SET [%s] %s %s",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      tmp,
      (ac < 4) ? "" : StrToupper(av[3]));

    return;
  }

  cmdptr = GetCommand(setcmds, av[2]);

  if (cmdptr && (cmdptr != (struct Command *) -1))
  {
    /* call cmdptr->func to execute command */
    (*cmdptr->func)(lptr, nptr, ac, av);
  }
  else
  {
    /* the option they gave was not valid */
    notice(n_ChanServ, lptr->nick,
      "%s switch [\002%s\002]",
      (cmdptr == (struct Command *) -1) ? "Ambiguous" : "Unknown",
      av[2]);
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "SET");
    return;
  }
} /* c_set() */

static void
c_set_topiclock(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;

  if (!(cptr = FindChan(av[1])))
    return;

  RecordCommand("%s: %s!%s@%s SET [%s] TOPICLOCK %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    (ac < 4) ? "" : StrToupper(av[3]));

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Topic Lock for channel %s is [\002%s\002]",
      cptr->name,
      (cptr->flags & CS_TOPICLOCK) ? "ON" : "OFF");
    return;
  }

  if (!strcasecmp(av[3], "ON"))
  {
    cptr->flags |= CS_TOPICLOCK;
    notice(n_ChanServ, lptr->nick,
      "Toggled Topic Lock for channel %s [\002ON\002]",
      cptr->name);
    return;
  }

  if (!strcasecmp(av[3], "OFF"))
  {
    cptr->flags &= ~CS_TOPICLOCK;
    notice(n_ChanServ, lptr->nick,
      "Toggled Topic Lock for channel %s [\002OFF\002]",
      cptr->name);
    return;
  }

  /* user gave an unknown param */
  notice(n_ChanServ, lptr->nick,
    "Syntax: \002SET <channel> TOPICLOCK {ON|OFF}\002");
  notice(n_ChanServ, lptr->nick,
    ERR_MORE_INFO,
    n_ChanServ,
    "SET TOPICLOCK");
} /* c_set_topiclock() */

static void
c_set_private(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;

  if (!(cptr = FindChan(av[1])))
    return;

  RecordCommand("%s: %s!%s@%s SET [%s] PRIVATE %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    (ac < 4) ? "" : StrToupper(av[3]));

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Privacy for channel %s is [\002%s\002]",
      cptr->name,
      (cptr->flags & CS_PRIVATE) ? "ON" : "OFF");
    return;
  }

  if (!strcasecmp(av[3], "ON"))
  {
    cptr->flags |= CS_PRIVATE;
    notice(n_ChanServ, lptr->nick,
      "Toggled Privacy for channel %s [\002ON\002]",
      cptr->name);
    return;
  }

  if (!strcasecmp(av[3], "OFF"))
  {
    cptr->flags &= ~CS_PRIVATE;
    notice(n_ChanServ, lptr->nick,
      "Toggled Privacy for channel %s [\002OFF\002]",
      cptr->name);
    return;
  }

  /* user gave an unknown param */
  notice(n_ChanServ, lptr->nick,
    "Syntax: \002SET <channel> PRIVATE {ON|OFF}\002");
  notice(n_ChanServ, lptr->nick,
    ERR_MORE_INFO,
    n_ChanServ,
    "SET PRIVATE");
} /* c_set_private() */

static void
c_set_secure(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;

  if (!(cptr = FindChan(av[1])))
    return;

  RecordCommand("%s: %s!%s@%s SET [%s] SECURE %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    (ac < 4) ? "" : StrToupper(av[3]));

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Security for channel %s is [\002%s\002]",
      cptr->name,
      (cptr->flags & CS_SECURE) ? "ON" : "OFF");
    return;
  }

  if (!strcasecmp(av[3], "ON"))
  {
    cptr->flags |= CS_SECURE;
    notice(n_ChanServ, lptr->nick,
      "Toggled Security for channel %s [\002ON\002]",
      cptr->name);
    return;
  }

  if (!strcasecmp(av[3], "OFF"))
  {
    cptr->flags &= ~CS_SECURE;
    notice(n_ChanServ, lptr->nick,
      "Toggled Security for channel %s [\002OFF\002]",
      cptr->name);
    return;
  }

  /* user gave an unknown param */
  notice(n_ChanServ, lptr->nick,
    "Syntax: \002SET <channel> SECURE {ON|OFF}\002");
  notice(n_ChanServ, lptr->nick,
    ERR_MORE_INFO,
    n_ChanServ,
    "SET SECURE");
} /* c_set_secure() */

static void
c_set_secureops(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;

  if (!(cptr = FindChan(av[1])))
    return;

  RecordCommand("%s: %s!%s@%s SET [%s] SECUREOPS %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    (ac < 4) ? "" : StrToupper(av[3]));

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "SecureOps for channel %s is [\002%s\002]",
      cptr->name,
      (cptr->flags & CS_SECUREOPS) ? "ON" : "OFF");
    return;
  }

  if (!strcasecmp(av[3], "ON"))
  {
    cptr->flags |= CS_SECUREOPS;
    notice(n_ChanServ, lptr->nick,
      "Toggled SecureOps for channel %s [\002ON\002]",
      cptr->name);
    return;
  }

  if (!strcasecmp(av[3], "OFF"))
  {
    cptr->flags &= ~CS_SECUREOPS;
    notice(n_ChanServ, lptr->nick,
      "Toggled SecureOps for channel %s [\002OFF\002]",
      cptr->name);
    return;
  }

  /* user gave an unknown param */
  notice(n_ChanServ, lptr->nick,
    "Syntax: \002SET <channel> SECUREOPS {ON|OFF}\002");
  notice(n_ChanServ, lptr->nick,
    ERR_MORE_INFO,
    n_ChanServ,
    "SET SECUREOPS");
} /* c_set_secureops() */

static void
c_set_restricted(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;

  if (!(cptr = FindChan(av[1])))
    return;

  RecordCommand("%s: %s!%s@%s SET [%s] RESTRICTED %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    (ac < 4) ? "" : StrToupper(av[3]));

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Restricted Access for channel %s is [\002%s\002]",
      cptr->name,
      (cptr->flags & CS_RESTRICTED) ? "ON" : "OFF");
    return;
  }

  if (!strcasecmp(av[3], "ON"))
  {
    cptr->flags |= CS_RESTRICTED;
    notice(n_ChanServ, lptr->nick,
      "Toggled Restricted Access for channel %s [\002ON\002]",
      cptr->name);
    return;
  }

  if (!strcasecmp(av[3], "OFF"))
  {
    cptr->flags &= ~CS_RESTRICTED;
    notice(n_ChanServ, lptr->nick,
      "Toggled Restricted Access for channel %s [\002OFF\002]",
      cptr->name);
    return;
  }

  /* user gave an unknown param */
  notice(n_ChanServ, lptr->nick,
    "Syntax: \002SET <channel> RESTRICTED {ON|OFF}\002");
  notice(n_ChanServ, lptr->nick,
    ERR_MORE_INFO,
    n_ChanServ,
    "SET RESTRICTED");
} /* c_set_restricted() */

static void
c_set_forget(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  char freason[MAXLINE];

  if (!(cptr = FindChan(av[1])))
    return;

  if (!IsFounder(lptr, cptr))
  {
    notice(n_ChanServ, lptr->nick,
      "Founder access is required for [\002SET FORGET\002]");
    RecordCommand("%s: %s!%s@%s failed SET [%s] FORGET %s",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      (ac < 4) ? "" : StrToupper(av[3]));
    return;
  }

  RecordCommand("%s: %s!%s@%s SET [%s] FORGET %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    (ac < 4) ? "" : StrToupper(av[3]));

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Forgotten for channel %s is [\002%s\002]",
      cptr->name,
      (cptr->flags & CS_FORGET) ? "ON" : "OFF");
    return;
  }

  if (!strcasecmp(av[3], "ON"))
  {
    cptr->flags |= CS_FORGET;
    notice(n_ChanServ, lptr->nick,
      "Toggled Forgotten for channel %s [\002ON\002]",
      cptr->name);

    if (cs_ShouldBeOnChan(cptr)) {
      cs_part(FindChannel(cptr->name)); /* leave the channel */
      
      sprintf(freason, "by founder (%s)", lptr->nick);
      cptr->topic = MyStrdup(freason);

      /* take care about seenserv */
      toserv(":%s PRIVMSG SeenServ :!ssctl leave %s\n",
        n_ChanServ,
        cptr->name);
    }
    return;
  }

  if (!strcasecmp(av[3], "OFF"))
  {
    cptr->flags &= ~CS_FORGET;
    notice(n_ChanServ, lptr->nick,
      "Toggled Forgotten for channel %s [\002OFF\002]",
      cptr->name);
    return;
  }

  /* user gave an unknown param */
  notice(n_ChanServ, lptr->nick,
    "Syntax: \002SET <channel> FORGET {ON|OFF}\002");
  notice(n_ChanServ, lptr->nick,
    ERR_MORE_INFO,
    n_ChanServ,
    "SET FORGET");
} /* c_set_forget() */

static void
c_set_guard(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  struct Channel *chptr;
  
  if (!(cptr = FindChan(av[1])))
    return;
  if (!(chptr = FindChannel(av[1])))
    return;
    
  RecordCommand("%s: %s!%s@%s SET [%s] GUARD %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    (ac < 4) ? "" : StrToupper(av[3]));

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "ChanGuard for channel %s is [\002%s\002]",
      cptr->name,
      (cptr->flags & CS_GUARD) ? "ON" : "OFF");
    return;
  }

  if (!strcasecmp(av[3], "ON"))
  {
    if ((chptr->numusers < 5) && !IsValidAdmin(lptr))
    {
	notice(n_ChanServ, lptr->nick,
	    "There are less than 5 users in channel %s. ChanGuard will remain [\002OFF\002]",
	    cptr->name);
	return;
    }
    cptr->flags |= CS_GUARD;
    notice(n_ChanServ, lptr->nick,
      "Toggled ChanGuard for channel %s [\002ON\002]",
      cptr->name);
      
    /* seenserv rullz */
    notice(n_ChanServ, lptr->nick,
      "You can put SeenServ in your channel using /MSG CS SET %s SEENSERV ON",
      cptr->name);
    ++Network->GuardChans;
    cs_join(cptr);
    return;
  }

  if (!strcasecmp(av[3], "OFF"))
  {
    cptr->flags &= ~CS_GUARD;
    notice(n_ChanServ, lptr->nick,
      "Toggled ChanGuard for channel %s [\002OFF\002]",
      cptr->name);
      
    /* seenserv rullz */
    notice(n_ChanServ, lptr->nick,
      "You can remove SeenServ from your channel using /MSG CS SET %s SEENSERV OFF",
      cptr->name);
    cs_part(FindChannel(cptr->name));
    --Network->GuardChans;
    return;
  }

  /* user gave an unknown param */
  notice(n_ChanServ, lptr->nick,
    "Syntax: \002SET <channel> GUARD {ON|OFF}\002");
  notice(n_ChanServ, lptr->nick,
    ERR_MORE_INFO,
    n_ChanServ,
    "SET GUARD");
} /* c_set_guard() */

static void
c_set_seenserv(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  struct Channel *chptr;
  
  if (!(cptr = FindChan(av[1])))
    return;
  if (!(chptr = FindChannel(av[1])))
    return;
    
  RecordCommand("%s: %s!%s@%s SET [%s] SEENSERV %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    (ac < 4) ? "" : StrToupper(av[3]));

  if (ac < 4)
  {
      notice(n_ChanServ, lptr->nick,
        "Syntax: \002SET <channel> SEENSERV {ON|OFF|PUBLIC|PRIVATE}\002");
      notice(n_ChanServ, lptr->nick,
        ERR_MORE_INFO,
        n_ChanServ,
        "SET SEENSERV");
    return;
  }

  if (!strcasecmp(av[3], "ON"))
  {
    if ((chptr->numusers < 5) && !IsValidAdmin(lptr))
    {
	notice(n_ChanServ, lptr->nick,
	    "There are less than 5 users in channel %s. SeenServ will not join",
	    cptr->name);
	return;
    }
    toserv(":%s PRIVMSG SeenServ :!ssctl join %s\n",
	n_ChanServ,
	cptr->name);
    notice(n_ChanServ, lptr->nick,
      "Toggled ChanSeen for channel %s [\002ON\002]",
      cptr->name);
    return;
  }

  if (!strcasecmp(av[3], "OFF"))
  {
    toserv(":%s PRIVMSG SeenServ :!ssctl leave %s\n",
	n_ChanServ,
	cptr->name);
    notice(n_ChanServ, lptr->nick,
      "Toggled ChanSeen for channel %s [\002OFF\002]",
      cptr->name);
    return;
  }

  if (!strcasecmp(av[3], "PUBLIC"))
  {
    toserv(":%s PRIVMSG SeenServ :!ssctl answerpub %s\n",
	n_ChanServ,
	cptr->name);
    notice(n_ChanServ, lptr->nick,
      "Toggled ChanSeen for channel %s [\002PUBLIC\002]",
      cptr->name);
    return;
  }

  if (!strcasecmp(av[3], "PRIVATE"))
  {
    toserv(":%s PRIVMSG SeenServ :!ssctl answerprivate %s\n",
	n_ChanServ,
	cptr->name);
    notice(n_ChanServ, lptr->nick,
      "Toggled ChanSeen for channel %s [\002PRIVATE\002]",
      cptr->name);
    return;
  }

  /* user gave an unknown param */
  notice(n_ChanServ, lptr->nick,
    "Syntax: \002SET <channel> SEENSERV {ON|OFF|PUBLIC|PRIVATE}\002");
  notice(n_ChanServ, lptr->nick,
    ERR_MORE_INFO,
    n_ChanServ,
    "SET SEENSERV");
} /* c_set_seenserv() */

static void
c_set_password(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;

  if (!(cptr = FindChan(av[1])))
    return;

  if (!IsFounder(lptr, cptr))
  {
    notice(n_ChanServ, lptr->nick,
      "Only founders can change the channel password");
    RecordCommand("%s: %s!%s@%s failed SET [%s] PASSWORD ...",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name);
    return;
  }

  RecordCommand("%s: %s!%s@%s SET [%s] PASSWORD ...",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name);

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002SET <channel> PASSWORD <newpass>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "SET PASSWORD");
    return;
  }

  if (!ChangeChanPass(cptr, av[3]))
  {
    notice(n_ChanServ, lptr->nick,
      "Password change failed");
    return;
  }
  else
  {
    notice(n_ChanServ, lptr->nick,
      "The password for %s has been changed to [\002%s\002]",
      cptr->name,
      av[3]);
    return;
  }
} /* c_set_password() */

static void
c_set_founder(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  struct NickInfo *fptr;

  if (!(cptr = FindChan(av[1])))
    return;

  if (!IsFounder(lptr, cptr))
  {
    notice(n_ChanServ, lptr->nick,
      "You must IDENTIFY as a founder before resetting the founder nickname");
    RecordCommand("%s: %s!%s@%s failed SET [%s] FOUNDER %s",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      (ac < 4) ? "" : av[3]);

    return;
  }

  if (ac < 5)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002SET <channel> FOUNDER <nickname> <perm pass>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "SET FOUNDER");
    return;
  }

  if (!(fptr = GetLink(av[3])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_NOT_REGGED,
      av[3]);
    return;
  }

  if (!pwmatch(cptr->permpass, av[4]))
    {
	notice(n_ChanServ, lptr->nick, ERR_BAD_PASS);
	RecordCommand("%s: %s!%s@%s failed SET [%s] FOUNDER %s",
          n_ChanServ,
          lptr->nick,
          lptr->username,
	  lptr->hostname,
          cptr->name,
          av[3]);
	return;
    }

  RecordCommand("%s: %s!%s@%s SET [%s] FOUNDER %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    (ac < 4) ? "" : av[3]);

  RemoveFounderChannelFromNick(&fptr, cptr);

  if (cptr->founder)
  {
    struct NickInfo *oldnick = GetLink(cptr->founder);

    if (oldnick)
      RemoveFounderChannelFromNick(&oldnick, cptr);

    MyFree(cptr->founder);
  }

  AddFounderChannelToNick(&fptr, cptr);

  cptr->founder = MyStrdup(fptr->nick);


  notice(n_ChanServ, lptr->nick,
    "The founder nickname for %s has been changed to [\002%s\002]",
    cptr->name,
    fptr->nick);
} /* c_set_founder() */

static void
c_set_successor(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  struct NickInfo *fptr;

  if (!(cptr = FindChan(av[1])))
    return;

  if (!IsFounder(lptr, cptr))
  {
    notice(n_ChanServ, lptr->nick,
      "You must IDENTIFY as a founder before setting the successor nickname");
    RecordCommand("%s: %s!%s@%s failed SET [%s] SUCCESSOR %s",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      (ac < 4) ? "" : av[3]);
    return;
  }

  RecordCommand("%s: %s!%s@%s SET [%s] SUCCESSOR %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    (ac < 4) ? "" : av[3]);

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002SET <channel> SUCCESSOR <nickname>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "SET SUCCESSOR");
    return;
  }

  if (!(fptr = FindNick(av[3])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_NOT_REGGED,
      av[3]);
    return;
  }

  if (cptr->successor)
    MyFree(cptr->successor);

  cptr->successor = MyStrdup(fptr->nick);

  notice(n_ChanServ, lptr->nick,
    "The successor nickname for %s has been changed to [\002%s\002]",
    cptr->name,
    fptr->nick);
} /* c_set_successor() */

static void
c_set_mlock(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  char modes[MAXLINE];
  struct Channel *chptr;
  int minus = 0,
      ii,
      argidx = 4;

  if (!(cptr = FindChan(av[1])))
    return;

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002SET <channel> MLOCK <modes>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "SET MLOCK");
    return;
  }

  RecordCommand("%s: %s!%s@%s SET [%s] MLOCK %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    av[3]);

  cptr->modes_on = cptr->modes_off = 0;
  cptr->limit = 0;
  if (cptr->key)
  {
    MyFree(cptr->key);
    cptr->key = NULL;
  }

  for (ii = 0; ii < strlen(av[3]); ++ii)
  {
    switch (av[3][ii])
    {
      case '+':
      {
        minus = 0;
        break;
      }
      case '-':
      {
        minus = 1;
        break;
      }

      case 'n':
      case 'N':
      {
        if (minus)
          cptr->modes_off |= MODE_N;
        else
          cptr->modes_on |= MODE_N;
        break;
      }

      case 't':
      case 'T':
      {
        if (minus)
          cptr->modes_off |= MODE_T;
        else
          cptr->modes_on |= MODE_T;
        break;
      }

      case 's':
      case 'S':
      {
        if (minus)
          cptr->modes_off |= MODE_S;
        else
          cptr->modes_on |= MODE_S;
        break;
      }

      case 'p':
      case 'P':
      {
        if (minus)
          cptr->modes_off |= MODE_P;
        else
          cptr->modes_on |= MODE_P;
        break;
      }

      case 'm':
      case 'M':
      {
        if (minus)
          cptr->modes_off |= MODE_M;
        else
          cptr->modes_on |= MODE_M;
        break;
      }

      case 'c':
      case 'C':
      {
        if (minus)
          cptr->modes_off |= MODE_C;
        else
          cptr->modes_on |= MODE_C;
        break;
      }

      case 'i':
      case 'I':
      {
        if (minus)
          cptr->modes_off |= MODE_I;
        else
          cptr->modes_on |= MODE_I;
        break;
      }

      case 'l':
      case 'L':
      {
        if (minus)
        {
          cptr->limit = 0;
          cptr->modes_off |= MODE_L;
        }
        else
        {
          if (ac >= (argidx + 1))
            if (IsNum(av[argidx]))
              cptr->limit = atoi(av[argidx++]);
        }
        break;
      }

      case 'k':
      case 'K':
      {
        if (minus)
        {
          if (cptr->key)
            MyFree(cptr->key);
          cptr->key = NULL;
          cptr->modes_off |= MODE_K;
        }
        else
          if (ac >= (argidx + 1))
          {
            if (cptr->key)
              MyFree(cptr->key);
            cptr->key = MyStrdup(av[argidx++]);
          }
        break;
      }

      default: break;
    } /* switch */
  }

  modes[0] = '\0';
  if (cptr->modes_off)
  {
    strcat(modes, "-");
    if (cptr->modes_off & MODE_S)
      strcat(modes, "s");
    if (cptr->modes_off & MODE_P)
      strcat(modes, "p");
    if (cptr->modes_off & MODE_N)
      strcat(modes, "n");
    if (cptr->modes_off & MODE_T)
      strcat(modes, "t");
    if (cptr->modes_off & MODE_M)
      strcat(modes, "m");
    if (cptr->modes_off & MODE_C)
      strcat(modes, "c");
    if (cptr->modes_off & MODE_I)
      strcat(modes, "i");
    if (cptr->modes_off & MODE_L)
      strcat(modes, "l");
    if (cptr->modes_off & MODE_K)
      strcat(modes, "k");
  }
  if (cptr->modes_on ||
      cptr->limit ||
      cptr->key)
  {
    strcat(modes, "+");
    if (cptr->modes_on & MODE_S)
      strcat(modes, "s");
    if (cptr->modes_on & MODE_P)
      strcat(modes, "p");
    if (cptr->modes_on & MODE_N)
      strcat(modes, "n");
    if (cptr->modes_on & MODE_T)
      strcat(modes, "t");
    if (cptr->modes_on & MODE_M)
      strcat(modes, "m");
    if (cptr->modes_on & MODE_C)
      strcat(modes, "c");
    if (cptr->modes_on & MODE_I)
      strcat(modes, "i");
    if (cptr->limit)
      strcat(modes, "l");
    if (cptr->key)
      strcat(modes, "k");

    if (cptr->limit)
    {
      char  temp[MAXLINE];

      sprintf(temp, "%s %ld", modes, cptr->limit);
      strcpy(modes, temp);
    }
    if (cptr->key)
    {
      strcat(modes, " ");
      strcat(modes, cptr->key);
    }
  }

  if ((chptr = FindChannel(cptr->name)))
  {
    toserv(":%s MODE %s %s %s\n",
      n_ChanServ,
      cptr->name,
      modes,
      ((cptr->modes_off & MODE_K) && (chptr->key)) ? chptr->key : "");
    UpdateChanModes(Me.csptr, n_ChanServ, chptr, modes);
  }

  notice(n_ChanServ, lptr->nick,
    "Now enforcing modes [\002%s\002] for %s",
    modes,
    cptr->name);
} /* c_set_mlock() */

static void
c_set_topic(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  char *topic;

  if (!(cptr = FindChan(av[1])))
    return;

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002SET <channel> TOPIC <topic|->\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "SET TOPIC");
    return;
  }

  if (!strcasecmp(av[3], "-"))
  {
    if (cptr->topic)
      MyFree(cptr->topic);
    cptr->topic = NULL;
    notice(n_ChanServ, lptr->nick,
      "The topic for %s has been cleared",
      cptr->name);
    RecordCommand("%s: %s!%s@%s SET [%s] TOPIC -",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name);
    return;
  }

  if (ac < 4)
    topic = NULL;
  else
    topic = GetString(ac - 3, av + 3);

  RecordCommand("%s: %s!%s@%s SET [%s] TOPIC %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    topic);

  if (cptr->topic)
    MyFree(cptr->topic);

  cptr->topic = topic;

  cs_SetTopic(FindChannel(cptr->name), cptr->topic);

  notice(n_ChanServ, lptr->nick,
    "The topic for %s has been set to [\002%s\002]",
    cptr->name,
    cptr->topic);
} /* c_set_topic() */

static void
c_set_entrymsg(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  char *emsg;

  if (!(cptr = FindChan(av[1])))
    return;

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002SET <channel> ENTRYMSG <message|->\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "SET ENTRYMSG");
    return;
  }

  if (!strcasecmp(av[3], "-"))
  {
    if (cptr->entrymsg)
      MyFree(cptr->entrymsg);
    cptr->entrymsg = NULL;
    notice(n_ChanServ, lptr->nick,
      "The entry message for %s has been cleared",
      cptr->name);
    RecordCommand("%s: %s!%s@%s SET [%s] ENTRYMSG -",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name);
    return;
  }

  if (ac < 4)
    emsg = NULL;
  else
    emsg = GetString(ac - 3, av + 3);

  RecordCommand("%s: %s!%s@%s SET [%s] ENTRYMSG %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    emsg);

  if (cptr->entrymsg)
    MyFree(cptr->entrymsg);

  cptr->entrymsg = emsg;

  notice(n_ChanServ, lptr->nick,
    "The entry message for %s has been set to [\002%s\002]",
    cptr->name,
    cptr->entrymsg);
} /* c_set_entrymsg() */

static void
c_set_email(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;

  if (!(cptr = FindChan(av[1])))
    return;

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002SET <channel> EMAIL <email address|->\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "SET EMAIL");
    return;
  }

  if (cptr->email)
    MyFree(cptr->email);

  if (!strcasecmp(av[3], "-"))
  {
    cptr->email = NULL;
    notice(n_ChanServ, lptr->nick,
      "The email address for %s has been cleared",
      cptr->name);
    RecordCommand("%s: %s!%s@%s SET [%s] EMAIL -",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name);
    return;
  }

  RecordCommand("%s: %s!%s@%s SET [%s] EMAIL %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    av[3]);

  cptr->email = MyStrdup(av[3]);

  notice(n_ChanServ, lptr->nick,
    "The email address for %s has been set to [\002%s\002]",
    cptr->name,
    cptr->email);
} /* c_set_email() */

static void
c_set_url(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;

  if (!(cptr = FindChan(av[1])))
    return;

  if (ac < 4)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002SET <channel> URL <http://url|->\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "SET URL");
    return;
  }

  if (cptr->url)
    MyFree(cptr->url);

  if (!strcasecmp(av[3], "-"))
  {
    cptr->url = NULL;
    notice(n_ChanServ, lptr->nick,
      "The url for %s has been cleared",
      cptr->name);
    RecordCommand("%s: %s!%s@%s SET [%s] URL -",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name);
    return;
  }

  RecordCommand("%s: %s!%s@%s SET [%s] URL %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    av[3]);

  cptr->url = MyStrdup(av[3]);

  notice(n_ChanServ, lptr->nick,
    "The url for %s has been set to [\002%s\002]",
    cptr->name,
    cptr->url);
} /* c_set_url() */

/*
c_invite()
  Invite lptr->nick to channel av[1]
*/

static void
c_invite(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  struct Channel *chptr;

  if (ac < 2)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002INVITE <channel>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "INVITE");
    return;
  }

  if (!(cptr = FindChan(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_CH_NOT_REGGED,
      av[1]);
    return;
  }

  if (cptr->flags & (CS_FORBID | CS_FORGET))
  {
    notice(n_ChanServ, lptr->nick,
      "[\002%s\002] is a forbidden channel",
      cptr->name);
    return;
  }

  if (!HasAccess(cptr, lptr, CA_CMDINVITE))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_NEED_ACCESS,
      cptr->access_lvl[CA_CMDINVITE],
      "INVITE",
      cptr->name);

    RecordCommand("%s: %s!%s@%s failed INVITE [%s]",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name);

    return;
  }

  RecordCommand("%s: %s!%s@%s INVITE [%s]",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name);

  if (IsChannelMember(FindChannel(av[1]), lptr))
  {
    notice(n_ChanServ, lptr->nick,
      "You are already on [\002%s\002]",
      cptr->name);
    return;
  }

  /*
   * Everything checks out - invite
   *
   * For ircd-hybrid 6+, ChanServ need to be inside the channel
   * to invite - have ChanServ join and part
   */

  if ((chptr = FindChannel(cptr->name)))
  {
    toserv(":%s SJOIN %ld %s + :@%s\n:%s INVITE %s %s\n",
      Me.name,
      chptr->since,
      cptr->name,
      n_ChanServ,
      n_ChanServ,
      lptr->nick,
      cptr->name);

    toserv(":%s NOTICE %s :Inviting %s\n:%s PART %s\n",
      n_ChanServ,
      cptr->name,
      lptr->nick,
      n_ChanServ,
      cptr->name);
  }
  else
  {
    notice(n_ChanServ, lptr->nick,
      "The channel [\002%s\002] is empty",
      cptr->name);
  }
} /* c_invite() */

/*
c_op()
  Op lptr->nick on channel av[1]
*/

static void
c_op(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  struct Channel *chptr;
  char *onicks, *dnicks;
  struct UserChannel *uchan;
  struct ChannelUser *cuser;
  struct Luser *currlptr;

  if (ac < 2)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002OP <channel> [nicks]\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "OP");
    return;
  }

  if (!strcasecmp(av[1], "ALL"))
    cptr = NULL;
  else if (!(cptr = FindChan(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_CH_NOT_REGGED,
      av[1]);
    return;
  }

  if (!cptr)
  {
    /*
     * They want to be opped in all channels they are currently
     * in.
     */

    for (uchan = lptr->firstchan; uchan; uchan = uchan->next)
    {
      if (HasAccess(FindChan(uchan->chptr->name), lptr, CA_CMDOP) &&
          !(uchan->flags & CH_OPPED))
      {
        toserv(":%s MODE %s +o %s\n",
          n_ChanServ,
          uchan->chptr->name,
          lptr->nick);
        uchan->flags |= CH_OPPED;

        if ((cuser = FindUserByChannel(uchan->chptr, lptr)))
          cuser->flags |= CH_OPPED;
      }
    }

    notice(n_ChanServ, lptr->nick,
      "You have been opped on all channels you have access to");

    RecordCommand("%s: %s!%s@%s OP ALL",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname);

    return;
  }

  if (!HasAccess(cptr, lptr, CA_CMDOP))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_NEED_ACCESS,
      cptr->access_lvl[CA_CMDOP],
      "OP",
      cptr->name);
    RecordCommand("%s: %s!%s@%s failed OP [%s]",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name);
    return;
  }

  chptr = FindChannel(av[1]);
  if (ac < 3)
  {
    if (HasFlag(lptr->nick, NS_NOCHANOPS))
      return;

    onicks = MyStrdup(lptr->nick);
    dnicks = MyStrdup("");
    if (!IsChannelMember(chptr, lptr))
    {
      notice(n_ChanServ, lptr->nick,
        "You are not on [\002%s\002]",
        cptr->name);
      return;
    }

    if (IsChannelOp(chptr, lptr))
    {
      notice(n_ChanServ, lptr->nick,
        "You are already opped on [\002%s\002]",
        cptr->name);
      return;
    }
  }
  else
  {
    int ii, arc;
    char *tempnix, *tempptr, **arv;

    /* they want to op other people */
    tempnix = GetString(ac - 2, av + 2);

    tempptr = tempnix;
    arc = SplitBuf(tempnix, &arv);
    onicks = (char *) MyMalloc(sizeof(char));
    onicks[0] = '\0';
    dnicks = (char *) MyMalloc(sizeof(char));
    dnicks[0] = '\0';
    for (ii = 0; ii < arc; ii++)
    {
      if (*arv[ii] == '-')
        currlptr = FindClient(arv[ii] + 1);
      else
        currlptr = FindClient(arv[ii]);

      if (!IsChannelMember(chptr, currlptr))
        continue;


      if ((arv[ii][0] == '-') && (IsChannelOp(chptr, currlptr)))
      {
        if (!strcasecmp(arv[ii] + 1, n_ChanServ) ||
	    !strcasecmp(arv[ii] + 1, n_OperServ))
	      continue;
        dnicks = (char *) MyRealloc(dnicks, strlen(dnicks) + strlen(arv[ii] + 1) + (2 * sizeof(char)));
        strcat(dnicks, arv[ii] + 1);
        strcat(dnicks, " ");
      }
      else if (!IsChannelOp(chptr, currlptr))
      {
        if ((cptr->flags & CS_SECUREOPS) &&
            (!HasAccess(cptr, FindClient(arv[ii]), CA_AUTOOP)))
          continue;

        /*
         * Don't op them if they are flagged
         */
        if (HasFlag(arv[ii], NS_NOCHANOPS))
          continue;
        
        onicks = (char *) MyRealloc(onicks, strlen(onicks) + strlen(arv[ii]) + (2 * sizeof(char)));
        strcat(onicks, arv[ii]);
        strcat(onicks, " ");
      }
    }

    MyFree(tempptr);
    MyFree(arv);
  }

  SetModes(n_ChanServ, 0, 'o', chptr, dnicks);
  SetModes(n_ChanServ, 1, 'o', chptr, onicks);

  RecordCommand("%s: %s!%s@%s OP [%s]%s%s%s%s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    strlen(onicks) ? " [+] " : "",
    strlen(onicks) ? onicks : "",
    strlen(dnicks) ? " [-] " : "",
    strlen(dnicks) ? dnicks : "");

  MyFree(onicks);
  MyFree(dnicks);
} /* c_op() */

/*
c_voice()
  (De)Voice nicks on channel av[1]
*/

static void
c_voice(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  struct Channel *chptr;
  char *vnicks, *dnicks;

  if (ac < 2)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002VOICE <channel>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "VOICE");
    return;
  }

  if (!(cptr = FindChan(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_CH_NOT_REGGED,
      av[1]);
    return;
  }

  if (!HasAccess(cptr, lptr, CA_CMDVOICE))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_NEED_ACCESS,
      cptr->access_lvl[CA_CMDVOICE],
      "VOICE",
      cptr->name);
    RecordCommand("%s: %s!%s@%s failed VOICE [%s]",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name);
    return;
  }

  chptr = FindChannel(av[1]);
  if (ac < 3)
  {
    vnicks = MyStrdup(lptr->nick);
    dnicks = MyStrdup("");
    if (!IsChannelMember(chptr, lptr))
    {
      notice(n_ChanServ, lptr->nick,
        "You are not on [\002%s\002]",
        cptr->name);
      MyFree(vnicks);
      MyFree(dnicks);
      return;
    }
  }
  else
  {
    int ii, arc;
    char *tempnix, *tempptr, **arv;
    struct Luser *currlptr;

    /* they want to voice other people */
    tempnix = GetString(ac - 2, av + 2);

    tempptr = tempnix;
    arc = SplitBuf(tempnix, &arv);
    vnicks = (char *) MyMalloc(sizeof(char));
    vnicks[0] = '\0';
    dnicks = (char *) MyMalloc(sizeof(char));
    dnicks[0] = '\0';
    for (ii = 0; ii < arc; ii++)
    {
      if (*arv[ii] == '-')
        currlptr = FindClient(arv[ii] + 1);
      else
        currlptr = FindClient(arv[ii]);

      if (!IsChannelMember(chptr, currlptr))
        continue;

      if (arv[ii][0] == '-')
      {
        dnicks = (char *) MyRealloc(dnicks, strlen(dnicks) + strlen(arv[ii] + 1) + (2 * sizeof(char)));
        strcat(dnicks, arv[ii] + 1);
        strcat(dnicks, " ");
      }
      else
      {
        vnicks = (char *) MyRealloc(vnicks, strlen(vnicks) + strlen(arv[ii]) + (2 * sizeof(char)));
        strcat(vnicks, arv[ii]);
        strcat(vnicks, " ");
      }
    }

    MyFree(tempptr);
    MyFree(arv);
  }

  SetModes(n_ChanServ, 0, 'v', chptr, dnicks);
  SetModes(n_ChanServ, 1, 'v', chptr, vnicks);

  RecordCommand("%s: %s!%s@%s VOICE [%s]%s%s%s%s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    strlen(vnicks) ? " [+] " : "",
    strlen(vnicks) ? vnicks : "",
    strlen(dnicks) ? " [-] " : "",
    strlen(dnicks) ? dnicks : "");

  MyFree(vnicks);
  MyFree(dnicks);

  return;
} /* c_voice() */

/*
c_unban()
  Unban all hostmasks matching lptr->nick on channel av[1]
*/

static void
c_unban(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  char chkstr[MAXLINE], chkstr2[MAXLINE], *bans, ipstr[16];
  struct Channel *chptr;
  struct ChannelBan *bptr;
  struct hostent *hent;
  int all = 0;

  if (ac < 2)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002UNBAN <channel> [ALL]\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "UNBAN");
    return;
  }

  if (!(cptr = FindChan(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_CH_NOT_REGGED,
      av[1]);
    return;
  }

  if (!HasAccess(cptr, lptr, CA_CMDUNBAN))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_NEED_ACCESS,
      cptr->access_lvl[CA_CMDUNBAN],
      "UNBAN",
      cptr->name);
    RecordCommand("%s: %s!%s@%s failed UNBAN [%s] %s",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      (ac >= 3) ? av[2] : "");
    return;
  }

  sprintf(chkstr, "%s!%s@%s",
    lptr->nick,
    lptr->username,
    lptr->hostname);


  if ((hent = gethostbyname(lptr->hostname)) != NULL) {
      inet_ntop(2, hent->h_addr_list[0], ipstr, 15);
      sprintf(chkstr2, "%s!%s@%s",
	lptr->nick,
	lptr->username,
	ipstr);
  } else
      chkstr2[0] = 0;

  if (ac >= 3)
    if (!strcasecmp(av[2], "ALL"))
    {
      if (!HasAccess(cptr, lptr, CA_SUPEROP))
      {
        notice(n_ChanServ, lptr->nick,
          "SuperOp privileges are required to unban \002ALL\002 on %s",
          cptr->name);
        return;
      }
      else
        all = 1;
    }

  if (!(chptr = FindChannel(cptr->name)))
  {
    notice(n_ChanServ, lptr->nick,
      "The channel [\002%s\002] is not active",
      cptr->name);
    return;
  }

  bans = (char *) MyMalloc(sizeof(char));
  bans[0] = '\0';
  for (bptr = chptr->firstban; bptr; bptr = bptr->next)
  {
    if (all || match(bptr->mask, chkstr) || (chkstr2[0] && match(bptr->mask, chkstr2)))
    {
      bans = (char *) MyRealloc(bans, strlen(bans) + strlen(bptr->mask) + (2 * sizeof(char)));
      strcat(bans, bptr->mask);
      strcat(bans, " ");
    }
  }

  SetModes(n_ChanServ, 0, 'b', chptr, bans);

  MyFree(bans);

  notice(n_ChanServ, lptr->nick,
    "All bans matching [\002%s\002] have been cleared on %s",
    (all) ? "*!*@*" : chkstr,
    cptr->name);

  RecordCommand("%s: %s!%s@%s UNBAN [%s] %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    (all) ? "ALL" : "");

  return;
} /* c_unban() */

/*
c_info()
  Display info about channel av[1]
*/

static void
c_info(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  char buf[MAXLINE];
  int online;
  struct NickInfo *ni;

  if (ac < 2)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002INFO <channel>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "INFO");
    return;
  }

  if (!(cptr = FindChan(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_CH_NOT_REGGED,
      av[1]);
    return;
  }

  if (((cptr->flags & CS_PRIVATE) || (cptr->flags & (CS_FORBID | CS_FORGET)))
  #ifdef EMPOWERADMINS
      && !(IsValidAdmin(lptr) || IsOperator(lptr)))
  #else
      )
  #endif
  {
    notice(n_ChanServ, lptr->nick,
      "Channel [\002%s\002] is private",
      cptr->name);
    return;
  }

  if ((cptr->flags & (CS_FORBID | CS_FORGET))
    && (IsValidAdmin(lptr) || IsOperator(lptr))) {
    notice(n_ChanServ, lptr->nick,
	"     Channel: %s",
	cptr->name);

    if (cptr->flags & CS_TFORBID) {
	notice(n_ChanServ, lptr->nick,
        	"     Options: Temporary Forbidden");
	notice(n_ChanServ, lptr->nick,
		"              -> at %s",
		ctime(&cptr->created));
	if (cptr->topic) {
	    notice(n_ChanServ, lptr->nick,
        	"              -> by %s", cptr->topic);
	}
	notice(n_ChanServ, lptr->nick,
		"              -> expires at %s",
		ctime(&cptr->lastused));

	return;
    }

    if (cptr->flags & CS_FORBID) {
	notice(n_ChanServ, lptr->nick,
        	"     Options: Permanently Forbidden");
	notice(n_ChanServ, lptr->nick,
		"              -> at %s",
		ctime(&cptr->created));
	if (cptr->topic) {
	    notice(n_ChanServ, lptr->nick,
        	"              -> by %s", cptr->topic);
	}
	return;
    }

    if (cptr->flags & CS_FORGET) {
	if (cptr->topic) {
	    notice(n_ChanServ, lptr->nick,
        	"     Options: Forgotten by %s",
		cptr->topic);
	} else {
	    notice(n_ChanServ, lptr->nick,
        	"     Options: Forgotten");
	}
	return;
    }
    return;
  }

  RecordCommand("%s: %s!%s@%s INFO [%s]",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name);

  online = 0;
  if ((ni = FindNick(cptr->founder)))
    if (ni->flags & NS_IDENTIFIED)
      online = 1;

  notice(n_ChanServ, lptr->nick,
    "     Channel: %s",
    cptr->name);
  notice(n_ChanServ, lptr->nick,
    "     Founder: %s %s",
    cptr->founder,
    online ? "<< ONLINE >>" : "");
  notice(n_ChanServ, lptr->nick,
    "  Registered: %s ago",
    timeago(cptr->created, 1));
  if (!FindChannel(cptr->name))
    notice(n_ChanServ, lptr->nick,
      "   Last Used: %s ago",
      timeago(cptr->lastused, 1));

  if (cptr->topic)
    notice(n_ChanServ, lptr->nick,
      "       Topic: %s",
      cptr->topic);

  if (cptr->email)
    notice(n_ChanServ, lptr->nick,
      "       Email: %s",
      cptr->email);

  if (cptr->url)
    notice(n_ChanServ, lptr->nick,
      "         Url: %s",
      cptr->url);

  buf[0] = '\0';
  if (cptr->flags & CS_TOPICLOCK)
    strcat(buf, "TopicLock, ");
  if (cptr->flags & CS_SECURE)
    strcat(buf, "Secure, ");
  if (cptr->flags & CS_SECUREOPS)
    strcat(buf, "SecureOps, ");
  if (cptr->flags & CS_GUARD)
    strcat(buf, "ChanGuard, ");
  if (cptr->flags & CS_RESTRICTED)
    strcat(buf, "Restricted, ");
  if (cptr->flags & CS_PRIVATE)
    strcat(buf, "Private, ");
  if (cptr->flags & CS_FORBID)
    strcat(buf, "Forbidden, ");
  if (cptr->flags & CS_FORGET)
    strcat(buf, "Forgotten, ");
  if (cptr->flags & CS_NOEXPIRE)
    strcat(buf, "NoExpire, ");

  if (*buf)
  {
    /* kill the trailing "," */
    buf[strlen(buf) - 2] = '\0';
    notice(n_ChanServ, lptr->nick,
      "     Options: %s",
      buf);
  }

  buf[0] = '\0';
  if (cptr->modes_off)
  {
    strcat(buf, "-");
    if (cptr->modes_off & MODE_S)
      strcat(buf, "s");
    if (cptr->modes_off & MODE_P)
      strcat(buf, "p");
    if (cptr->modes_off & MODE_N)
      strcat(buf, "n");
    if (cptr->modes_off & MODE_T)
      strcat(buf, "t");
    if (cptr->modes_off & MODE_M)
      strcat(buf, "m");
    if (cptr->modes_off & MODE_C)
      strcat(buf, "c");
    if (cptr->modes_off & MODE_I)
      strcat(buf, "i");
    if (cptr->modes_off & MODE_L)
      strcat(buf, "l");
    if (cptr->modes_off & MODE_K)
      strcat(buf, "k");
  }
  if (cptr->modes_on || cptr->limit || cptr->key)
  {
    strcat(buf, "+");
    if (cptr->modes_on & MODE_S)
      strcat(buf, "s");
    if (cptr->modes_on & MODE_P)
      strcat(buf, "p");
    if (cptr->modes_on & MODE_N)
      strcat(buf, "n");
    if (cptr->modes_on & MODE_T)
      strcat(buf, "t");
    if (cptr->modes_on & MODE_M)
      strcat(buf, "m");
    if (cptr->modes_on & MODE_C)
      strcat(buf, "c");
    if (cptr->modes_on & MODE_I)
      strcat(buf, "i");
    if (cptr->limit)
      strcat(buf, "l");
    if (cptr->key)
      strcat(buf, "k");
  }
  if (buf[0])
    notice(n_ChanServ, lptr->nick,
      "   Mode Lock: %s",
      buf);
} /* c_info() */

/*
c_clear()
  Clear modes/bans/ops/voices from a channel
*/

static void
c_clear(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  struct Command *cmdptr;

  if (ac < 3)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002CLEAR <channel> {OPS|VOICES|MODES|BANS|GECOSBANS|ALL|USERS}\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "CLEAR");
    return;
  }

  if (!(cptr = FindChan(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_CH_NOT_REGGED,
      av[1]);
    return;
  }

  if (!HasAccess(cptr, lptr, CA_CMDCLEAR))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_NEED_ACCESS,
      cptr->access_lvl[CA_CMDCLEAR],
      "CLEAR",
      cptr->name);
    RecordCommand("%s: %s!%s@%s failed CLEAR [%s] %s",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      av[2]);
    return;
  }

  cmdptr = GetCommand(clearcmds, av[2]);

  if (cmdptr && (cmdptr != (struct Command *) -1))
  {
    RecordCommand("%s: %s!%s@%s CLEAR [%s] [%s]",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      cptr->name,
      cmdptr->cmd);

    /* call cmdptr->func to execute command */
    (*cmdptr->func)(lptr, nptr, ac, av);
  }
  else
  {
    /* the option they gave was not valid */
    notice(n_ChanServ, lptr->nick,
      "%s switch [\002%s\002]",
      (cmdptr == (struct Command *) -1) ? "Ambiguous" : "Unknown",
      av[2]);
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "CLEAR");
    return;
  }

  notice(n_ChanServ, lptr->nick,
    "[%s] have been cleared on [\002%s\002]",
    cmdptr->cmd,
    cptr->name);
} /* c_clear() */

static void
c_clear_ops(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChannelUser *cuser;
  char *ops;
  struct Channel *chptr;

  if (!(chptr = FindChannel(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      "The channel [\002%s\002] is not active",
      av[1]);
    return;
  }

  ops = (char *) MyMalloc(sizeof(char));
  ops[0] = '\0';
  for (cuser = chptr->firstuser; cuser; cuser = cuser->next)
  {
    if (FindService(cuser->lptr) ||
        !(cuser->flags & CH_OPPED))
      continue;

    ops = (char *) MyRealloc(ops, strlen(ops) + strlen(cuser->lptr->nick) + (2 * sizeof(char)));
    strcat(ops, cuser->lptr->nick);
    strcat(ops, " ");
  }

  SetModes(n_ChanServ, 0, 'o', chptr, ops);

  MyFree(ops);
} /* c_clear_ops() */

static void
c_clear_voices(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct Channel *chptr;
  struct ChannelUser *cuser;
  char *voices;

  if (!(chptr = FindChannel(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      "The channel [\002%s\002] is not active",
      av[1]);
    return;
  }

  voices = (char *) MyMalloc(sizeof(char));
  voices[0] = '\0';
  for (cuser = chptr->firstuser; cuser; cuser = cuser->next)
  {
    if (FindService(cuser->lptr) ||
        !(cuser->flags & CH_VOICED))
      continue;

    voices = (char *) MyRealloc(voices, strlen(voices) + strlen(cuser->lptr->nick) + (2 * sizeof(char)));
    strcat(voices, cuser->lptr->nick);
    strcat(voices, " ");
  }

  SetModes(n_ChanServ, 0, 'v', chptr, voices);

  MyFree(voices);
} /* c_clear_voices() */

static void
c_clear_modes(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  char modes[MAXLINE];
  struct Channel *chptr;

  if (!(chptr = FindChannel(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      "The channel [\002%s\002] is not active",
      av[1]);
    return;
  }

  strcpy(modes, "-");

  if (chptr->modes & MODE_S)
    strcat(modes, "s");
  if (chptr->modes & MODE_P)
    strcat(modes, "p");
  if (chptr->modes & MODE_N)
    strcat(modes, "n");
  if (chptr->modes & MODE_T)
    strcat(modes, "t");
  if (chptr->modes & MODE_M)
    strcat(modes, "m");
  if (chptr->modes & MODE_C)
    strcat(modes, "c");
  if (chptr->modes & MODE_I)
    strcat(modes, "i");
  if (chptr->limit)
    strcat(modes, "l");
  if (chptr->key)
  {
    strcat(modes, "k ");
    strcat(modes, chptr->key);
  }

  toserv(":%s MODE %s %s\n",
    n_ChanServ,
    chptr->name,
    modes);

  UpdateChanModes(Me.csptr, n_ChanServ, chptr, modes);
} /* c_clear_modes() */

static void
c_clear_bans(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct Channel *chptr;
  struct ChannelBan *bptr;
  char *bans;

  if (!(chptr = FindChannel(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      "The channel [\002%s\002] is not active",
      av[1]);
    return;
  }

  bans = (char *) MyMalloc(sizeof(char));
  bans[0] = '\0';

  for (bptr = chptr->firstban; bptr; bptr = bptr->next)
  {
    bans = (char *) MyRealloc(bans, strlen(bans) + strlen(bptr->mask) + (2 * sizeof(char)));
    strcat(bans, bptr->mask);
    strcat(bans, " ");
  }

  SetModes(n_ChanServ, 0, 'b', chptr, bans);

  MyFree(bans);
} /* c_clear_bans() */

static void
c_clear_users(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct Channel *chptr;
  struct ChannelUser *cuser;
  char *knicks,
       *ops; /* deop before kicking */

  if (!(chptr = FindChannel(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      "The channel [\002%s\002] is not active",
      av[1]);
    return;
  }

  knicks = (char *) MyMalloc(sizeof(char));
  knicks[0] = '\0';
  ops = (char *) MyMalloc(sizeof(char));
  ops[0] = '\0';

  for (cuser = chptr->firstuser; cuser; cuser = cuser->next)
  {
    if (FindService(cuser->lptr))
      continue;

    if (cuser->flags & CH_OPPED)
    {
      ops = (char *) MyRealloc(ops, strlen(ops) + strlen(cuser->lptr->nick) + (2 * sizeof(char)));
      strcat(ops, cuser->lptr->nick);
      strcat(ops, " ");
    }

    knicks = (char *) MyRealloc(knicks, strlen(knicks) + strlen(cuser->lptr->nick) + (2 * sizeof(char)));
    strcat(knicks, cuser->lptr->nick);
    strcat(knicks, " ");
  }

  SetModes(n_ChanServ, 0, 'o', chptr, ops);
  MyFree(ops);

  toserv(":%s MODE %s +i\n",
    n_ChanServ,
    chptr->name);

  UpdateChanModes(Me.csptr, n_ChanServ, chptr, "+i");

  KickBan(0, n_ChanServ, chptr, knicks, "Clearing Users");
  MyFree(knicks);

  toserv(":%s MODE %s -i\n",
    n_ChanServ,
    chptr->name);

  UpdateChanModes(Me.csptr, n_ChanServ, chptr, "-i");
} /* c_clear_users() */

static void
c_clear_all(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  c_clear_ops(lptr, nptr, ac, av);
  c_clear_voices(lptr, nptr, ac, av);
  c_clear_modes(lptr, nptr, ac, av);
  c_clear_bans(lptr, nptr, ac, av);
  c_clear_gecos_bans( lptr, nptr, ac, av);
} /* c_clear_all() */

#ifdef UNIBG_SERVICES
static void
c_clear_gecos_bans(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct Channel *chptr;
  struct ChannelGecosBan *bptr;
  char *bans;
  
  if (!(chptr = FindChannel(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      "The channel [\002%s\002] is not active",
      av[1]);
    return;
  }

  bans = (char *) MyMalloc(sizeof(char));
  bans[0] = '\0';

  for (bptr = chptr->firstgecosban; bptr; bptr = bptr->next)
  {
    bans = (char *) MyRealloc(bans, strlen(bans) + strlen(bptr->mask) + (2 * sizeof(char)));
    strcat(bans, bptr->mask);
    strcat(bans, " ");    
  }

  SetModes(n_ChanServ, 0, 'd', chptr, bans);

  MyFree(bans);
} /* c_clear_gecos_bans() */
#endif /* UNIBG_SERVICES */

#ifdef EMPOWERADMINS

/*
c_forbid()
  Prevent anyone from using channel av[1]
*/

static void
c_forbid(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  char *reason;
  char freason[MAXLINE];
  long expires;

  if (ac < 3)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002FORBID [time] <channel> <comment>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "FORBID");
    return;
  }

  if ((expires = timestr(av[1])) <= 0) {
      /* perm forbid */
      
      reason = GetString(ac - 2, av + 2);
  
      RecordCommand("%s: %s!%s@%s FORBID %s [%s]",
        n_ChanServ,
        lptr->nick,
        lptr->username,
        lptr->hostname,
        av[1],
        reason);

      if ((cptr = FindChan(av[1])))
      {
        /* the channel is already registered */
        if (cptr->flags & CS_FORBID)
        {
          notice(n_ChanServ, lptr->nick,
            "The channel [\002%s\002] is already forbidden",
            cptr->name);
          return;
        }

        /*
         * delete channel to kill the access/akick/founder list etc
         */
        DeleteChan(cptr);
      }
      else
      {
        if (av[1][0] != '#')
        {
          notice(n_ChanServ, lptr->nick,
            "[\002%s\002] is an invalid channel",
            av[1]);
          return;
        }
      }
      cptr = MakeChan();
      cptr->name = MyStrdup(av[1]);
      cptr->created = time(NULL);
      cptr->flags = CS_FORBID;

      notice(n_ChanServ, lptr->nick,
        "The channel [\002%s\002] is now forbidden",
            av[1]);

      toserv(":%s WALLOPS :%s!%s@%s FORBID %s [%s]\n",
          n_ChanServ,
          lptr->nick,
          lptr->username,
          lptr->hostname,
          av[1],
          reason);
  }   /* perm forbid */
  else
  {   /* timed forbid */

      if (ac < 4)
      {
        notice(n_ChanServ, lptr->nick,
          "Syntax: \002FORBID [time] <channel> <comment>\002");
        notice(n_ChanServ, lptr->nick,
          ERR_MORE_INFO,
          n_ChanServ,
          "FORBID");
        return;
      }

      reason = GetString(ac - 3, av + 3);

      RecordCommand("%s: %s!%s@%s FORBID %s [%s] (%s)",
        n_ChanServ,
        lptr->nick,
        lptr->username,
        lptr->hostname,
        av[2],
        reason,
	av[1]);

      if ((cptr = FindChan(av[2])))
      {
        /* the channel is already registered */
        if (cptr->flags & CS_FORBID)
        {
          notice(n_ChanServ, lptr->nick,
            "The channel [\002%s\002] is already forbidden",
            cptr->name);
          return;
        }

        /*
         * delete channel to kill the access/akick/founder list etc
         */
        DeleteChan(cptr);
      }
      else
      {
        if (av[2][0] != '#')
        {
          notice(n_ChanServ, lptr->nick,
            "[\002%s\002] is an invalid channel",
            av[2]);
          return;
        }
      }
      cptr = MakeChan();
      cptr->name = MyStrdup(av[2]);
      cptr->created = time(NULL);
      cptr->lastused = time(NULL) + expires;
      cptr->flags |= CS_FORBID;
      cptr->flags |= CS_TFORBID;

      notice(n_ChanServ, lptr->nick,
        "The channel [\002%s\002] is now forbidden",
            av[2]);

      toserv(":%s WALLOPS :%s!%s@%s FORBID %s [%s] (%s)\n",
          n_ChanServ,
          lptr->nick,
          lptr->username,
          lptr->hostname,
          av[2],
          reason,
	  av[1]);
  } /* timed forbid */

  SetChanDefAccess(cptr);
    
  sprintf(freason, "%s: %s", lptr->nick, reason);
  cptr->topic = MyStrdup(freason);

  AddChan(cptr);
  cs_join(cptr);

} /* c_forbid() */


/*
c_setpass()
  Change the channel (av[1]) password to av[2]
*/

static void
c_setpass(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;

  if (ac < 3)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002SETPASS <channel> <password>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "SETPASS");
    return;
  }

  RecordCommand("%s: %s!%s@%s SETPASS [%s] ...",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    av[1]);

  if (!(cptr = FindChan(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_CH_NOT_REGGED,
      av[1]);
    return;
  }

  if (!ChangeChanPass(cptr, av[2]))
  {
    notice(n_ChanServ, lptr->nick,
      "Password change failed");
    RecordCommand("%s: Failed to change password for [%s] to [%s] (SETPASS)",
      n_ChanServ,
      cptr->name,
      av[2]);
    return;
  }

  notice(n_ChanServ, lptr->nick,
    "Founder password for %s has been changed to [\002%s\002]",
    cptr->name,
    av[2]);
  toserv(":%s WALLOPS :%s!%s@%s SETPASS [%s]\n",
      n_ChanServ,
      lptr->nick,
      lptr->username,
      lptr->hostname,
      av[1]);
} /* c_setpass() */

/*
c_status()
  Give the current access level of nick av[2] on channel av[1]
*/

static void
c_status(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  struct Luser *tmpuser;

  if (ac < 3)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002STATUS <channel> <nickname>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "STATUS");
    return;
  }

  if (!(cptr = FindChan(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_CH_NOT_REGGED,
      av[1]);
    return;
  }

  if (!(tmpuser = FindClient(av[2])))
  {
    notice(n_ChanServ, lptr->nick,
      "[\002%s\002] is not currently online",
      av[2]);
    return;
  }

  notice(n_ChanServ, lptr->nick,
    "%s has an access level of [\002%d\002] on %s",
    tmpuser->nick,
    GetAccess(cptr, tmpuser),
    cptr->name);

  RecordCommand("%s: %s!%s@%s STATUS [%s] %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    cptr->name,
    tmpuser->nick);
} /* c_status() */

/*
c_forget()
  "forget" that a channel exists (ie: allow users to enter the channel,
but don't let anyone register it)
*/

static void
c_forget(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;
  struct Channel *chptr;
  char *reason;
  char freason[MAXLINE];
  
  if (ac < 3)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002FORGET <channel> <comment>\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "FORGET");
    return;
  }
  
  reason = GetString(ac - 2, av + 2);

  RecordCommand("%s: %s!%s@%s FORGET [%s] - %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    av[1],
    reason);

  if ((cptr = FindChan(av[1])))
  {
    /* the channel is already registered */
    if (cptr->flags & CS_FORGET)
    {
      notice(n_ChanServ, lptr->nick,
        "The channel [\002%s\002] is already forgotten",
        cptr->name);
      return;
    }

    chptr = FindChannel(cptr->name);
    if (IsChannelMember(chptr, Me.csptr))
      cs_part(chptr);

    /*
     * delete channel to kill the access/akick/founder list etc
     */
    DeleteChan(cptr);
  }
  else
  {
    if (av[1][0] != '#')
    {
      notice(n_ChanServ, lptr->nick,
        "[\002%s\002] is an invalid channel",
        av[1]);
      return;
    }
  }

  cptr = MakeChan();
  cptr->name = MyStrdup(av[1]);
  cptr->created = time(NULL);
  cptr->flags = CS_FORGET;

  SetChanDefAccess(cptr);

  sprintf(freason, "%s: %s", lptr->nick, reason);
  cptr->topic = MyStrdup(freason);
  
  AddChan(cptr);

  notice(n_ChanServ, lptr->nick,
    "The channel [\002%s\002] is now forgotten",
    av[1]);

  toserv(":%s WALLOPS :%s!%s@%s FORGET [%s] - %s\n",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    av[1],
    reason);
} /* c_forget() */

/*
c_noexpire()
 Prevent channel av[1] from expiring
*/

static void
c_noexpire(struct Luser *lptr, struct NickInfo *nptr, int ac, char **av)

{
  struct ChanInfo *cptr;

  if (ac < 2)
  {
    notice(n_ChanServ, lptr->nick,
      "Syntax: \002NOEXPIRE <channel> {ON|OFF}\002");
    notice(n_ChanServ, lptr->nick,
      ERR_MORE_INFO,
      n_ChanServ,
      "NOEXPIRE");
    return;
  }

  if (!(cptr = FindChan(av[1])))
  {
    notice(n_ChanServ, lptr->nick,
      ERR_CH_NOT_REGGED,
      av[1]);
    return;
  }

  if (ac < 3)
  {
    notice(n_ChanServ, lptr->nick,
      "NoExpire for channel %s is [\002%s\002]",
      cptr->name,
      (cptr->flags & CS_NOEXPIRE) ? "ON" : "OFF");
    return;
  }

  RecordCommand("%s: %s!%s@%s NOEXPIRE [%s] %s",
    n_ChanServ,
    lptr->nick,
    lptr->username,
    lptr->hostname,
    av[1],
    (ac < 3) ? "" : StrToupper(av[2]));

  if (!strcasecmp(av[2], "ON"))
  {
    toserv(":%s WALLOPS :%s!%s@%s NOEXPIRE [%s] ON\n",
	n_ChanServ,
	lptr->nick,
	lptr->username,
	lptr->hostname,
	cptr->name);
    cptr->flags |= CS_NOEXPIRE;
    notice(n_ChanServ, lptr->nick,
      "Toggled NoExpire for channel %s [\002ON\002]",
      cptr->name);
    ++Network->NoExpireChans;
    return;
  }

  if (!strcasecmp(av[2], "OFF"))
  {
    toserv(":%s WALLOPS :%s!%s@%s NOEXPIRE [%s] OFF\n",
	n_ChanServ,
	lptr->nick,
	lptr->username,
	lptr->hostname,
	cptr->name);
    cptr->flags &= ~CS_NOEXPIRE;
    notice(n_ChanServ, lptr->nick,
      "Toggled NoExpire for channel %s [\002OFF\002]",
      cptr->name);
    --Network->NoExpireChans;
    return;
  }

  /* user gave an unknown param */
  notice(n_ChanServ, lptr->nick,
    "Syntax: \002NOEXPIRE <channel> {ON|OFF}\002");
  notice(n_ChanServ, lptr->nick,
    ERR_MORE_INFO,
    n_ChanServ,
    "NOEXPIRE");
} /* c_noexpire() */

#endif /* EMPOWERADMINS */

#endif /* defined(NICKSERVICES) && defined(CHANNELSERVICES) */
